library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use std.textio.all;
use work.notech.all;

entity v is
  port(
-- I/O to connect to operator blocs
    BUS_DONNEES_1_v : inout word;
    BUS_DONNEES_2_v : inout word;
    BUS_DONNEES_3_v : inout word;
    enable, rstb, clk : in std_logic
 );
end;
architecture v_arch of v is
  attribute period:string;
  attribute period of clk : signal is "10 ns";
  -- states of the FSM
  type state_type is (S0,S1,S2,S3,S4,S5,S6,S7,S8,
                      S9,S10,S11,S12,S13,S14,S15,S16,S17,S18);
  --sequential coding for state_type
  --attribute enum_encoding : string;
  --attribute enum_encoding of state_type : type is "sequential";
  --sequentail coding for state_type for synplify
  --attribute syn_encoding : string;
  --attribute syn_encoding of state_type : type is "sequential";
  signal next_state : state_type;
  signal state : state_type;
  -- signals to connect registers to operators
  signal comp_0_add_op_a : word;
  signal comp_0_add_op_b : word;
  signal comp_0_add_op_o : word;
  signal comp_2_mul_op_a : word;
  signal comp_2_mul_op_b : word;
  signal comp_2_mul_op_o : word;
  signal comp_3_mul_op_a : word;
  signal comp_3_mul_op_b : word;
  signal comp_3_mul_op_o : word;
  signal comp_4_mul_op_a : word;
  signal comp_4_mul_op_b : word;
  signal comp_4_mul_op_o : word;
  signal comp_5_mul_op_a : word;
  signal comp_5_mul_op_b : word;
  signal comp_5_mul_op_o : word;
  -- data path registers
  signal reg_14 : word;
  signal reg_15 : word;
  signal reg_16 : word;
  signal reg_17 : word;
  signal reg_18 : word;
  signal reg_19 : word;
  signal reg_20 : word;
  signal reg_21 : word;
  signal reg_22 : word;
  signal reg_25 : word;
  signal reg_31 : word;
  signal reg_37 : word;
  signal reg_43 : word;
  -- hardwire constant
  constant reg_0 : word := std_logic_vector( to_signed(0, 16));
  constant reg_1 : word := std_logic_vector( to_signed(0, 16));
  constant reg_2 : word := std_logic_vector( to_signed(0, 16));
  constant reg_3 : word := std_logic_vector( to_signed(0, 16));
  constant reg_4 : word := std_logic_vector( to_signed(0, 16));
  constant reg_5 : word := std_logic_vector( to_signed(0, 16));
  constant reg_6 : word := std_logic_vector( to_signed(0, 16));
  constant reg_7 : word := std_logic_vector( to_signed(0, 16));
  constant reg_8 : word := std_logic_vector( to_signed(0, 16));
  constant reg_9 : word := std_logic_vector( to_signed(0, 16));
  constant reg_10 : word := std_logic_vector( to_signed(0, 16));
  constant reg_11 : word := std_logic_vector( to_signed(0, 16));
  constant reg_12 : word := std_logic_vector( to_signed(0, 16));
  constant reg_13 : word := std_logic_vector( to_signed(0, 16));

begin
  -- instance of operator add_op to implement function add
  comp_0_add_op : add_op
    port map(
      -- input ports
      a => comp_0_add_op_a,
      b => comp_0_add_op_b,
      -- output ports
      o => comp_0_add_op_o
    );
  -- instance of operator mul_op to implement function mul
  comp_2_mul_op : mul_op
    port map(
      -- input ports
      a => comp_2_mul_op_a,
      b => comp_2_mul_op_b,
      -- output ports
      o => comp_2_mul_op_o
    );
  -- instance of operator mul_op to implement function mul
  comp_3_mul_op : mul_op
    port map(
      -- input ports
      a => comp_3_mul_op_a,
      b => comp_3_mul_op_b,
      -- output ports
      o => comp_3_mul_op_o
    );
  -- instance of operator mul_op to implement function mul
  comp_4_mul_op : mul_op
    port map(
      -- input ports
      a => comp_4_mul_op_a,
      b => comp_4_mul_op_b,
      -- output ports
      o => comp_4_mul_op_o
    );
  -- instance of operator mul_op to implement function mul
  comp_5_mul_op : mul_op
    port map(
      -- input ports
      a => comp_5_mul_op_a,
      b => comp_5_mul_op_b,
      -- output ports
      o => comp_5_mul_op_o
    );

  -- The synchronous process
  SYNC_PROC: process (clk)
  begin
    if (clk'event and clk = '1') then
      if (rstb = '0') then
        state <= S18;
      elsif (enable = '1') then
        state <= next_state;
      end if;
    end if;
  end process;

  MUX_REGISTER: process (clk)
  begin
  if (clk'event and clk = '1') then
    if (rstb = '0') then
    -- reset value on registers
    reg_14 <= (others=>'0');
    reg_15 <= (others=>'0');
    reg_16 <= (others=>'0');
    reg_17 <= (others=>'0');
    reg_18 <= (others=>'0');
    reg_19 <= (others=>'0');
    reg_20 <= (others=>'0');
    reg_21 <= (others=>'0');
    reg_22 <= (others=>'0');
    reg_25 <= (others=>'0');
    reg_31 <= (others=>'0');
    reg_37 <= (others=>'0');
    reg_43 <= (others=>'0');
    elsif (enable = '1') then
    case (state) is
      when S0 => -- time 0
        -- static memory read from buses 
        reg_16 <= BUS_DONNEES_1_v;
        reg_15 <= BUS_DONNEES_2_v;
        reg_14 <= BUS_DONNEES_3_v;
      when S1 => -- time 10
      when S2 => -- time 20
        -- outputs of operation op0
        reg_20 <= comp_2_mul_op_o;
        -- outputs of operation op2
        reg_19 <= comp_3_mul_op_o;
        -- outputs of operation op4
        reg_18 <= comp_4_mul_op_o;
        -- outputs of operation op6
        reg_17 <= comp_5_mul_op_o;
      when S3 => -- time 30
        -- outputs of operation op1
        reg_22 <=reg_20;
        -- outputs of operation op3
        reg_20 <=reg_19;
        -- outputs of operation op32
        reg_21 <= comp_0_add_op_o;
        -- outputs of operation op5
        reg_19 <=reg_18;
      when S4 => -- time 40
        -- outputs of operation op10
        reg_22 <= comp_4_mul_op_o;
        -- outputs of operation op12
        reg_19 <= comp_3_mul_op_o;
        -- outputs of operation op15
        reg_25 <= comp_5_mul_op_o;
        -- outputs of operation op33
        reg_21 <= comp_0_add_op_o;
        -- outputs of operation op8
        reg_20 <= comp_2_mul_op_o;
      when S5 => -- time 50
        -- outputs of operation op34
        reg_21 <= comp_0_add_op_o;
      when S6 => -- time 60
        -- outputs of operation op13
        reg_31 <= comp_2_mul_op_o;
        -- outputs of operation op14
        reg_19 <= comp_3_mul_op_o;
        -- outputs of operation op16
        reg_18 <= comp_4_mul_op_o;
        -- outputs of operation op17
        reg_25 <= comp_5_mul_op_o;
        -- outputs of operation op35
        reg_21 <= comp_0_add_op_o;
      when S7 => -- time 70
        -- outputs of operation op36
        reg_21 <= comp_0_add_op_o;
      when S8 => -- time 80
        -- outputs of operation op18
        reg_37 <= comp_4_mul_op_o;
        -- outputs of operation op22
        reg_31 <= comp_2_mul_op_o;
        -- outputs of operation op24
        reg_19 <= comp_3_mul_op_o;
        -- outputs of operation op26
        reg_25 <= comp_5_mul_op_o;
        -- outputs of operation op37
        reg_21 <= comp_0_add_op_o;
      when S9 => -- time 90
        -- outputs of operation op38
        reg_21 <= comp_0_add_op_o;
      when S10 => -- time 100
        -- outputs of operation op19
        reg_19 <= comp_3_mul_op_o;
        -- outputs of operation op23
        reg_31 <= comp_2_mul_op_o;
        -- outputs of operation op25
        reg_43 <= comp_5_mul_op_o;
        -- outputs of operation op28
        reg_18 <= comp_4_mul_op_o;
        -- outputs of operation op39
        reg_21 <= comp_0_add_op_o;
      when S11 => -- time 110
        -- outputs of operation op40
        reg_21 <= comp_0_add_op_o;
      when S12 => -- time 120
        -- outputs of operation op20
        reg_18 <= comp_4_mul_op_o;
        -- outputs of operation op27
        reg_19 <= comp_3_mul_op_o;
        -- outputs of operation op29
        reg_25 <= comp_5_mul_op_o;
        -- outputs of operation op30
        reg_31 <= comp_2_mul_op_o;
        -- outputs of operation op41
        reg_21 <= comp_0_add_op_o;
      when S13 => -- time 130
        -- outputs of operation op42
        reg_21 <= comp_0_add_op_o;
      when S14 => -- time 140
        -- outputs of operation op11
        reg_18 <= comp_4_mul_op_o;
        -- outputs of operation op21
        reg_19 <= comp_3_mul_op_o;
        -- outputs of operation op31
        reg_31 <= comp_2_mul_op_o;
        -- outputs of operation op43
        reg_21 <= comp_0_add_op_o;
        -- outputs of operation op7
        reg_17 <= comp_5_mul_op_o;
      when S15 => -- time 150
        -- outputs of operation op44
        reg_21 <= comp_0_add_op_o;
      when S16 => -- time 160
        -- outputs of operation op45
        reg_21 <= comp_0_add_op_o;
        -- outputs of operation op9
        reg_20 <= comp_2_mul_op_o;
      when S17 => -- time 170
        -- outputs of operation op46
        reg_21 <= comp_0_add_op_o;
      when S18 => -- time 180
    end case; end if;
    end if;
  end process;

  MUX_OPERATOR_TRI_BUS_NEXT_STATE: process (state,reg_14,reg_15,reg_16,
                      reg_17,reg_18,reg_19,reg_20,reg_21,reg_22,reg_25,reg_31,reg_37,reg_43)
  begin
    -- high impedance on buses to avoid bus pollution
    BUS_DONNEES_1_v <= (others => 'Z');
    BUS_DONNEES_2_v <= (others => 'Z');
    BUS_DONNEES_3_v <= (others => 'Z');
    -- signals to connect registers to operators
    comp_0_add_op_a <= (others => 'X');
    comp_0_add_op_b <= (others => 'X');
    comp_2_mul_op_a <= (others => 'X');
    comp_2_mul_op_b <= (others => 'X');
    comp_3_mul_op_a <= (others => 'X');
    comp_3_mul_op_b <= (others => 'X');
    comp_4_mul_op_a <= (others => 'X');
    comp_4_mul_op_b <= (others => 'X');
    comp_5_mul_op_a <= (others => 'X');
    comp_5_mul_op_b <= (others => 'X');
    case (state) is
      when S0 => -- time 0
        next_state <= S1;
      when S1 => -- time 10
        -- inputs of operation op0
        comp_2_mul_op_a <= reg_0;
        comp_2_mul_op_b <= reg_16;
        -- inputs of operation op2
        comp_3_mul_op_a <= reg_6;
        comp_3_mul_op_b <= reg_15;
        -- inputs of operation op4
        comp_4_mul_op_a <= reg_10;
        comp_4_mul_op_b <= reg_14;
        -- inputs of operation op6
        comp_5_mul_op_a <= reg_16;
        comp_5_mul_op_b <= reg_16;
        next_state <= S2;
      when S2 => -- time 20
        -- inputs of operation op0
        comp_2_mul_op_a <= reg_0;
        comp_2_mul_op_b <= reg_16;
        -- inputs of operation op2
        comp_3_mul_op_a <= reg_6;
        comp_3_mul_op_b <= reg_15;
        -- inputs of operation op4
        comp_4_mul_op_a <= reg_10;
        comp_4_mul_op_b <= reg_14;
        -- inputs of operation op6
        comp_5_mul_op_a <= reg_16;
        comp_5_mul_op_b <= reg_16;
        next_state <= S3;
      when S3 => -- time 30
        -- inputs of operation op10
        comp_4_mul_op_a <= reg_14;
        comp_4_mul_op_b <= reg_14;
        -- inputs of operation op12
        comp_3_mul_op_a <= reg_17;
        comp_3_mul_op_b <= reg_1;
        -- inputs of operation op15
        comp_5_mul_op_a <= reg_3;
        comp_5_mul_op_b <= reg_16;
        -- inputs of operation op32
        comp_0_add_op_a <= reg_19;
        comp_0_add_op_b <= reg_20;
        -- inputs of operation op8
        comp_2_mul_op_a <= reg_15;
        comp_2_mul_op_b <= reg_15;
        next_state <= S4;
      when S4 => -- time 40
        -- outputs on buses
        BUS_DONNEES_1_v <= reg_22;
        BUS_DONNEES_2_v <= reg_20;
        BUS_DONNEES_3_v <= reg_19;
        -- inputs of operation op33
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_18;
        -- inputs of operation op10
        comp_4_mul_op_a <= reg_14;
        comp_4_mul_op_b <= reg_14;
        -- inputs of operation op12
        comp_3_mul_op_a <= reg_17;
        comp_3_mul_op_b <= reg_1;
        -- inputs of operation op15
        comp_5_mul_op_a <= reg_3;
        comp_5_mul_op_b <= reg_16;
        -- inputs of operation op8
        comp_2_mul_op_a <= reg_15;
        comp_2_mul_op_b <= reg_15;
        next_state <= S5;
      when S5 => -- time 50
        -- inputs of operation op13
        comp_2_mul_op_a <= reg_20;
        comp_2_mul_op_b <= reg_7;
        -- inputs of operation op14
        comp_3_mul_op_a <= reg_22;
        comp_3_mul_op_b <= reg_11;
        -- inputs of operation op16
        comp_4_mul_op_a <= reg_25;
        comp_4_mul_op_b <= reg_15;
        -- inputs of operation op17
        comp_5_mul_op_a <= reg_5;
        comp_5_mul_op_b <= reg_16;
        -- inputs of operation op34
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_18;
        next_state <= S6;
      when S6 => -- time 60
        -- inputs of operation op35
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_19;
        -- inputs of operation op13
        comp_2_mul_op_a <= reg_20;
        comp_2_mul_op_b <= reg_7;
        -- inputs of operation op14
        comp_3_mul_op_a <= reg_22;
        comp_3_mul_op_b <= reg_11;
        -- inputs of operation op16
        comp_4_mul_op_a <= reg_25;
        comp_4_mul_op_b <= reg_15;
        -- inputs of operation op17
        comp_5_mul_op_a <= reg_5;
        comp_5_mul_op_b <= reg_16;
        next_state <= S7;
      when S7 => -- time 70
        -- inputs of operation op18
        comp_4_mul_op_a <= reg_25;
        comp_4_mul_op_b <= reg_14;
        -- inputs of operation op22
        comp_2_mul_op_a <= reg_14;
        comp_2_mul_op_b <= reg_2;
        -- inputs of operation op24
        comp_3_mul_op_a <= reg_16;
        comp_3_mul_op_b <= reg_8;
        -- inputs of operation op26
        comp_5_mul_op_a <= reg_14;
        comp_5_mul_op_b <= reg_9;
        -- inputs of operation op36
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_31;
        next_state <= S8;
      when S8 => -- time 80
        -- inputs of operation op37
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_19;
        -- inputs of operation op18
        comp_4_mul_op_a <= reg_25;
        comp_4_mul_op_b <= reg_14;
        -- inputs of operation op22
        comp_2_mul_op_a <= reg_14;
        comp_2_mul_op_b <= reg_2;
        -- inputs of operation op24
        comp_3_mul_op_a <= reg_16;
        comp_3_mul_op_b <= reg_8;
        -- inputs of operation op26
        comp_5_mul_op_a <= reg_14;
        comp_5_mul_op_b <= reg_9;
        next_state <= S9;
      when S9 => -- time 90
        -- inputs of operation op19
        comp_3_mul_op_a <= reg_4;
        comp_3_mul_op_b <= reg_16;
        -- inputs of operation op23
        comp_2_mul_op_a <= reg_31;
        comp_2_mul_op_b <= reg_17;
        -- inputs of operation op25
        comp_5_mul_op_a <= reg_19;
        comp_5_mul_op_b <= reg_20;
        -- inputs of operation op28
        comp_4_mul_op_a <= reg_16;
        comp_4_mul_op_b <= reg_12;
        -- inputs of operation op38
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_18;
        next_state <= S10;
      when S10 => -- time 100
        -- inputs of operation op39
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_18;
        -- inputs of operation op19
        comp_3_mul_op_a <= reg_4;
        comp_3_mul_op_b <= reg_16;
        -- inputs of operation op23
        comp_2_mul_op_a <= reg_31;
        comp_2_mul_op_b <= reg_17;
        -- inputs of operation op25
        comp_5_mul_op_a <= reg_19;
        comp_5_mul_op_b <= reg_20;
        -- inputs of operation op28
        comp_4_mul_op_a <= reg_16;
        comp_4_mul_op_b <= reg_12;
        next_state <= S11;
      when S11 => -- time 110
        -- inputs of operation op20
        comp_4_mul_op_a <= reg_19;
        comp_4_mul_op_b <= reg_15;
        -- inputs of operation op27
        comp_3_mul_op_a <= reg_25;
        comp_3_mul_op_b <= reg_20;
        -- inputs of operation op29
        comp_5_mul_op_a <= reg_18;
        comp_5_mul_op_b <= reg_22;
        -- inputs of operation op30
        comp_2_mul_op_a <= reg_15;
        comp_2_mul_op_b <= reg_13;
        -- inputs of operation op40
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_37;
        next_state <= S12;
      when S12 => -- time 120
        -- inputs of operation op41
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_31;
        -- inputs of operation op20
        comp_4_mul_op_a <= reg_19;
        comp_4_mul_op_b <= reg_15;
        -- inputs of operation op27
        comp_3_mul_op_a <= reg_25;
        comp_3_mul_op_b <= reg_20;
        -- inputs of operation op29
        comp_5_mul_op_a <= reg_18;
        comp_5_mul_op_b <= reg_22;
        -- inputs of operation op30
        comp_2_mul_op_a <= reg_15;
        comp_2_mul_op_b <= reg_13;
        next_state <= S13;
      when S13 => -- time 130
        -- inputs of operation op11
        comp_4_mul_op_a <= reg_22;
        comp_4_mul_op_b <= reg_14;
        -- inputs of operation op21
        comp_3_mul_op_a <= reg_18;
        comp_3_mul_op_b <= reg_14;
        -- inputs of operation op31
        comp_2_mul_op_a <= reg_31;
        comp_2_mul_op_b <= reg_22;
        -- inputs of operation op42
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_43;
        -- inputs of operation op7
        comp_5_mul_op_a <= reg_17;
        comp_5_mul_op_b <= reg_16;
        next_state <= S14;
      when S14 => -- time 140
        -- inputs of operation op43
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_19;
        -- inputs of operation op11
        comp_4_mul_op_a <= reg_22;
        comp_4_mul_op_b <= reg_14;
        -- inputs of operation op21
        comp_3_mul_op_a <= reg_18;
        comp_3_mul_op_b <= reg_14;
        -- inputs of operation op31
        comp_2_mul_op_a <= reg_31;
        comp_2_mul_op_b <= reg_22;
        -- inputs of operation op7
        comp_5_mul_op_a <= reg_17;
        comp_5_mul_op_b <= reg_16;
        next_state <= S15;
      when S15 => -- time 150
        -- outputs on buses
        BUS_DONNEES_1_v <= reg_17;
        BUS_DONNEES_2_v <= reg_18;
        -- inputs of operation op44
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_25;
        -- inputs of operation op9
        comp_2_mul_op_a <= reg_20;
        comp_2_mul_op_b <= reg_15;
        next_state <= S16;
      when S16 => -- time 160
        -- inputs of operation op45
        comp_0_add_op_a <= reg_21;
        comp_0_add_op_b <= reg_31;
        -- inputs of operation op9
        comp_2_mul_op_a <= reg_20;
        comp_2_mul_op_b <= reg_15;
        next_state <= S17;
      when S17 => -- time 170
        -- outputs on buses
        BUS_DONNEES_2_v <= reg_20;
        -- inputs of operation op46
        comp_0_add_op_a <= reg_19;
        comp_0_add_op_b <= reg_21;
        next_state <= S18;
      when S18 => -- time 180
        -- outputs on buses
        BUS_DONNEES_1_v <= reg_21;
        next_state <= S0;
    end case;
  end process;

end v_arch;

