library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use std.textio.all;
use work.notech.all;

entity vr is
  port(
-- I/O to connect to operator blocs
    BUS_DONNEES_1_vr : inout word;
    BUS_DONNEES_2_vr : inout word;
    BUS_DONNEES_3_vr : inout word;
    BUS_DONNEES_4_vr : inout word;
    BUS_DONNEES_5_vr : out word;
    BUS_DONNEES_6_vr : out word;
    BUS_DONNEES_7_vr : inout word;
    BUS_DONNEES_8_vr : out word;
    enable, rstb, clk : in std_logic
 );
end;
architecture vr_arch of vr is
  attribute period:string;
  attribute period of clk : signal is "10 ns";
  -- states of the FSM
  type state_type is (S0,S1,S2,S3,S4,S5,S6,S7,S8,
                      S9,S10,S11,S12,S13,S14,S15,S16,S17,S18);
  --sequential coding for state_type
  --attribute enum_encoding : string;
  --attribute enum_encoding of state_type : type is "sequential";
  --sequentail coding for state_type for synplify
  --attribute syn_encoding : string;
  --attribute syn_encoding of state_type : type is "sequential";
  signal next_state : state_type;
  signal state : state_type;
  -- signals to connect registers to operators
  signal comp_0_add_op_a : word;
  signal comp_0_add_op_b : word;
  signal comp_0_add_op_o : word;
  signal comp_2_mul_op_a : word;
  signal comp_2_mul_op_b : word;
  signal comp_2_mul_op_o : word;
  signal comp_3_mul_op_a : word;
  signal comp_3_mul_op_b : word;
  signal comp_3_mul_op_o : word;
  signal comp_5_mul_op_a : word;
  signal comp_5_mul_op_b : word;
  signal comp_5_mul_op_o : word;
  signal comp_4_sll_op_a : word;
  signal comp_4_sll_op_b : word;
  signal comp_4_sll_op_o : word;
  -- data path registers
  signal reg_2 : word;
  signal reg_3 : word;
  signal reg_4 : word;
  signal reg_5 : word;
  signal reg_6 : word;
  signal reg_7 : word;
  signal reg_8 : word;
  signal reg_9 : word;
  signal reg_10 : word;
  signal reg_11 : word;
  signal reg_12 : word;
  signal reg_13 : word;
  signal reg_14 : word;
  signal reg_24 : word;
  signal reg_32 : word;
  -- hardwire constant
  constant reg_0 : word := std_logic_vector( to_signed(1, 16));
  constant reg_1 : word := std_logic_vector( to_signed(0, 16));

begin
  -- instance of operator add_op to implement function add
  comp_0_add_op : add_op
    port map(
      -- input ports
      a => comp_0_add_op_a,
      b => comp_0_add_op_b,
      -- output ports
      o => comp_0_add_op_o
    );
  -- instance of operator mul_op to implement function mul
  comp_2_mul_op : mul_op
    port map(
      -- input ports
      a => comp_2_mul_op_a,
      b => comp_2_mul_op_b,
      -- output ports
      o => comp_2_mul_op_o
    );
  -- instance of operator mul_op to implement function mul
  comp_3_mul_op : mul_op
    port map(
      -- input ports
      a => comp_3_mul_op_a,
      b => comp_3_mul_op_b,
      -- output ports
      o => comp_3_mul_op_o
    );
  -- instance of operator mul_op to implement function mul
  comp_5_mul_op : mul_op
    port map(
      -- input ports
      a => comp_5_mul_op_a,
      b => comp_5_mul_op_b,
      -- output ports
      o => comp_5_mul_op_o
    );
  -- instance of operator sll_op to implement function sll
  comp_4_sll_op : sll_op
    port map(
      -- input ports
      a => comp_4_sll_op_a,
      b => comp_4_sll_op_b,
      -- output ports
      o => comp_4_sll_op_o
    );

  -- The synchronous process
  SYNC_PROC: process (clk)
  begin
    if (clk'event and clk = '1') then
      if (rstb = '0') then
        state <= S18;
      elsif (enable = '1') then
        state <= next_state;
      end if;
    end if;
  end process;

  MUX_REGISTER: process (clk)
  begin
  if (clk'event and clk = '1') then
    if (rstb = '0') then
    -- reset value on registers
    reg_2 <= (others=>'0');
    reg_3 <= (others=>'0');
    reg_4 <= (others=>'0');
    reg_5 <= (others=>'0');
    reg_6 <= (others=>'0');
    reg_7 <= (others=>'0');
    reg_8 <= (others=>'0');
    reg_9 <= (others=>'0');
    reg_10 <= (others=>'0');
    reg_11 <= (others=>'0');
    reg_12 <= (others=>'0');
    reg_13 <= (others=>'0');
    reg_14 <= (others=>'0');
    reg_24 <= (others=>'0');
    reg_32 <= (others=>'0');
    elsif (enable = '1') then
    case (state) is
      when S0 => -- time 0
        -- static memory read from buses 
        reg_7 <= BUS_DONNEES_1_vr;
        reg_6 <= BUS_DONNEES_2_vr;
      when S1 => -- time 10
        -- outputs of operation op0
        reg_12 <= comp_2_mul_op_o;
        -- outputs of operation op1
        reg_13 <= comp_3_mul_op_o;
        -- outputs of operation op2
        reg_14 <= comp_5_mul_op_o;
        -- outputs of operation op42
        reg_8 <=reg_6;
        -- outputs of operation op43
        reg_9 <=reg_6;
        -- outputs of operation op44
        reg_10 <=reg_6;
        -- inputs from buses
        reg_6 <= BUS_DONNEES_1_vr;
        reg_3 <= BUS_DONNEES_2_vr;
        reg_4 <= BUS_DONNEES_3_vr;
        -- static memory read from buses 
        reg_2 <= BUS_DONNEES_4_vr;
      when S2 => -- time 20
        -- outputs of operation op19
        reg_8 <= comp_4_sll_op_o;
        -- outputs of operation op27
        reg_9 <= comp_0_add_op_o;
        -- outputs of operation op39
        reg_13 <=reg_12;
        -- outputs of operation op40
        reg_12 <=reg_13;
        -- outputs of operation op41
        reg_10 <=reg_14;
      when S3 => -- time 30
        -- outputs of operation op10
        reg_13 <= comp_3_mul_op_o;
        -- outputs of operation op11
        reg_11 <= comp_5_mul_op_o;
        -- outputs of operation op8
        reg_12 <= comp_2_mul_op_o;
        -- inputs from buses
        reg_6 <= BUS_DONNEES_1_vr;
        reg_3 <= BUS_DONNEES_2_vr;
        reg_4 <= BUS_DONNEES_3_vr;
        -- static memory read from buses 
        reg_11 <= BUS_DONNEES_7_vr;
      when S4 => -- time 40
        -- outputs of operation op12
        reg_32 <= comp_4_sll_op_o;
        -- outputs of operation op28
        reg_9 <= comp_0_add_op_o;
      when S5 => -- time 50
        -- outputs of operation op14
        reg_12 <= comp_2_mul_op_o;
        -- outputs of operation op15
        reg_3 <= comp_3_mul_op_o;
        -- outputs of operation op17
        reg_14 <= comp_5_mul_op_o;
        -- outputs of operation op29
        reg_9 <= comp_0_add_op_o;
        -- inputs from buses
        reg_24 <= BUS_DONNEES_1_vr;
        reg_4 <= BUS_DONNEES_2_vr;
        -- static memory read from buses 
        reg_24 <= BUS_DONNEES_3_vr;
      when S6 => -- time 60
        -- outputs of operation op30
        reg_9 <= comp_0_add_op_o;
      when S7 => -- time 70
        -- outputs of operation op20
        reg_6 <= comp_2_mul_op_o;
        -- outputs of operation op21
        reg_11 <= comp_5_mul_op_o;
        -- outputs of operation op3
        reg_13 <= comp_3_mul_op_o;
        -- outputs of operation op31
        reg_9 <= comp_0_add_op_o;
        -- inputs from buses
        reg_8 <= BUS_DONNEES_1_vr;
        reg_4 <= BUS_DONNEES_2_vr;
        -- static memory read from buses 
        reg_10 <= BUS_DONNEES_3_vr;
      when S8 => -- time 80
        -- outputs of operation op16
        reg_12 <=reg_13;
        -- outputs of operation op18
        reg_32 <=reg_13;
        -- outputs of operation op32
        reg_9 <= comp_0_add_op_o;
      when S9 => -- time 90
        -- outputs of operation op23
        reg_24 <= comp_2_mul_op_o;
        -- outputs of operation op25
        reg_10 <= comp_5_mul_op_o;
        -- outputs of operation op33
        reg_9 <= comp_0_add_op_o;
        -- outputs of operation op6
        reg_3 <= comp_3_mul_op_o;
        -- inputs from buses
        reg_4 <= BUS_DONNEES_1_vr;
      when S10 => -- time 100
        -- outputs of operation op22
        reg_3 <=reg_3;
        -- outputs of operation op34
        reg_9 <= comp_0_add_op_o;
        -- outputs of operation op7
        reg_8 <=reg_3;
      when S11 => -- time 110
        -- outputs of operation op24
        reg_13 <= comp_3_mul_op_o;
        -- outputs of operation op26
        reg_14 <= comp_5_mul_op_o;
        -- outputs of operation op35
        reg_9 <= comp_0_add_op_o;
        -- outputs of operation op4
        reg_6 <= comp_2_mul_op_o;
        -- static memory read from buses 
        reg_24 <= BUS_DONNEES_1_vr;
      when S12 => -- time 120
        -- outputs of operation op13
        reg_3 <=reg_6;
        -- outputs of operation op36
        reg_9 <= comp_0_add_op_o;
        -- outputs of operation op5
        reg_2 <=reg_6;
        -- outputs of operation op9
        reg_4 <=reg_6;
      when S13 => -- time 130
        -- outputs of operation op37
        reg_9 <= comp_0_add_op_o;
        -- static memory read from buses 
        reg_10 <= BUS_DONNEES_1_vr;
      when S14 => -- time 140
        -- outputs of operation op38
        reg_9 <= comp_0_add_op_o;
      when S15 => -- time 150
        -- outputs of operation op45
        reg_9 <= comp_0_add_op_o;
      when S16 => -- time 160
      when S17 => -- time 170
      when S18 => -- time 180
        -- inputs from buses
        reg_5 <= BUS_DONNEES_1_vr;
        reg_2 <= BUS_DONNEES_2_vr;
        reg_3 <= BUS_DONNEES_3_vr;
        reg_4 <= BUS_DONNEES_4_vr;
    end case; end if;
    end if;
  end process;

  MUX_OPERATOR_TRI_BUS_NEXT_STATE: process (state,reg_2,reg_3,reg_4,
                      reg_5,reg_6,reg_7,reg_8,reg_9,reg_10,reg_11,reg_12,reg_13,reg_14,reg_24,reg_32)
  begin
    -- high impedance on buses to avoid bus pollution
    BUS_DONNEES_1_vr <= (others => 'Z');
    BUS_DONNEES_2_vr <= (others => 'Z');
    BUS_DONNEES_3_vr <= (others => 'Z');
    BUS_DONNEES_4_vr <= (others => 'Z');
    BUS_DONNEES_5_vr <= (others => 'Z');
    BUS_DONNEES_6_vr <= (others => 'Z');
    BUS_DONNEES_7_vr <= (others => 'Z');
    BUS_DONNEES_8_vr <= (others => 'Z');
    -- signals to connect registers to operators
    comp_0_add_op_a <= (others => 'X');
    comp_0_add_op_b <= (others => 'X');
    comp_2_mul_op_a <= (others => 'X');
    comp_2_mul_op_b <= (others => 'X');
    comp_3_mul_op_a <= (others => 'X');
    comp_3_mul_op_b <= (others => 'X');
    comp_5_mul_op_a <= (others => 'X');
    comp_5_mul_op_b <= (others => 'X');
    comp_4_sll_op_a <= (others => 'X');
    comp_4_sll_op_b <= (others => 'X');
    case (state) is
      when S0 => -- time 0
        -- inputs of operation op0
        comp_2_mul_op_a <= reg_5;
        comp_2_mul_op_b <= reg_2;
        -- inputs of operation op1
        comp_3_mul_op_a <= reg_3;
        comp_3_mul_op_b <= reg_5;
        -- inputs of operation op2
        comp_5_mul_op_a <= reg_4;
        comp_5_mul_op_b <= reg_5;
        next_state <= S1;
      when S1 => -- time 10
        -- inputs of operation op19
        comp_4_sll_op_a <= reg_7;
        comp_4_sll_op_b <= reg_0;
        -- inputs of operation op0
        comp_2_mul_op_a <= reg_5;
        comp_2_mul_op_b <= reg_2;
        -- inputs of operation op1
        comp_3_mul_op_a <= reg_3;
        comp_3_mul_op_b <= reg_5;
        -- inputs of operation op2
        comp_5_mul_op_a <= reg_4;
        comp_5_mul_op_b <= reg_5;
        next_state <= S2;
      when S2 => -- time 20
        -- outputs on buses
        BUS_DONNEES_1_vr <= reg_8;
        BUS_DONNEES_2_vr <= reg_9;
        BUS_DONNEES_3_vr <= reg_10;
        -- inputs of operation op10
        comp_3_mul_op_a <= reg_3;
        comp_3_mul_op_b <= reg_11;
        -- inputs of operation op11
        comp_5_mul_op_a <= reg_4;
        comp_5_mul_op_b <= reg_1;
        -- inputs of operation op27
        comp_0_add_op_a <= reg_13;
        comp_0_add_op_b <= reg_12;
        -- inputs of operation op8
        comp_2_mul_op_a <= reg_6;
        comp_2_mul_op_b <= reg_2;
        -- inputs of operation op19
        comp_4_sll_op_a <= reg_7;
        comp_4_sll_op_b <= reg_0;
        next_state <= S3;
      when S3 => -- time 30
        -- outputs on buses
        BUS_DONNEES_4_vr <= reg_13;
        BUS_DONNEES_5_vr <= reg_12;
        BUS_DONNEES_6_vr <= reg_10;
        -- inputs of operation op12
        comp_4_sll_op_a <= reg_14;
        comp_4_sll_op_b <= reg_0;
        -- inputs of operation op10
        comp_3_mul_op_a <= reg_3;
        comp_3_mul_op_b <= reg_11;
        -- inputs of operation op11
        comp_5_mul_op_a <= reg_4;
        comp_5_mul_op_b <= reg_1;
        -- inputs of operation op8
        comp_2_mul_op_a <= reg_6;
        comp_2_mul_op_b <= reg_2;
        next_state <= S4;
      when S4 => -- time 40
        -- inputs of operation op14
        comp_2_mul_op_a <= reg_6;
        comp_2_mul_op_b <= reg_24;
        -- inputs of operation op15
        comp_3_mul_op_a <= reg_3;
        comp_3_mul_op_b <= reg_1;
        -- inputs of operation op17
        comp_5_mul_op_a <= reg_4;
        comp_5_mul_op_b <= reg_10;
        -- inputs of operation op28
        comp_0_add_op_a <= reg_9;
        comp_0_add_op_b <= reg_12;
        -- inputs of operation op12
        comp_4_sll_op_a <= reg_14;
        comp_4_sll_op_b <= reg_0;
        next_state <= S5;
      when S5 => -- time 50
        -- inputs of operation op29
        comp_0_add_op_a <= reg_9;
        comp_0_add_op_b <= reg_32;
        -- inputs of operation op14
        comp_2_mul_op_a <= reg_6;
        comp_2_mul_op_b <= reg_24;
        -- inputs of operation op15
        comp_3_mul_op_a <= reg_3;
        comp_3_mul_op_b <= reg_1;
        -- inputs of operation op17
        comp_5_mul_op_a <= reg_4;
        comp_5_mul_op_b <= reg_10;
        next_state <= S6;
      when S6 => -- time 60
        -- inputs of operation op20
        comp_2_mul_op_a <= reg_8;
        comp_2_mul_op_b <= reg_24;
        -- inputs of operation op21
        comp_5_mul_op_a <= reg_4;
        comp_5_mul_op_b <= reg_1;
        -- inputs of operation op3
        comp_3_mul_op_a <= reg_5;
        comp_3_mul_op_b <= reg_5;
        -- inputs of operation op30
        comp_0_add_op_a <= reg_9;
        comp_0_add_op_b <= reg_13;
        next_state <= S7;
      when S7 => -- time 70
        -- inputs of operation op31
        comp_0_add_op_a <= reg_9;
        comp_0_add_op_b <= reg_11;
        -- inputs of operation op20
        comp_2_mul_op_a <= reg_8;
        comp_2_mul_op_b <= reg_24;
        -- inputs of operation op21
        comp_5_mul_op_a <= reg_4;
        comp_5_mul_op_b <= reg_1;
        -- inputs of operation op3
        comp_3_mul_op_a <= reg_5;
        comp_3_mul_op_b <= reg_5;
        next_state <= S8;
      when S8 => -- time 80
        -- inputs of operation op23
        comp_2_mul_op_a <= reg_8;
        comp_2_mul_op_b <= reg_24;
        -- inputs of operation op25
        comp_5_mul_op_a <= reg_4;
        comp_5_mul_op_b <= reg_10;
        -- inputs of operation op32
        comp_0_add_op_a <= reg_9;
        comp_0_add_op_b <= reg_12;
        -- inputs of operation op6
        comp_3_mul_op_a <= reg_5;
        comp_3_mul_op_b <= reg_1;
        next_state <= S9;
      when S9 => -- time 90
        -- static write to memory
        BUS_DONNEES_2_vr <= reg_12;
        BUS_DONNEES_3_vr <= reg_32;
        -- inputs of operation op33
        comp_0_add_op_a <= reg_9;
        comp_0_add_op_b <= reg_3;
        -- inputs of operation op23
        comp_2_mul_op_a <= reg_8;
        comp_2_mul_op_b <= reg_24;
        -- inputs of operation op25
        comp_5_mul_op_a <= reg_4;
        comp_5_mul_op_b <= reg_10;
        -- inputs of operation op6
        comp_3_mul_op_a <= reg_5;
        comp_3_mul_op_b <= reg_1;
        next_state <= S10;
      when S10 => -- time 100
        -- inputs of operation op24
        comp_3_mul_op_a <= reg_5;
        comp_3_mul_op_b <= reg_7;
        -- inputs of operation op26
        comp_5_mul_op_a <= reg_4;
        comp_5_mul_op_b <= reg_2;
        -- inputs of operation op34
        comp_0_add_op_a <= reg_9;
        comp_0_add_op_b <= reg_14;
        -- inputs of operation op4
        comp_2_mul_op_a <= reg_13;
        comp_2_mul_op_b <= reg_5;
        next_state <= S11;
      when S11 => -- time 110
        -- static write to memory
        BUS_DONNEES_2_vr <= reg_8;
        BUS_DONNEES_3_vr <= reg_3;
        -- inputs of operation op35
        comp_0_add_op_a <= reg_9;
        comp_0_add_op_b <= reg_6;
        -- inputs of operation op24
        comp_3_mul_op_a <= reg_5;
        comp_3_mul_op_b <= reg_7;
        -- inputs of operation op26
        comp_5_mul_op_a <= reg_4;
        comp_5_mul_op_b <= reg_2;
        -- inputs of operation op4
        comp_2_mul_op_a <= reg_13;
        comp_2_mul_op_b <= reg_5;
        next_state <= S12;
      when S12 => -- time 120
        -- static write to memory
        BUS_DONNEES_1_vr <= reg_13;
        -- inputs of operation op36
        comp_0_add_op_a <= reg_9;
        comp_0_add_op_b <= reg_11;
        next_state <= S13;
      when S13 => -- time 130
        -- static write to memory
        BUS_DONNEES_2_vr <= reg_32;
        BUS_DONNEES_3_vr <= reg_12;
        BUS_DONNEES_4_vr <= reg_2;
        BUS_DONNEES_5_vr <= reg_3;
        BUS_DONNEES_6_vr <= reg_4;
        -- inputs of operation op37
        comp_0_add_op_a <= reg_9;
        comp_0_add_op_b <= reg_24;
        next_state <= S14;
      when S14 => -- time 140
        -- inputs of operation op38
        comp_0_add_op_a <= reg_9;
        comp_0_add_op_b <= reg_10;
        next_state <= S15;
      when S15 => -- time 150
        -- static write to memory
        BUS_DONNEES_1_vr <= reg_3;
        BUS_DONNEES_2_vr <= reg_8;
        -- inputs of operation op45
        comp_0_add_op_a <= reg_9;
        comp_0_add_op_b <= reg_14;
        next_state <= S16;
      when S16 => -- time 160
        -- outputs on buses
        BUS_DONNEES_1_vr <= reg_9;
        next_state <= S17;
      when S17 => -- time 170
        next_state <= S18;
      when S18 => -- time 180
        -- static write to memory
        BUS_DONNEES_5_vr <= reg_13;
        BUS_DONNEES_6_vr <= reg_2;
        BUS_DONNEES_7_vr <= reg_3;
        BUS_DONNEES_8_vr <= reg_4;
        next_state <= S0;
    end case;
  end process;

end vr_arch;

