/*
 * =====================================================================================
 *
 *       Filename:  ConvertDfg2Ted.cc
 *    Description:
 *        Created:  01/11/2010 02:31:41 PM
 *         Author:  Daniel F. Gomez-Prado (dgomezpr), dgomezpr@ecs.umass.edu
 *        Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 283                                          $: Revision of last commit
 *  $Author:: daniel                                         $: Author of last commit
 *  $Date:: 2013-02-11 12:52:04 -0500 (Mon, 11 Feb 2013)     $: Date of last commit
 * =====================================================================================
 */

#include <cassert>
#include <string>
#include <iostream>
using namespace std;

#include "TedMan.h"
#include "TedNode.h"
#include "ATedNode.h"
#include "ETedNode.h"
#include "TedNodeRoot.h"
using namespace ted;

#include "DfgMan.h"
#include "DfgNode.h"
using namespace dfg;

#include "util.h"
using namespace util;

#include "ConvertDfg2Ted.h"
namespace convert {

	ATedNode * ConvertDfg2Ted::translate(DfgNode * pNode) {
		// shut down compiler warning
		pNode = pNode;
		return NULL;
	}

	void ConvertDfg2Ted::translate(void) {
		translateDFS();
	}

	void ConvertDfg2Ted::translateLevels(void) {
		assert(ATedNode::checkContainer(_tMan->getContainer()));
		Mark<TedNode>::newMark();
		Mark<TedNode>::newPMark();
		Mark<TedKids>::newPMark();
		_tMan->getContainer().update_PMark();
		DFGVector dfs_nodes;
		DfgParents parents;
		_dMan->collectParents(parents,dfs_nodes);

		map<DfgNode*,unsigned int> parents_ref;
		for (DfgParents::iterator it = parents.begin(); it != parents.end(); it++) {
			if (it->first) {
				parents_ref[it->first] = it->second.size();
			}
		}
		DFGLevels wave;
		unsigned int latency = _dMan->levelize(wave);
		DFGSet& pi = wave[0];
#ifdef FIND_STRUCTURAL_SIMILARITIES
		map<wedge,ETedNode*> constMap;
		map<string,ETedNode*> varMap;
#endif
		for (DFGSet::iterator it = pi.begin(); it != pi.end(); it++) {
			DfgNode* pNode =*it;
			ETedNode* pANode = NULL;
			if (pNode->isConst()) {
#ifdef FIND_STRUCTURAL_SIMILARITIES
				wedge val = pNode->getConst();
				if (constMap.find(val) ==constMap.end()) {
					pANode = new ETedNode(pNode->getConst());
					constMap[val] = pANode;
				} else {
					pANode = constMap[val];
				}
#else
				pANode = new ETedNode(pNode->getConst());
#endif
			} else if (pNode->isPI() || pNode->isVarConst()) {
				string name = pNode->getName();
				TedVar* pvar = pNode->getTedVarIfExist();
				if(pvar && pvar->isRetimed()) {
					name = pvar->getBase()->getName();
				}
#ifdef FIND_STRUCTURAL_SIMILARITIES
				if (varMap.find(name) ==varMap.end()) {
					pANode = new ETedNode(name);
					varMap[name] = pANode;
				} else {
					pANode = varMap[name];
				}
#else
				pANode = new ETedNode(name);
#endif
			} else {
				throw(string("Operator ")+string(DfgOperator::getOpName(pNode->getOp()))+ string(" is at a PI level") );
			}
			mVisited[pNode] = pANode;
			pANode->update_cone_PMark();
		}

		for (unsigned int index = 1; index <= latency; index++) {
			DFGSet& level = wave[index];
			cout << "Level " << index << " with " << level.size() << " DFG nodes and a TED container with " << ATedNode::transitory_container_size() << endl;
			for (DFGSet::iterator it = level.begin(); it != level.end(); it++) {
				DfgNode* pNode =*it;
				ETedNode* pANode = NULL;
				ETedNode* pELeft = mVisited[pNode->getLeft()];
				ETedNode* pERight = mVisited[pNode->getRight()];
				if (pELeft && pERight) {
					assert(parents.find(pNode->getLeft())!=parents.end());
					assert(parents.find(pNode->getRight())!=parents.end());
					parents_ref[pNode->getLeft()]--;
					parents_ref[pNode->getRight()]--;
					if (pNode->isReg()) {
						assert(pELeft==pERight);
						TedRegister retime = pNode->getRegisterValue();
						pANode = dynamic_cast<ETedNode*>(ETedNode::BuildRecursive(pELeft,retime));
						assert(pANode);
					} else if (pNode->isAssign()) {
						assert(pELeft==pERight);
						pANode = pELeft;
						if (parents[pNode->getLeft()].size()>1) {
							pANode = dynamic_cast<ETedNode*>(pELeft->duplicate());
							assert(pANode);
						}
					} else if (pNode->isLoopback()) {
						assert(pELeft==pERight);
						if (parents[pNode->getLeft()].size()>1) {
							pANode = dynamic_cast<ETedNode*>(pELeft->duplicate());
							assert(pANode);
						}
					} else {
						if (pNode->isAdd()) {
							pANode = dynamic_cast<ETedNode*>(pELeft->add_safe(pERight));
						} else if (pNode->isSub()) {
							pANode = dynamic_cast<ETedNode*>(pELeft->sub_safe(pERight));
						} else if (pNode->isMul()) {
							pANode = dynamic_cast<ETedNode*>(pELeft->mul_safe(pERight));
						} else if (pNode->isLSH()) {
							assert(pERight->isConst());
							wedge val = 2;
							wedge exponent = pERight->getConst();
							for (int i = 1; i < exponent; i++) {
								val *= 2;
							}
							pERight->setConst(val);
							pANode = dynamic_cast<ETedNode*>(pELeft->mul_safe(pERight));
						} else {
							throw(string("Operator ")+string(DfgOperator::getOpName(pNode->getOp()))+ string(" prevents ted2dfg. Use dfg2ntl and then ntl2ted") );
						}
					}
				} else if (pELeft || pERight) {
					if (pNode->isAdd()) {
						pANode = dynamic_cast<ETedNode*>(pELeft ? pELeft->duplicate() : pERight->duplicate());
					} else if (pNode->isSub()) {
						if (pELeft) {
							pANode = dynamic_cast<ETedNode*>(pELeft->duplicate());
						} else {
							ETedNode* inv = new ETedNode(-1);
							pANode = dynamic_cast<ETedNode*>(pERight->mul_safe(inv));
						}
					} else if (pNode->isMul()) {
						pANode = NULL;
					} else if (pNode->isLSH()) {
						if (pELeft) {
							pANode = pELeft;
						} else {
							pANode = NULL;
						}
					} else {
						throw(string("Operator ")+string(DfgOperator::getOpName(pNode->getOp()))+ string(" has been reduced") );
					}
				} else {
					pANode = NULL;
				}
				mVisited[pNode] = pANode;
				//#define PARANOIC_GARBAGE_COLLECTION
#undef PARANOIC_GARBAGE_COLLECTION
#if defined(PARANOIC_GARBAGE_COLLECTION)
				if (pANode) {
					Mark<TedNode>::newMark();
					pANode->update_cone_PMark();
				}
				if (ATedNode::transitory_container_size()>=2*threshold) {
					cout << "Container was " << ATedNode::transitory_container_size();
					std::flush(std::cout);
					TedKids::garbage_collect_unmarked(&Mark<TedKids>::isPMarked);
					ATedNode::garbage_collect_unmarked(&Mark<TedNode>::isPMarked);
					cout << " -> Container is " << ATedNode::transitory_container_size() << endl;
					cout << "Threshold is " << threshold;
					std::flush(std::cout);
					if (threshold < ATedNode::transitory_container_size()) {
						threshold = max(threshold*2,ATedNode::transitory_container_size());
					}
				}
#endif
			}
			Mark<TedNode>::newMark();
			Mark<TedNode>::newPMark();
			Mark<TedKids>::newPMark();
			for (DFGSet::iterator it = pi.begin(); it != pi.end(); it++) {
				DfgNode* pNode =*it;
				ETedNode* pANode = mVisited[pNode];
				pANode->update_cone_PMark();
			}
			for (unsigned int jndex = 1; jndex <= index; jndex++) {
				DFGSet& mark_level = wave[jndex];
				for (DFGSet::iterator it = mark_level.begin(); it != mark_level.end(); it++) {
					DfgNode* pNode =*it;
#ifdef FIND_STRUCTURAL_SIMILARITIES
					if (parents_ref.find(pNode)!=parents_ref.end()) {
						if (parents_ref[pNode]>0) {
							ETedNode* pANode = mVisited[pNode];
							if (pANode) {
								pANode->update_cone_PMark();
							}
						} else {
							// no longer referenced therefore removed
							mVisited[pNode] = NULL;
						}
					}
#else
					if (parents_ref[pNode]>0) {
						ETedNode* pANode = mVisited[pNode];
						if (pANode) {
							pANode->update_cone_PMark();
						}
					}
#endif
				}
			}
			for (DfgMan::PrimaryOutputs::iterator p = _dMan->_mPos.begin(); p != _dMan->_mPos.end(); p++) {
				DfgNode* pNode = p->second;
				if (mVisited.find(pNode) != mVisited.end()) {
					ETedNode* pANode = mVisited[pNode];
					if (pANode) {
						pANode->update_cone_PMark();
					}
				}
			}
			cout << " ---------------------------- " << endl;
			cout << " container was " << ATedNode::transitory_container_size();
			std::flush(std::cout);
			TedKids::garbage_collect_unmarked(&Mark<TedKids>::isPMarked);
			ATedNode::garbage_collect_unmarked(&Mark<TedNode>::isPMarked);
			cout << "  new container is " << ATedNode::transitory_container_size() << endl;
#ifdef FIND_STRUCTURAL_SIMILARITIES
			cout << " -+-+-+-+-+-+-+-+-+-+-+-+-+-+ " << endl;
			//
			// now let's compact those nodes even more
			//	 PI names are fixed
			//   const names can be added in the process
			Mark<TedNode>::newMark();
			Mark<TedNode>::newPMark();
			Mark<TedKids>::newPMark();

			set<ETedNode*> goodgrief;
			map<ETedNode*,ETedNode*> compress;
			for (DFGSet::iterator it = pi.begin(); it != pi.end(); it++) {
				DfgNode* pNode =*it;
				ETedNode* pANode = mVisited[pNode];
				pANode->update_cone_PMark();
			}
			for (unsigned int jndex = 1; jndex <= index; jndex++) {
				DFGSet& mark_level = wave[jndex];
				for (DFGSet::iterator it = mark_level.begin(); it != mark_level.end(); it++) {
					DfgNode* pNode =*it;
					ETedNode* pANode = mVisited[pNode];
					if (pANode) {
						if (pANode->isConst()) {
							if (constMap.find(pANode->getConst()) ==constMap.end()) {
								pANode->update_cone_PMark();
								constMap[pANode->getConst()] = pANode;
							} else {
								compress[pANode] = constMap[pANode->getConst()];
							}
						} else {
							unsigned int index1 = 0;
							ETedNode* pKid1 = NULL;
							ETedNode* candidate = NULL;
							FOREACH_KID_OF_NODE(pANode) {
								pKid1 = _iterkids.Node<ETedNode>();
								index1 = _iterkids.getIndex();
								map<ETedNode*,ETedNode*>::iterator kt = compress.find(pKid1);
								if (kt!=compress.end()) {
									candidate = kt->second;
									assert(candidate!=pKid1);
									pANode->replaceKid(index1,candidate,_iterkids.getWeight(),_iterkids.getRegister());
								}
							}
							bool eq_found = false;
							for ( set<ETedNode*>::iterator jt = goodgrief.begin(); jt != goodgrief.end(); jt++) {
								ETedNode* tmp = (*jt);
								if (tmp->getVar() ==pANode->getVar() && *tmp->getKids() ==*pANode->getKids()) {
									eq_found = true;
									compress[pANode] = tmp;
									break;
								}
							}
							if (!eq_found) {
								pANode->update_cone_PMark();
								goodgrief.insert(pANode);
							}
						}
					}
				}
			}
			for (map <DfgNode *, ETedNode *>::iterator it = mVisited.begin(); it != mVisited.end();it++) {
				ETedNode* oldkey = it->second;
				if (oldkey && compress.find(oldkey)!= compress.end()) {
					ETedNode* newkey = compress[oldkey];
					mVisited[it->first] = newkey;
				}
			}
			for (DfgMan::PrimaryOutputs::iterator p = _dMan->_mPos.begin(); p != _dMan->_mPos.end(); p++) {
				DfgNode* pNode = p->second;
				if (mVisited.find(pNode) != mVisited.end()) {
					ETedNode* pANode = mVisited[pNode];
					if (pANode) {
						pANode->update_cone_PMark();
					}
				}
			}
			goodgrief.clear();
			compress.clear();
			TedKids::garbage_collect_unmarked(&Mark<TedKids>::isPMarked);
			ATedNode::garbage_collect_unmarked(&Mark<TedNode>::isPMarked);
			cout << "  final container is " << ATedNode::transitory_container_size() << endl;
#endif
		}
		for (DfgMan::PrimaryOutputs::iterator p = _dMan->_mPos.begin(); p != _dMan->_mPos.end(); p++) {
			ATedNode * pANode = mVisited[p->second];
			if (pANode) {
				pANode = pANode->duplicateRecursive();
			}
			TedNodeRoot pin(_tMan->normalize(pANode));
			_tMan->linkTedToPO(pin.Node(), pin.getWeight(), p->first);
		}
	}


	// the following flag passed all the tests and should be incorporated in the method, but as it makes
	// the code much difficult to read, is left with #ifdef ... #endif brackets.
#define REDUCE_CONES_NO_LONGER_NEEDED 1
	void ConvertDfg2Ted::translateDFS(void) {
		assert(ATedNode::checkContainer(_tMan->getContainer()));
		Mark<TedNode>::newMark();
		Mark<TedNode>::newPMark();
		Mark<TedKids>::newPMark();
		_tMan->getContainer().update_PMark();
		DFGVector dfs_nodes;
		DfgParents parents;
		_dMan->collectParents(parents,dfs_nodes);
#if REDUCE_CONES_NO_LONGER_NEEDED
		map<DfgNode*,unsigned int> parents_ref;
		for (DfgParents::iterator it = parents.begin(); it != parents.end(); it++) {
			if (it->first) {
				parents_ref[it->first] = it->second.size();
			}
		}
		for (DfgMan::PrimaryOutputs::iterator p = _dMan->_mPos.begin(); p != _dMan->_mPos.end(); p++) {
			DfgNode* dNode = p->second;
			if (parents_ref.find(dNode)!=parents_ref.end()) {
				parents_ref[dNode]++;
			} else {
				parents_ref[dNode] = 1;
			}
		}
		ETedNode* deadbeaf = reinterpret_cast<ETedNode*>(0xdeadbeaf);
#endif
		TEDSet::size_type threshold = 1000;
		DFGVector::size_type count = 0;
		bool memory_fragmented = false;
		for(DFGVector::iterator it = dfs_nodes.begin(); it != dfs_nodes.end(); it++) {
			char msg[64];
			//                   10        20        30        40        50
			//          "12345678901234567890123456789012345678901234567890"
			//           size=999,999,999 [memory fragmented]            x
			sprintf(msg,"size=%d %s         ",ATedNode::transitory_container_size(),(memory_fragmented?"[memory fragmented]":"                   "));
			string buf(msg);
			Util::progressBar(count++,dfs_nodes.size()+_dMan->_mPos.size(),buf);
			DfgNode* pNode =*it;
			ETedNode* pANode = NULL;
			if (pNode->isConst()) {
				wedge val = pNode->getConst();
				pANode = new ETedNode(val);
			} else if (pNode->isPI() || pNode->isVarConst()) {
				string name = pNode->getName();
				TedVar* pvar = pNode->getTedVarIfExist();
				if(pvar && pvar->isRetimed()) {
					name = pvar->getBase()->getName();
				}
				pANode = new ETedNode(name);
			} else {
				ETedNode* pELeft = mVisited[pNode->getLeft()];
				ETedNode* pERight = mVisited[pNode->getRight()];
#ifdef REDUCE_CONES_NO_LONGER_NEEDED
				assert(pELeft!=deadbeaf);
				assert(pERight!=deadbeaf);
#endif
				if (pELeft && pERight) {
					assert(parents.find(pNode->getLeft())!=parents.end());
					assert(parents.find(pNode->getRight())!=parents.end());
#ifdef REDUCE_CONES_NO_LONGER_NEEDED
					parents_ref[pNode->getLeft()]--;
					parents_ref[pNode->getRight()]--;
#endif
					if (pNode->isReg()) {
						assert(pELeft==pERight);
						assert(pNode->getLeft() ==pNode->getRight());
						TedRegister retime = pNode->getRegisterValue();
						pANode = dynamic_cast<ETedNode*>(ETedNode::BuildRecursive(pELeft,retime));
						assert(pANode);
#ifdef REDUCE_CONES_NO_LONGER_NEEDED
						if (parents_ref[pNode->getLeft()]==0) {
							mVisited[pNode->getLeft()] = deadbeaf;
							mVisited[pNode->getRight()] = deadbeaf;
						}
#endif

					} else if (pNode->isAssign()) {
						assert(pELeft==pERight);
						pANode = pELeft;
						if (parents[pNode->getLeft()].size()>1) {
							pANode = dynamic_cast<ETedNode*>(pELeft->duplicate());
							assert(pANode);
						}
					} else if (pNode->isLoopback()) {
						assert(pELeft==pERight);
						if (parents[pNode->getLeft()].size()>1) {
							pANode = dynamic_cast<ETedNode*>(pELeft->duplicate());
							assert(pANode);
						}
					} else {
#ifdef REDUCE_CONES_NO_LONGER_NEEDED
						if (parents_ref[pNode->getLeft()]==0) {
							mVisited[pNode->getLeft()] = deadbeaf;
						}
						if (parents_ref[pNode->getRight()]==0) {
							mVisited[pNode->getRight()] = deadbeaf;
						}
#endif
						if (pNode->isAdd()) {
							pANode = dynamic_cast<ETedNode*>(pELeft->add_safe(pERight));
						} else if (pNode->isSub()) {
							pANode = dynamic_cast<ETedNode*>(pELeft->sub_safe(pERight));
						} else if (pNode->isMul()) {
							pANode = dynamic_cast<ETedNode*>(pELeft->mul_safe(pERight));
						} else if (pNode->isLSH()) {
							assert(pERight->isConst());
							wedge val = 2;
							wedge exponent = pERight->getConst();
							for (int i = 1; i < exponent; i++) {
								val *= 2;
							}
							//pERight->setConst(val);
							pERight = new ETedNode(val);
							pANode = dynamic_cast<ETedNode*>(pELeft->mul_safe(pERight));
						} else {
							throw(string("Operator ")+string(DfgOperator::getOpName(pNode->getOp()))+ string(" prevents ted2dfg. Use dfg2ntl and then ntl2ted") );
						}
					}
				} else if (pELeft || pERight) {
					if (pNode->isAdd()) {
						pANode = dynamic_cast<ETedNode*>(pELeft ? pELeft->duplicate() : pERight->duplicate());
					} else if (pNode->isSub()) {
						if (pELeft) {
							pANode = dynamic_cast<ETedNode*>(pELeft->duplicate());
						} else {
							ETedNode* inv = new ETedNode(-1);
							pANode = dynamic_cast<ETedNode*>(pERight->mul_safe(inv));
						}
					} else if (pNode->isMul()) {
						pANode = NULL;
					} else if (pNode->isLSH()) {
						if (pELeft) {
							pANode = pELeft;
						} else {
							pANode = NULL;
						}
					} else {
						throw(string("Operator ")+string(DfgOperator::getOpName(pNode->getOp()))+ string(" has been reduced") );
					}
				} else {
					pANode = NULL;
				}

			}
			mVisited[pNode] = pANode;
#ifndef REDUCE_CONES_NO_LONGER_NEEDED
			if (pANode) {
				// the property mark is kept the same through out and each cone is updated with the mark
				// all cones maintain the mark and they are preserved through out
				Mark<TedNode>::newMark();
				pANode->update_cone_PMark();
			}
#endif
			if (ATedNode::transitory_container_size()>threshold) {
#if REDUCE_CONES_NO_LONGER_NEEDED
				// at each point that requires the container reduction the property mark is updated
				// therefore all cones that are still active (not null or not deadbeaf) are updated
				// all other cones are removed.
				Mark<TedNode>::newMark();
				Mark<TedNode>::newPMark();
				Mark<TedKids>::newPMark();
				memory_fragmented = (ATedNode::transitory_container_size()>threshold*3);
				for (map <DfgNode *, ETedNode *>::iterator mviter = mVisited.begin(); mviter != mVisited.end(); mviter++) {
					ETedNode* ptr = mviter->second;
					if (ptr && ptr != deadbeaf) {
						if (memory_fragmented) {
							ETedNode* new_ptr = dynamic_cast<ETedNode*>(ptr->duplicateRecursive());
							mviter->second = new_ptr;
							ptr = new_ptr;
						}
						ptr->update_cone_PMark();
					}
				}
#endif
				TedKids::garbage_collect_unmarked(&Mark<TedKids>::isPMarked);
				ATedNode::garbage_collect_unmarked(&Mark<TedNode>::isPMarked);
				threshold = (min(threshold,ATedNode::transitory_container_size())*3)>>1;
			}
		}
		for (DfgMan::PrimaryOutputs::iterator p = _dMan->_mPos.begin(); p != _dMan->_mPos.end(); p++) {
			Util::progressBar(count++,dfs_nodes.size()+_dMan->_mPos.size(),"                       ");
			ATedNode * pANode = mVisited[p->second];
			if (pANode) {
				pANode = pANode->duplicateRecursive();
			}
			TedNodeRoot pin(_tMan->normalize(pANode));
			_tMan->linkTedToPO(pin.Node(), pin.getWeight(), p->first);
		}
		Util::progressBar(1,1," ");
		cout << endl;
	}

	void ConvertDfg2Ted::flatten(void) {
		assert(ATedNode::checkContainer(_tMan->getContainer()));
		multimap<int,pair<string,int> > h;
		_dMan->hierarchy(h);
#if 0
		Mark<TedNode>::newPMark();
		_tMan->getContainer().update_PMark();
#endif
		map<string,ATedNode*> hash_name;
		map<string,int> hash_count;
		DfgParents parents;
		_dMan->collectParents(parents);
		Mark<DfgNode>::newMark();
		mVisited.clear();
		for (multimap<int,pair<string,int> >::iterator jt = h.begin(), jtend=h.end(); jt!=jtend; jt++) {
			string po_name = jt->second.first;
			DfgNode* po_node = _dMan->_mPos[po_name];
			ATedNode* pANode = flatten(hash_count,hash_name,parents,po_node);
			hash_name[po_name] = pANode;
			hash_count[po_name] = jt->second.second;
#if 0
			Mark<TedNode>::newMark();
			pANode->update_cone_PMark();
			TedKids::garbage_collect_unmarked(&Mark<TedKids>::isPMarked);
			ATedNode::garbage_collect_unmarked(&Mark<TedNode>::isPMarked);
			cout << "po_name="<<po_name<<endl;
#endif
		}
		for (multimap<int,pair<string,int> >::iterator jt = h.begin(), jtend=h.end(); jt!=jtend; jt++) {
			if (jt->second.second==0) {
				ATedNode * pANode = hash_name[jt->second.first];
				TedNodeRoot pin(_tMan->normalize(pANode));
				_tMan->linkTedToPO(pin.Node(), pin.getWeight(), jt->second.first);
			}
		}
	}

	ATedNode* ConvertDfg2Ted::flatten(map<string,int>& hash_count, map<string,ATedNode*>& hash_name, DfgParents& parents, DfgNode* pNode) {
		if (mVisited.find(pNode)!=mVisited.end()) {
			return mVisited[pNode];
		}
		ATedNode* pANode = NULL;
		if (pNode->isConst()) {
			pANode = new ETedNode(pNode->getConst());
		} else if (pNode->isPI() || pNode->isVarConst()) {
			string name = pNode->getName();
			map<string,ATedNode*>::iterator it = hash_name.find(name);
			if (it!=hash_name.end()) {
				pANode = hash_name[name];
				//			if (hash_count[name]>1) {
				//				pANode = pANode->duplicateRecursive();
				//				hash_count[name]--;
				//			}
			} else {
				pANode = new ETedNode(pNode->getName());
			}
		} else {
			assert(pNode->isOp());
			ATedNode* pELeft = flatten(hash_count,hash_name,parents,pNode->getLeft());
			ATedNode* pERight = flatten(hash_count,hash_name,parents,pNode->getRight());
			assert(parents.find(pNode->getLeft())!=parents.end());
			if (parents[pNode->getLeft()].size()>1) {
				pELeft = pELeft->duplicateRecursive();
			}
			assert(parents.find(pNode->getRight())!=parents.end());
			if (parents[pNode->getRight()].size()>1) {
				pERight = pERight->duplicateRecursive();
			}
			if (pNode->isAdd()) {
				pANode = pELeft->add(pERight);
			} else if (pNode->isSub()) {
				pANode = pELeft->sub(pERight);
			} else if (pNode->isMul()) {
				pANode = pELeft->mul(pERight);
			} else if (pNode->isLSH()) {
				assert(pERight->isConst());
				wedge val = 2;
				wedge exponent = pERight->getConst();
				for (int i = 1; i < exponent; i++) {
					val *= 2;
				}
				pERight->setConst(val);
				pANode = pELeft->mul(pERight);
			}
		}
		mVisited[pNode] = dynamic_cast<ETedNode*>(pANode);
		return pANode;
	}

}
