/*
 * =====================================================================================
 *
 *       Filename:  MapDfgNode.cc
 *    Description:
 *        Created:  10/02/2009 12:20:42 AM
 *         Author:  Daniel f. Gomez-Prado (dgomezpr), dgomezpr@ecs.umass.edu
 *        Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 255                                          $: Revision of last commit
 *  $Author:: daniel@win                                     $: Author of last commit
 *  $Date:: 2012-08-29 21:48:37 -0400 (Wed, 29 Aug 2012)     $: Date of last commit
 * =====================================================================================
 */

#include <cassert>

#include "TedNode.h"
using namespace ted;

#include "DfgNode.h"
using namespace dfg;

#include "MapDfgNode.h"
namespace convert {

	void MapDfgNode::registrate(OuterKey dNode, InnerKey dw, NodeType* dnode) {
		Database::iterator found = _data.find(dNode);
		if (found == _data.end()) {
			Data datum;
			datum.insert(pair<InnerKey,NodeType*>(dw,dnode));
			_data.insert(pair<OuterKey,Data>(dNode,datum));
		} else {
			throw(string("08002. Registering an existing node"));
		}
	}

	MapDfgNode::NodeType* MapDfgNode::get(OuterKey dNode, InnerKey dw) {
		//PRE: dNode has already been registered
		assert(_data.find(dNode) != _data.end());
		Data& datum = _data[dNode];
		assert(!datum.empty());
#if 1
		Data::iterator found = datum.find(dw);
		NodeType* ret = NULL;
		if (found==datum.end()) {
			ret = dNode->mul(dw);
			datum[dw] = ret;
		} else {
			ret = found->second;
		}
#else
		NodeType * ret = datum[dw];
		if (!ret) {
			ret = dNode->mul(dw);
			datum[dw] = ret;
		}
#endif
		return ret;
	}

	void MapDfgNodeR::registrate(OuterKey dNode, InnerKey dw, NodeType* dnode) {
		Database::iterator found = _data.find(dNode);
		if (found == _data.end() && 1==dw) {
			Data datum;
			datum.insert(pair<InnerKey,NodeType*>(dw,dnode));
			_data.insert(pair<OuterKey,Data>(dNode,datum));
		} else {
			throw(string("08003. Bad initial retimed node registration"));
		}
	}

	MapDfgNodeR::NodeType* MapDfgNodeR::get(OuterKey dNode, InnerKey dw) {
		//PRE: dNode has already been registered
		if (0==dw)
			return dNode;
		if (_data.find(dNode) == _data.end()) {
			registrate(dNode,1,dNode->reg(1));
		}
		Data& datum = _data[dNode];
		assert(!datum.empty());
		Data::iterator found = datum.find(dw);
		NodeType* ret = NULL;
		if (found==datum.end()) {
			assert(!datum.empty());
			ret = datum[datum.size()];
			for (Data::size_type index = datum.size()+1; index <= dw; index++) {
				ret = ret->reg(1);
				datum[index] = ret;
			}
		} else {
			ret = found->second;
		}
		return ret;
	}



	void MapDfgNodeP::registrate(OuterKey varname,InnerKey index, NodeType* dnode) {
		// PRE varname is not yet on the database
		assert(!has(varname));
		Data datum;
		datum.insert(pair<InnerKey,NodeType*>(index,dnode));
		_data.insert(pair<OuterKey,Data>(varname,datum));
	}

	MapDfgNodeP::NodeType* MapDfgNodeP::get(OuterKey varname, InnerKey index) {
		//PRE: varname has been registered on the map
		if (!has(varname)) {
			registrate(varname,1,varname);
		}
		Data& datum = _data[varname];
		NodeType* ret = datum[index];
		if (!ret) {
			ret = generate(datum,index);
		}
		return ret;
	}

	MapDfgNodeP::NodeType* MapDfgNodeP::generate(Data& datum, InnerKey index) {
		assert(index!=1);
		double val = ((double)index)/2;
		double n = ceil(val);
		double m = floor(val);
		int index_n = (int) n;
		int index_m = (int) m;
		NodeType * dnode_n = datum[index_n];
		if( !dnode_n ) {
			dnode_n = generate(datum,index_n);
		}
		NodeType * dnode_m = datum[index_m];
		if( !dnode_m ) {
			dnode_m = generate(datum,index_m);
		}
		NodeType* dnode_index = new NodeType(_pMan,DfgOperator::MUL,dnode_n,dnode_m);
		datum[index] = dnode_index;
		assert(dnode_index);
		return dnode_index;
	}


}
