

#include "DfgChain.h"
#include "DfgMan.h"
#include "DfgNode.h"


namespace dfg {


	void DfgChain::insert_at(unsigned int delay, DfgNode* dnode) {
		insert(DfgChain::value_type(delay,dnode));
	}

	void DfgChain::create_mul_tree(void) {
		assert (this->size() !=1);
		reorder_tree();
		if (this->size()>2) {
			DfgChain::iterator ms1 = this->begin();
			DfgChain::iterator ms2 = ms1;
			ms2++;
			DfgChain::iterator ms3 = ms2;
			ms3++;
			while ( ms3!=this->end() ) {
				unsigned int combine_delay = (ms2->first) + DfgNode::delayMul();
				DfgNode* new_element = ms1->second->mul(ms2->second);
				this->erase(ms1);
				this->erase(ms2);
				this->insert(DfgChain::value_type(combine_delay,new_element));
				ms1 = this->begin();
				ms2 = ms1;
				ms2++;
				ms3 = ms2;
				ms3++;
			}
			assert(this->size() ==2);
		}
	}

	// invert all additions by subtractions and
	// invert all subtractions by additions
	void DfgChain::invert_add_tree(void) {
		DfgChain::iterator ms1 = this->begin();
		long mask = ~0L-1;
		while (ms1 != this->end()) {
			bool ms1_inv = (long) ms1->second & 1;
			bool invert_ms1 = !ms1_inv;
			ms1->second = reinterpret_cast<DfgNode*>( ((long) ms1->second & mask) | invert_ms1 );
			ms1++;
		}
	}

	void DfgChain::recompose_add_tree(DfgNode* minusOne) {
		reorder_tree();
		//
		// recover level one adders from top chain adders (including constant multiplications)
		// ie: (a+b+c)*d --> translates due to ordering into --> a*d+b*d+c*d
		// this method recover the adders as if the order were  d*(a+b+c+d)
		//
		DfgChain recomposed;
		DfgChain recomposing;
		DfgChain::iterator ms1 = this->begin();
		long mask = ~0L-1;
		//
		// runtime can be improved by using a matrix to discover all pair connectivity
		//
		while (ms1!=this->end()) {
			bool ms1_inv = (long) ms1->second & 1;
			DfgNode* node_1 = reinterpret_cast<DfgNode*>( (long)ms1->second & mask);
			unsigned int ms1_delay = ms1->first;
			this->erase(ms1);
			assert( !( (long)node_1->getLeft() & 1) && !( (long)node_1->getRight() & 1));
			if (node_1->isInput() || !node_1->isMul() ) {
				recomposed.insert(DfgChain::value_type(ms1_delay, reinterpret_cast<DfgNode*>((long)node_1 | ms1_inv)));
			} else {
				assert(node_1->isMul());
				DfgNode* key = NULL;
				DfgChain::iterator ms2 = this->begin();
				while(ms2 != this->end()) {
					bool ms2_inv = (long) ms2->second & 1;
					DfgNode* node_2 = reinterpret_cast<DfgNode*>( (long)ms2->second & mask);
					assert( !((long)node_2->getLeft() & 1) && !((long)node_2->getRight() & 1));
					bool to_recomposing = false;
					if (!node_2->isInput()) {
						if (key) {
							if (key == node_2->getLeft()) {
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)node_2->getRight() | ms2_inv)));
								to_recomposing = true;
							} else if (key == node_2->getRight()) {
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)node_2->getLeft() | ms2_inv)));
								to_recomposing = true;
							} else if (node_2->getLeft()->isMul() && key == node_2->getLeft()->getLeft()) {
								DfgNode* retouch = node_2->getRight()->mul(node_2->getLeft()->getRight());
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)retouch | ms2_inv)));
								to_recomposing = true;
							} else if (node_2->getLeft()->isMul() && key == node_2->getLeft()->getRight()) {
								DfgNode* retouch = node_2->getRight()->mul(node_2->getLeft()->getLeft());
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)retouch | ms2_inv)));
								to_recomposing = true;
							} else if (node_2->getRight()->isMul() && key == node_2->getRight()->getLeft()) {
								DfgNode* retouch = node_2->getLeft()->mul(node_2->getRight()->getRight());
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)retouch | ms2_inv)));
								to_recomposing = true;
							} else if (node_2->getRight()->isMul() && key == node_2->getRight()->getRight()) {
								DfgNode* retouch = node_2->getLeft()->mul(node_2->getRight()->getLeft());
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)retouch | ms2_inv)));
								to_recomposing = true;
							}
						} else {
							if (node_1->getLeft() == node_2->getLeft()) {
								key = node_1->getLeft();
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)node_1->getRight() | ms1_inv)));
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)node_2->getRight() | ms2_inv)));
								to_recomposing = true;
							} else if (node_1->getLeft() == node_2->getRight()) {
								key = node_1->getLeft();
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)node_1->getRight() | ms1_inv)));
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)node_2->getLeft() | ms2_inv)));
								to_recomposing = true;
							} else if (node_1->getRight() == node_2->getLeft()) {
								key = node_1->getRight();
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)node_1->getLeft() | ms1_inv)));
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)node_2->getRight() | ms2_inv)));
								to_recomposing = true;
							} else if (node_1->getRight() == node_2->getRight() ) {
								key = node_1->getRight();
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)node_1->getLeft() | ms1_inv)));
								recomposing.insert(DfgChain::value_type(0,reinterpret_cast<DfgNode*>((long)node_2->getLeft() | ms2_inv)));
								to_recomposing = true;
							}
						}
					}
					if (to_recomposing) {
						this->erase(ms2);
						ms2 = this->begin();
					} else {
						ms2++;
					}
				}
				if (recomposing.empty()) {
					assert(!key);
					recomposed.insert(DfgChain::value_type(ms1_delay,node_1));
				} else {
					assert(key);
					recomposing.create_add_tree();
					DfgChain::iterator ms3 = recomposing.begin();
					long mask = ~0L-1;
					bool invert = ((long)ms3->second & 1);
					DfgNode* ms3_node = reinterpret_cast<DfgNode*>( (long)ms3->second & mask);
					unsigned int new_tree_delay = DfgNode::delayMul();
					DfgNode* new_node = NULL;
					assert(ms1_delay >= DfgNode::delayMul());
					ms1_delay -= DfgNode::delayMul();
					if (invert) {
						DfgNode* to_x = NULL;
						if (ms3->first > ms1_delay) {
							new_tree_delay += ms1_delay;
							new_node = key->mul(minusOne);
							to_x = ms3_node;
						} else {
							new_tree_delay += ms3->first;
							new_node = ms3_node->mul(minusOne);
							to_x = key;
						}
						new_node = new_node->mul(to_x);
						new_tree_delay = max(new_tree_delay,max(ms3->first,ms1_delay))+DfgNode::delayMul();
					} else {
						new_tree_delay = max(ms3->first,ms1_delay)+DfgNode::delayMul();
						new_node = key->mul(ms3_node);
					}
					recomposed.insert( DfgChain::value_type(new_tree_delay,new_node) );
					recomposing.clear();
				}
			}
			ms1 = this->begin();
		}
		assert(this->empty());
		this->insert(recomposed.begin(),recomposed.end());
	}

	void DfgChain::reorder_tree(void) {
		multimap<DfgNode*,unsigned int, _DfgNode_MapOrder> order_map;

		for (DfgChain::iterator it = this->begin(), it_end = this->end(); it != it_end; it++ ) {
			order_map.insert(pair<DfgNode*,unsigned int>(it->second,it->first));
		}
		this->clear();
		for (multimap<DfgNode*,unsigned int, _DfgNode_MapOrder>::iterator it = order_map.begin(), it_end = order_map.end(); it != it_end; it++) {
			this->insert(DfgChain::value_type(it->second,it->first));
		}
		order_map.clear();
	}

	void DfgChain::create_add_tree(void) {

		// limit number of multipliers in parallel
		int max_number_mpy = 2;
		int mul_limit = max_number_mpy;

		if (this->size() ==1) {
			// the adder tree has already been resolved by the recompose_add_tree method
			return;
		}
		reorder_tree();
		DfgChain::iterator ms1 = this->begin();
		DfgChain::iterator ms2 = ms1;
		ms2++;
		long mask = ~0L-1;
		while ( ms2!=this->end()) {
			unsigned int combine_delay = (ms2->first) + DfgNode::delayAdd();
			bool ms1_inv = (long)ms1->second & 1;
			bool ms2_inv = (long)ms2->second & 1;
			DfgNode* new_element = NULL;
			if (ms1_inv && ms2_inv) {
				DfgNode* inv1 = reinterpret_cast<DfgNode*>( (long)ms1->second & mask);
				DfgNode* inv2 = reinterpret_cast<DfgNode*>( (long)ms2->second & mask);
				if (inv1->isMul())
					mul_limit--;
				if (inv2->isMul())
					mul_limit--;
				new_element = inv1->add(inv2);
				new_element = reinterpret_cast<DfgNode*>( (long)new_element | 1);
			} else if (ms1_inv) {
				DfgNode* inv1 = reinterpret_cast<DfgNode*>( (long)ms1->second & mask);
				if (inv1->isMul())
					mul_limit--;
				if (ms2->second->isMul())
					mul_limit--;
				new_element = ms2->second->sub(inv1);
			} else if (ms2_inv) {
				DfgNode* inv2 = reinterpret_cast<DfgNode*>( (long)ms2->second & mask);
				if (inv2->isMul())
					mul_limit--;
				if (ms1->second->isMul())
					mul_limit--;
				new_element = ms1->second->sub(inv2);
			} else {
				if (ms2->second->isMul())
					mul_limit--;
				if (ms1->second->isMul())
					mul_limit--;
				new_element = ms1->second->add(ms2->second);
			}
			this->erase(ms1);
			this->erase(ms2);
#if 0
			//* ***************************************
			//USED WITH PRODMAT ***************************
			//*********************************************
			if (mul_limit==0) {
				mul_limit = max_number_mpy;
				combine_delay = 0;
			} else if (mul_limit<0) {
				mul_limit = max_number_mpy;
				combine_delay += 1;
			}
#endif
			this->insert(DfgChain::value_type(combine_delay,new_element));
			ms1 = this->begin();
			ms2 = ms1;
			ms2++;
		}
		//assert( !((long)this->begin()->second & 1) );
		assert(this->size() ==1);
	}

}
