/*
 * =====================================================================================
 *
 *       Filename:  TedDecompose.c
 *    Description:  TedDecompose class
 *        Created:  04/30/2007 08:59:36 PM EDT
 *         Author:  Daniel Gomez-Prado(from jan 08), Qian Ren(up to dec 07)qren@ren-chao.com
 *        Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 279                                          $: Revision of last commit
 *  $Author:: daniel                                         $: Author of last commit
 *  $Date:: 2010-11-18 08:13:53 -0500(Thu, 18 Nov 2010)$: Date of last commit
 * =====================================================================================
 */

#include <cassert>
#include <cstring>
#include <cstdlib>
#include <sstream>
#include <fstream>
#include <iostream>
#include <climits>
#include <cmath>

#ifndef UINT_MAX
# define UINT_MAX 0xFFFF
#endif

#include "util.h"
using util::Util;

#include "Environment.h"
using tds::Environment;


#include "DfgNode.h"
#include "DfgMan.h"
#include "DfgOperator.h"

#include "Bitwidth.h"
using data::Bitwidth;

#include "TedVarMan.h"
using ted::TedVarMan;

namespace dfg {
#ifndef INLINE
#include "DfgNode.inl"
#endif


	unsigned int DfgNode::_addDelay=0, DfgNode::_subDelay=0, DfgNode::_mpyDelay=0;
	unsigned int DfgNode::_regDelay=0, DfgNode::_lshDelay=0, DfgNode::_rshDelay=0;
	unsigned int DfgNode::_divDelay=3;

	DfgNode::DfgNode(DfgMan* pMan, DfgOperator::Type op, DfgNode* left, DfgNode* right):
		_name(NULL), _left(left), _right(right), _pMan(pMan), _op(NULL), _level(0), _value(0) {
			_op = new DfgOperator(op);
			_data.ptr = NULL;
			_pMan->registerNode(this);
			_nrefs = 0;
			if(_left && _right) {
				setArrivalTime((_left->getArrivalTime()> _right->getArrivalTime() ? _left->getArrivalTime(): _right->getArrivalTime())+ getOpDelay());
			} else {
				_arrival_time= 0;
			}
			_required_time = UINT_MAX;
		}

	DfgNode::DfgNode(DfgMan* pMan, const string& varname, DfgOperator::Type op, wedge data_value) {
		assert(op==DfgOperator::CONST || op==DfgOperator::VARCONST);
		_op = new DfgOperator(op);
		_name = Util::strsav(varname.c_str());
		_left  = NULL;
		_right = NULL;
		_pMan  = pMan;
		_pMan->registerNode(this);
		_data.value = data_value;

		_level = 0;
		_value = 0;
		_nrefs = 0;
		_arrival_time = 0;
		_required_time = UINT_MAX;
	}

	DfgNode::DfgNode(DfgMan* pMan, const string& varname, DfgOperator::Type op) {
		_op = new DfgOperator(op);
		_name = Util::strsav(varname.c_str());
		_left  = NULL;
		_right = NULL;
		_pMan  = pMan;
		_pMan->registerNode(this);

		_level = 0;
		_value = 0;
		_nrefs = 0;
		_arrival_time = 0;
		_required_time = UINT_MAX;
	}

	DfgNode::DfgNode(DfgMan* pMan, DfgNode* prev, unsigned int regamount):
		_name(NULL), _left(prev), _right(prev), _pMan(pMan), _op(NULL), _level(0), _value(0) {
			_op = new DfgOperator(DfgOperator::REG);
			setArrivalTime(prev->getArrivalTime()+ getOpDelay());
			string name = Util::itoa(regamount)+ DfgNode::TOKEN_REG;
			_name = Util::strsav(name.c_str());
			_required_time = UINT_MAX;
			_nrefs = 0;
			_pMan->registerNode(this);
		}

	DfgNode::DfgNode(DfgMan* pMan, wedge constvalue) {
		_name = NULL;
		_op = new DfgOperator(DfgOperator::CONST);
		_data.value = constvalue;
		_left  = NULL;
		_right = NULL;
		_pMan  = pMan;
		_pMan->registerNode(this);

		_level = 0;
		_value = 0;
		_nrefs = 0;
		_arrival_time = 0;
		_required_time= UINT_MAX;
	}

	DfgNode::~DfgNode(void) {
		_pMan->unregisterNode(this);
		if(isPI()) {
			free(_name);
		}
	}

	void DfgNode::setKids(DfgNode* left, DfgNode* right) {
		_left = left;
		_right = right;
		if(_left && _right)
			setArrivalTime((_left->getArrivalTime()> _right->getArrivalTime() ? _left->getArrivalTime(): _right->getArrivalTime())+ getOpDelay());
	}


	const char* DfgNode::getOpStr(void)const {
		switch(getOp()) {
		case DfgOperator::ADD: return "+";
		case DfgOperator::MUL: return "*";
		case DfgOperator::DIV: return "/";
		case DfgOperator::SUB: return "-";
		case DfgOperator::LSH: return "<<";
		case DfgOperator::RSH: return ">>";
		case DfgOperator::REG: return _name;
		case DfgOperator::EQ: return "assign";
		case DfgOperator::LOOPBACK: return "loopback";
		case DfgOperator::VAR: //fall through
		case DfgOperator::VARCONST: //fall through
		case DfgOperator::CONST: return _name;
		default: return "unknown";
		}
	}

	static DfgNode* _RemoveNegPhase_rec(DfgMan* pMan, DfgNode* bNode, bool & bNegs) {
		if(bNode->isConst()&& bNode->getConst() == -1) {
			bNegs ^= true;
			return NULL;
		} else if(!bNode->isMul()) {
			return bNode;
		}

		DfgNode* left  = _RemoveNegPhase_rec(pMan, bNode->getLeft(),  bNegs);
		DfgNode* right = _RemoveNegPhase_rec(pMan, bNode->getRight(),  bNegs);

		if(left != NULL && right != NULL) {
			bNode->setKids(left, right);
			return bNode;
		} else if(left == NULL && right != NULL) {
			return right;
		} else if(right == NULL && left != NULL) {
			return left;
		} else {
			return new DfgNode(pMan, 1);
		}
	}

	DfgNode* DfgNode::add(DfgNode* t) {
		if(isConst()&& t->isConst()) {
			return new DfgNode(_pMan, getConst()+ t->getConst());
		}

		if(isConst()) {
			if(getConst() == 0) {
				return t;
			} else if(getConst()< 0) {
				return t->sub(new DfgNode(_pMan, getConst()*-1));
			}
		}

		if(t->isConst()) {
			if(t->getConst() == 0) {
				return this;
			} else if(t->getConst()< 0) {
				return this->sub(new DfgNode(_pMan, t->getConst()* -1));
			}
		}

		DfgNode* tt = NULL;
		bool bLeftNeg = false, bRightNeg = false;
		DfgNode *left = NULL;
		DfgNode* right = NULL;
		if(this->isMul()) {
			left = _RemoveNegPhase_rec(_pMan, this, bLeftNeg);
		}  else {
			left = this;
		}

		if(t->isMul()) {
			right = _RemoveNegPhase_rec(_pMan, t, bRightNeg);
		}  else {
			right = t;
		}

		if(bLeftNeg == bRightNeg) {
			tt = new DfgNode(_pMan, DfgOperator::ADD, left, right);
			if(bLeftNeg) {
				tt = tt->mul(new DfgNode(_pMan, -1));
			}
		} else {
			if(bLeftNeg) {
				tt = right->sub(left);
			} else {
				tt = left->sub(right);
			}
		}
		return tt;
	}

	//  a - b  is represented as:
	// Node (-)
	//  Left  = a
	//  Right = b
	DfgNode* DfgNode::sub(DfgNode* right) {

		if(isConst()&& right->isConst()) {
			return new DfgNode(_pMan, getConst()- right->getConst());
		}
		if(right->isConst()&& right->getConst() == 0) {
			return this;
		}

		// try to recover predecessor with a multiplication -1
		DfgNode* pTemp = new DfgNode(_pMan, DfgOperator::SUB, this, right);
		return pTemp;
	}

	DfgNode* DfgNode::reg(unsigned int value) {
		return(value==0) ? this : new DfgNode(_pMan,this,value);
	}

	DfgNode* _NegateAddOrSubNode(DfgNode* t) {
		DfgNode* l,* r;

		if(!t->isAddorSub()) {
			return NULL;
		}

		if(t->isSub()) {
			t->setKids(t->getRight(), t->getLeft());
			return t;
		} else if(t->isAdd()) {
			l = _NegateAddOrSubNode(t->getLeft());
			r = _NegateAddOrSubNode(t->getRight());

			if(l != NULL && r != NULL) {
				t->setKids(l, r);
			} else {
				return NULL;
			}
		}
		return NULL;
	}


	DfgNode* DfgNode::mul(DfgNode* t) {
		DfgNode* pLeft,* pRight,* pLeftNeg;

		if(isConst()&& t->isConst()) {
			return new DfgNode(_pMan, getConst()* t->getConst());
		}

		if(isConst()) {
			pLeft = t;
			pRight = this;
		} else {
			pLeft = this;
			pRight = t;
		}

		if(pRight->isConst()&& pRight->getConst() == 1) {
			return pLeft;
		}
#if 0
		// the only thing the ->neg() method prevents is
		// the existence when possible of a -1 multiplier
		//
		// When possible should be over stressed here
		// as a chain of positive multiplications won't
		// be able to reduce the -1 multiplier NOR a chain
		// of positive adders. The presence of only one
		// SUB is neither a guarantee of optimization because
		// the negation has to be performed in a cut of the
		// entire cone, and not only in one operand.
		//
		if(pRight->isConst()&& pRight->getConst()< 0) {
			pLeftNeg = pLeft->neg();
			if(pLeftNeg != 0) {
				return pLeftNeg->mul(new DfgNode(_pMan, pRight->getConst()*-1));
			}
		}
#else
		// let's try to minimize multipliers by -1 preceded by a subtraction
		if (pRight->isConst() && pRight->getConst() ==-1) {
			if (pLeft->isSub()) {
				//swap the Left and Right of the sub operator
				return new DfgNode(_pMan, DfgOperator::SUB, pLeft->getRight(),pLeft->getLeft());
			}
		}
#endif
		return  new DfgNode(_pMan, DfgOperator::MUL, pLeft, pRight);
	}

	DfgNode* DfgNode::hasInvertion(void) const {
		if (this->isMul()) {
			if (_left->isConst() && _left->getConst() ==-1) {
				return _right;
			}
			if (_right->isConst() && _right->getConst() ==-1) {
				return _left;
			}
		}
		return NULL;
	}

	/**
	 * @brief Recursive traversal of DFG nodes, storing all nodes into
	 *        vNodes, and annotating in each node its level
	 *
	 * @pre   a new mark should have been started
	 * @see   Mark::startNewMark
	 * @param vNodes Container for all DFG Nodes.
	 * @param vNodes Maps each node to a particular level.
	 *
	 * @return The maximum number of levels in the DFG graph, or subgraph.
	 **/
	unsigned int DfgNode::collectDFS(vector <DfgNode*> & vNodes) {
		if(_visited.isMarked())
			return _level;
		_visited.setMark();
		if(isInput()) {
			// Variables and Constant values do not increment the DFG level
			_level = 0;
		} else {
			int left_level = getLeft()->collectDFS(vNodes);
			int right_level = getRight()->collectDFS(vNodes);
			// From both left and right subgraphs, choose the max level
			_level = MAX(left_level, right_level)+1;
		}
		// pair "this" node with current "level"
		vNodes.push_back(this);
		return _level;
	}

	void DfgNode::collectDFSonly(vector <DfgNode*> & vNodes) {
		if(_visited.isMarked())
			return;
		_visited.setMark();
		if(!isInput()) {
			getLeft()->collectDFSonly(vNodes);
			getRight()->collectDFSonly(vNodes);
			// From both left and right subgraphs, choose the max level
		}
		// pair "this" node with current "level"
		vNodes.push_back(this);
	}

	void DfgNode::computeSNR(void) {
		if(_visited.isMarked())
			return;
		DfgOperator* left = NULL;
		DfgOperator* right = NULL;
		if(isOp()) {
			left = getLeft()->_op;
			right = getRight()->_op;
			if(!left || !right) {
				throw(string("01031. Error."));
			}
			getLeft()->computeSNR();
			getRight()->computeSNR();
		}
		if(!_op) {
			throw(string("01032. Error."));
		} else {
			if(_op->getError()) {
				_op->clearError();
			}
		}
		switch(getOp()) {
		case DfgOperator::SUB:
		case DfgOperator::ADD:
			{
				assert(left->getError());
				assert(right->getError());
				Error* err = left->getError()->add(right->getError());
				err->insert(_op->getBitwidth()->getSize(),_op->getBitwidth()->getError());
				_op->set(err);
				break;
			}
		case DfgOperator::MUL:
			{
				assert(left->getError());
				assert(right->getError());
				Error* err = left->getError()->mpy(right->getError());
				err->insert(_op->getBitwidth()->getSize(),_op->getBitwidth()->getError());
				_op->set(err);
				break;
			}
		case DfgOperator::LSH:
		case DfgOperator::RSH:
			{
				throw(string("01033. DfgOperator::LSH, DfgOperator::RSH not implemented."));
				break;
			}
		case DfgOperator::CONST:
			{
				Error* err = new Error((long double)_data.value,(long double)_data.value,0.0);
				_op->set(err);
				break;
			}
		case DfgOperator::VARCONST:
		case DfgOperator::VAR:
			{
				Error* err = TedVarMan::instance().getIfExist(getName())->getError();
				if(err) {
					err = new Error(err);
					_op->set(err);
				} else {
					err = new Error(1,1,0);
					_op->set(err);
					//throw(string("01018. No bitwidth defined for variable \"")+getName()+string("\"."));
				}
				break;
			}
		default:
			throw(string("01034. Unknown DfgNode type."));
			break;
		}
		_visited.setMark();
	}

	/** @brief compute the bitwidth of the node by traversing its cone up to the operators at level(default PIs=0)*/
	void DfgNode::computeBitwidth(bool(DfgNode::*Functor)(unsigned int), unsigned int level) {
		if(Functor &&(this->*Functor)(level))
			return;
		if(_visited.isMarked())
			return;
		if(isOp()) {
			getLeft()->computeBitwidth(Functor,level);
			getRight()->computeBitwidth(Functor,level);
		}
		computeBitwidth();
		_visited.setMark();
	}

	void DfgNode::computeBitwidth(void) {
		DfgOperator* left = NULL;
		DfgOperator* right = NULL;
		if(isOp()) {
			left = getLeft()->getOperator();
			right = getRight()->getOperator();
			if(!left || !right) {
				throw(string("01020. No bitwidth set yet."));
			}
		}
		if(!_op) {
			_op = new DfgOperator(getOp());
		} else {
			if(_op->getBitwidth()) {
				_op->clearBitwidth();
			}
		}
		switch(getOp()) {
		case DfgOperator::SUB:
		case DfgOperator::ADD:
			{
				Bitwidth* pre = left->getBitwidth()->clone();
				pre->add(right->getBitwidth());
				_op->set(pre);
				break;
			}
		case DfgOperator::MUL:
			{
				Bitwidth* pre = left->getBitwidth()->clone();
				pre->mpy(right->getBitwidth());
				_op->set(pre);
				break;
			}
		case DfgOperator::LSH:
			{
				unsigned int amount = getRight()->getValue();
				assert(amount>0);
				Bitwidth* pre = left->getBitwidth()->clone();
				pre->lsh(amount);
				_op->set(pre);
				break;
			}
		case DfgOperator::RSH:
			{
				unsigned int amount = getRight()->getValue();
				assert(amount>0);
				Bitwidth* pre = left->getBitwidth()->clone();
				pre->rsh(amount);
				_op->set(pre);
				break;
			}
		case DfgOperator::CONST:
			{
				// what about floating constants
				long double weight = _data.value;
				if(weight <1)
					weight*= -1;
				double bits_needed = log((double)weight)/log((double)2);
				double bits_upbound = ceil(bits_needed);
				unsigned int bits = (unsigned int)bits_upbound;
				if(bits_upbound == bits_needed) {
					//i.e. 16 decimal do not require 4 bits but 5.
					bits++;
				}
				Bitwidth* pre = new Integer(bits);
				_op->set(pre);
				break;
			}
		case DfgOperator::VARCONST:
		case DfgOperator::VAR:
			{
				//what about floating variable ranges
				Bitwidth* pre = TedVarMan::instance().getIfExist(getName())->getBitwidth();
				if(pre) {
					pre = pre->clone();
					_op->set(pre);
				} else {
					//delete _op;
					//_op = NULL;
					throw(string("01018. No bitwidth defined for variable \"")+getName()+string("\"."));
				}
				break;
			}
		case DfgOperator::REG:
			{
				// easily attainable just skip the register and show whatever is the input of the register
				Bitwidth* pre = left->getBitwidth()->clone();
				_op->set(pre);
				//throw(string("01036. Bitwidth computation does not handle registers yet."));
				break;
			}
		default:
			throw(string("01017. Not supported DfgNode type for bitwidth propagation."));
			break;
		}
	}


	/** @brief compute the bitwidth of the node by traversing its cone up to the operators at level(default PIs=0)*/
	void DfgNode::recomputeBitwidth(bool(DfgNode::*Functor)(unsigned int), unsigned int level) {
		if(Functor &&(this->*Functor)(level))
			return;
		if(_visited.isMarked())
			return;
		if(isOp()&& !isReg()) {
			getLeft()->recomputeBitwidth(Functor,level);
			getRight()->recomputeBitwidth(Functor,level);
		}
		recomputeBitwidth();
		_visited.setMark();
	}



	void DfgNode::recomputeBitwidth(void) {
		DfgOperator* left = NULL;
		DfgOperator* right = NULL;
		if(isOp()&& !isReg()) {
			left = getLeft()->getOperator();
			right = getRight()->getOperator();
			if(!left || !right) {
				throw(string("01020. No bitwidth set yet."));
			}
		}
		Bitwidth* oldpre = NULL;
		Bitwidth* pre = NULL;
		if(!_op) {
			_op = new DfgOperator(getOp());
		} else {
			oldpre = _op->getBitwidth();
		}
		switch(getOp()) {
		case DfgOperator::SUB:
		case DfgOperator::ADD:
			{
				pre = left->getBitwidth()->clone();
				pre->add(right->getBitwidth());
				break;
			}
		case DfgOperator::MUL:
			{
				pre = left->getBitwidth()->clone();
				pre->mpy(right->getBitwidth());
				break;
			}
		case DfgOperator::LSH:
			{
				unsigned int amount = getRight()->getValue();
				assert(amount>0);
				pre = left->getBitwidth()->clone();
				pre->lsh(amount);
				break;
			}
		case DfgOperator::RSH:
			{
				unsigned int amount = getRight()->getValue();
				assert(amount>0);
				pre = left->getBitwidth()->clone();
				pre->rsh(amount);
				break;
			}
		case DfgOperator::CONST:
			{
				// what about floating constants
				long double weight = _data.value;
				if(weight <1)
					weight*= -1;
				double bits_needed = log((double)weight)/log((double)2);
				double bits_upbound = ceil(bits_needed);
				unsigned int bits = (unsigned int)bits_upbound;
				if(bits_upbound == bits_needed) {
					//i.e. 16 decimal do not require 4 bits but 5.
					bits++;
				}
				pre = new Integer(bits);
				break;
			}
		case DfgOperator::VARCONST:
		case DfgOperator::VAR:
			{
				//what about floating variable ranges
				Bitwidth* pre = TedVarMan::instance().getIfExist(getName())->getBitwidth();
				if(pre) {
					pre = pre->clone();
				} else {
					//delete _op;
					//_op = NULL;
					throw(string("01018. No bitwidth defined for variable \"")+getName()+string("\"."));
				}
				break;
			}
		case DfgOperator::REG:
			// easily attainable just skip the register and show whatever is the input of the register
			throw(string("01036. Bitwidth computation does not handle registers yet."));
		default:
			throw(string("01017. Not supported DfgNode type for bitwitdh propagation."));
			break;
		}
		assert(pre);
		if (oldpre) {
			if (pre->getSize() < oldpre->getSize()) {
				_op->set(pre);
				delete oldpre;
			} else {
				delete pre;
			}
		} else {
			_op->set(pre);
		}
	}

#ifdef FUNCTIONS_DEPRECATED
	unsigned int DfgNode::collectDFS(vector <DfgNode*> & vNodes,
									 map <DfgNode*, unsigned int>*  mNodeLevels) {
		map <DfgNode*, unsigned int>::iterator mp;
		unsigned int left_level, right_level, level;
		assert(mNodeLevels);
		if((mp = mNodeLevels->find(this))!= mNodeLevels->end())
			return mp->second;
		if(isOp()) {
			left_level = getLeft()->collectDFS(vNodes, mNodeLevels);
			right_level = getRight()->collectDFS(vNodes, mNodeLevels);
			// From both left and right subgraphs, choose the max level
			level = MAX(left_level, right_level)+1;
		} else {
			// Variables and Constant values do not increment the DFG level
			level = 0;
		}
		// pair "this" node with current "level"
		(*mNodeLevels)[this] = level;
		vNodes.push_back(this);
		return level;
	}
#endif

	void DfgNode::_CollectBalanceSupportAndCone(vector <DfgNode*> &vSupp, vector <DfgNode*> &vCone) {
		vCone.push_back(this);
		if(this->getLeft()) {
			if(this->getLeft()->numRefs() == 1 && ( (this->getLeft()->isMul()&& this->isMul()) || (this->getLeft()->isAddorSub()&&this->isAddorSub()) ) ) {
				this->getLeft()->_CollectBalanceSupportAndCone(vSupp, vCone);
			} else {
				vSupp.push_back(this->getLeft());
			}
		}
		if (this->isOnechild()) {
			return;
		}
		if(this->getRight()) {
			if(this->getRight()->numRefs() == 1 &&((this->getRight()->isMul()&& this->isMul())||(this->getRight()->isAddorSub()&&  this->isAddorSub()))) {
				this->getRight()->_CollectBalanceSupportAndCone(vSupp, vCone);
			} else {
				vSupp.push_back(this->getRight());
			}
		}
	}

	void DfgNode::_UpdateSupportPhase_rec(map <DfgNode*, bool> & mPhase, unsigned int nNegs) {
		map <DfgNode*, bool>::iterator p;

		if((p = mPhase.find(this))!= mPhase.end()) {
			if((nNegs%2) == 0) {
				p->second = true;
			} else {
				p->second = false;
			}
			return;
		}
		getLeft()->_UpdateSupportPhase_rec(mPhase, nNegs);
		if(this->isSub()) {
			getRight()->_UpdateSupportPhase_rec(mPhase, nNegs + 1);
		} else {
			getRight()->_UpdateSupportPhase_rec(mPhase, nNegs);
		}
	}

	void DfgNode::balanceAreaRecovery(set <DfgNode*> & sVisited) {
		DfgNode* pLeft,* pRight,* pTemp;

		if(sVisited.find(this)!= sVisited.end())return;
		if((pLeft = getLeft())!= NULL)pLeft->balanceAreaRecovery(sVisited);
		if((pRight= getRight())!= NULL)pRight->balanceAreaRecovery(sVisited);

		if(isOp()&& getRequiredTime()> getArrivalTime()
		   && pLeft != NULL && pLeft->isOp()&& pLeft->getRequiredTime()> pLeft->getArrivalTime()
		   && pRight != NULL && pRight->isOp()&& pRight->getRequiredTime()> pRight->getArrivalTime()) {
			bool bFlag = true;
			if(isAddorSub()) {
				if(pRight->isAddorSub()) {
					;
				} else if(pLeft->isAddorSub()) {
					pTemp = pRight;
					pRight = pLeft;
					pLeft = pTemp;
				} else {
					bFlag = false;

				}
			}
			if(isMul()) {
				if(pRight->isMul()) {
					;
				} else if(pLeft->isMul()) {
					pTemp = pRight;
					pRight = pLeft;
					pLeft = pTemp;
				} else {
					bFlag = false;
				}
			}

			if(bFlag && pRight->getRight()!= NULL && pRight->getLeft()!= NULL && pRight->numRefs() == 1) {
				pTemp = pRight->getRight();
				pRight->setKids(pLeft, pRight->getLeft());
				setKids(pRight, pTemp);
			} else if(bFlag && pLeft->getRight()!= NULL && pLeft->getLeft()!= NULL && pLeft->numRefs() == 1) {
				pTemp = pLeft->getRight();
				pLeft->setKids(pLeft->getLeft(), pRight);
				setKids(pTemp, pLeft);
			}
		}
		sVisited.insert(this);
	}

	void DfgNode::balance(DfgMan* pMan) {

		DFGVector vSupp;
		DFGVector vCone;

		this->_CollectBalanceSupportAndCone(vSupp, vCone);
		if(vSupp.size() == 0) {
			return;
		}
		for(DFGVector::size_type i = 0; i < vSupp.size(); i++) {
			vSupp[i]->balance(pMan);
		}
		_BalanceCone(pMan, vSupp, vCone);
	}


	void DfgNode::_BalanceCone(DfgMan* pMan, DFGVector& vSupp, DFGVector& vCone) {
		if (this->isOnechild() || this->isDiv() || this->isRSH() || this->isLSH()) {
			// balance correctly preserve these node types:
			// - it doesn't make any sense to have a tree of registers
			// - dividers such as (3/5)/(2/4) will translate to (3*4)/(2*5)
			//   in which case the operator divider can not be translated to
			//   TED so there is no point in optimizing these nodes
			// - RSH are still dividers by two, which is not supported in the TED
			return;
		}

		multiset <DfgNode*, _DfgNode_ltDelay> sNodes;
		multiset <DfgNode*, _DfgNode_ltDelay>::iterator p;
		map <DfgNode*, bool> mNodePositivePhase;
		DfgNode* left,* right,* pNode;
		bool bPhase, bPhaseLeft, bPhaseRight;

		for(DFGVector::size_type i = 0; i < vSupp.size(); i++) {
			sNodes.insert(vSupp[i]);
			mNodePositivePhase[vSupp[i]] = true;
		}

		if(this->isAddorSub()) {
			_UpdateSupportPhase_rec(mNodePositivePhase);
		}

		while(sNodes.size()> 1) {
			p = sNodes.begin();
			left  =*p;
			sNodes.erase(p);
			p = sNodes.begin();
			right =*p;
			sNodes.erase(p);
			switch (this->getOp()) {
			case DfgOperator::LSH: // correct fall through
			case DfgOperator::MUL: // correct fall through
				{
					pNode = new DfgNode(pMan, this->getOp(), left, right);
					sNodes.insert(pNode);
					break;
				}
			case DfgOperator::ADD: // fall through
			case DfgOperator::SUB:
				{
					bPhaseLeft = mNodePositivePhase[left];
					bPhaseRight = mNodePositivePhase[right];
					if(bPhaseLeft && bPhaseRight) {
						pNode = new DfgNode(pMan, DfgOperator::ADD, left, right);
						bPhase = true;
					} else if(bPhaseLeft && !bPhaseRight) {
						pNode = new DfgNode(pMan, DfgOperator::SUB, left, right);
						bPhase = true;
					} else if(!bPhaseLeft && bPhaseRight) {
						pNode = new DfgNode(pMan, DfgOperator::SUB, right, left);
						bPhase = true;
					} else {
						pNode = new DfgNode(pMan, DfgOperator::ADD, left, right);
						bPhase = false;
					}
					sNodes.insert(pNode);
					mNodePositivePhase[pNode] = bPhase;
					break;
				}
			case DfgOperator::REG:
			case DfgOperator::EQ: // on trial
			case DfgOperator::LOOPBACK: // on trial
				assert(left==right);
			case DfgOperator::DIV: // on trial
			case DfgOperator::RSH: // on trial
			default:
				{
					throw string("01035. Balance does not support all operations present in the DFG.");
					break;
				}
			}
		}

		p = sNodes.begin();
		pNode =*p;
		sNodes.erase(p);

		if(this->isAddorSub()) {
			assert(pNode->isAddorSub());
			bPhase = pNode->_UpdateConePhase_rec(mNodePositivePhase);
			if(bPhase == false) {
				pNode = pNode->mul(new DfgNode(pMan, -1));
			}
		}

		this->setType(pNode->getOp());
		this->setKids(pNode->getLeft(), pNode->getRight());

		//delete pNode;
		for(DFGVector::size_type i = 0; i < vCone.size(); i++) {
			if(vCone[i] == this)continue;
			delete vCone[i];
		}
	}

	/*
	   void DfgNode::UpdateDelayAndNumRefs(void) {
	   vector <DfgNode*> vNodes;
	   unsigned int dLeft, dRight;
	   collectDFS(vNodes);

	   for(unsigned int i = 0; i < vNodes.size(); i++) {
	   vNodes[i]->_nrefs = 0;
	   if(vNodes[i]->getLeft())vNodes[i]->getLeft()->incRef();
	   if(vNodes[i]->getRight())vNodes[i]->getRight()->incRef();

	   dLeft = dRight = 0;
	   if(vNodes[i]->getLeft())dLeft  = vNodes[i]->getLeft()->getArrivalTime();
	   if(vNodes[i]->getRight())dRight = vNodes[i]->getRight()->getArrivalTime();
	   vNodes[i]->setArrivalTime((dLeft>dRight?dLeft:dRight)+ vNodes[i]->getOpDelay());
	   }
	   }
	   */

	bool DfgNode::_UpdateConePhase_rec(map <DfgNode*, bool> & mPhase) {
		map <DfgNode*, bool>::iterator p;
		bool bRet = false;

		if((p = mPhase.find(this))!= mPhase.end()) {
			return p->second;
		}
		bool bLeft  = getLeft()->_UpdateConePhase_rec(mPhase);
		bool bRight = getRight()->_UpdateConePhase_rec(mPhase);

		unsigned int c = 0;
		if(bLeft)c ^= 0x01;
		if(bRight)c ^= 0x02;

		switch(c) {
		case 0x0: //both negative
			bRet = false;
			break;
		case 0x2: //getLeft  negative
			swapKid();  //than follow  case 1
		case 0x1: //getRight negative
			//this->_type == '-';
			this->setType(DfgOperator::SUB);
			bRet = true;
			break;
		case 0x3: //both positive
			bRet = true;
			break;
		}

		return bRet;
	}

	unsigned int DfgNode::getOpDelay(void)const {
		if(this->isMul()) {
			return delayMul();
		} else if(this->isAdd()) {
			return delayAdd();
		} else if(this->isSub()) {
			return delaySub();
		} else if(this->isLSH() || this->isRSH()) {
			return delayLSH();
		} else if(this->isReg()) {
			return delayReg();
		} else if(this->isAssign() || this->isLoopback()) {
			return 1;
		} else if(this->isDiv()) {
			return delayDiv();
		} else {
			return 0;
		}
	}

	char* DfgNode::getSymbolSave(void) {
		char* p;
		string s;
		switch(getOp()) {
		case DfgOperator::CONST:
			s = Util::itoa(getConst());
			break;
		case DfgOperator::VARCONST: //fall through
		case DfgOperator::VAR:
			s = getName();
			break;
		case DfgOperator::REG: //fall through
		case DfgOperator::ADD: //fall through
		case DfgOperator::SUB: //fall through
		case DfgOperator::MUL: //fall through
		case DfgOperator::DIV: //fall through
		case DfgOperator::LSH: //fall through
		case DfgOperator::RSH: //fall through
		case DfgOperator::EQ: //fall through
		case DfgOperator::LOOPBACK:
			s = getOpStr();
			break;
		default:
			throw(string("01024. UNKNOWN DFG node type"));
		}
		return(p = (char*)Util::strsav(s.c_str()));
	}

	DfgNode* DfgNode::neg(void) {
		DfgNode* pLeft,* pRight;
		switch(getOp()) {
		case DfgOperator::SUB:
			pLeft= getLeft()->neg();
			pRight = getRight()->neg();
			/*
			   if(pLeft != NULL && pRight != NULL) {
			   return(pLeft->sub(pRight));
			   } else if(pLeft == NULL && pRight != NULL) {
			   return(getLeft()->add(pRight));
			   } else if(pLeft != NULL && pRight == NULL) {
			   return(pLeft->add(getRight()));
			   } else {
			   return getRight()->sub(getLeft());
			   }
			   */
			return getRight()->sub(getLeft());
		case DfgOperator::ADD:
			pLeft= getLeft()->neg();
			pRight = getRight()->neg();
			if(pLeft != NULL && pRight != NULL) {
				return getLeft()->neg()->add(getRight()->neg());
			} else {
				return NULL;
			}
		case DfgOperator::MUL:
			pLeft= getLeft()->neg();
			pRight = getRight()->neg();
			if(pLeft == NULL && pRight == NULL) {
				return NULL;
			} else if(pLeft != NULL) {
				return pLeft->mul(getRight());
			} else {
				return pRight->mul(getLeft());
			}
			break;
		case DfgOperator::VAR:
		case DfgOperator::VARCONST:
		case DfgOperator::CONST:
		default:
			return NULL;
		}
	}

	void DfgNode::setEnvOpDelays(void) {
		_addDelay = Environment::getInt("delayADD");
		_subDelay = Environment::getInt("delaySUB");
		_mpyDelay = Environment::getInt("delayMPY");
		_divDelay = 3; //Environment::getInt("delayDfgOperator::DIV");
		_lshDelay = Environment::getInt("delayLSH");
		_rshDelay = Environment::getInt("delayRSH");
		_regDelay = Environment::getInt("delayREG");
	}

	TedVar* DfgNode::getTedVarIfExist(void)const {
		return TedVarMan::instance().getIfExist(getName());
	}

	int DfgNode::getRegisterValue(void)const {
		if (!isReg()) {
			return 0;
		}
		string name = this->getName();
		Util::strSubstitute(name,"R");
		pair<bool,wedge> test = Util::atoi(name.c_str());
		assert(test.first);
		return test.second;
	}
}

