/*
 * =====================================================================================
 *
 *       Filename:  DfgOperator.cc
 *    Description:
 *        Created:  11/22/2009 09:54:13 PM
 *         Author:  Daniel f. Gomez-Prado (dgomezpr), dgomezpr@ecs.umass.edu
 *        Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 110                                          $: Revision of last commit
 *  $Author:: daniel                                         $: Author of last commit
 *  $Date:: 2009-11-20 19:54:30 -0500 (Fri, 20 Nov 2009)     $: Date of last commit
 * =====================================================================================
 */

#include <cstring>

#include "DfgNode.h"
#include "DfgOperator.h"

namespace dfg {

#ifndef INLINE
#include "DfgNode.inl"
#endif

	unsigned int DfgOperator::idcount = 0;
	map<unsigned int,DfgOperator*> DfgOperator::manager;

	DfgOperator::DfgOperator(Bitwidth* pre, Type type): _bitw(pre), _rounding(DN), _id(++idcount), _count(0), _type(type), _error(NULL) {
		manager.insert(pair<unsigned int,DfgOperator*>(_id,this));
	}
	DfgOperator::DfgOperator(Type type): _bitw(NULL), _rounding(DN), _id(++idcount), _count(0), _type(type), _error(NULL) {
		manager.insert(pair<unsigned int,DfgOperator*>(_id,this));
	}
	DfgOperator::~DfgOperator(void) {
		if(_bitw)
			delete _bitw;
		manager.erase(_id);
	};

	DfgOperator::Type DfgOperator::queryType(const char* opStr) {
		Type nodeType = UNKNOWN;
		static const char* REG_XML = "DELAY";
		if(!strcmp(opStr,getOpName(CONST)) || !strcmp(opStr,getOpNameCDFG(CONST))) {
			nodeType=CONST;
		} else if(!strcmp(opStr,getOpName(VAR)) || !strcmp(opStr,getOpNameCDFG(VAR))) {
			nodeType=VAR;
		} else if(!strcmp(opStr,getOpName(VARCONST)) || !strcmp(opStr,getOpNameCDFG(VARCONST))) {
			nodeType=VARCONST;
		} else if(!strcmp(opStr,getOpName(DIV)) || !strcmp(opStr,getOpNameCDFG(DIV))) {
			nodeType=DIV;
		} else if(!strcmp(opStr,getOpName(ADD)) || !strcmp(opStr,getOpNameCDFG(ADD)) ) {
			nodeType=ADD;
		} else if(!strcmp(opStr,getOpName(MUL)) || !strcmp(opStr,getOpNameCDFG(MUL))) {
			nodeType=MUL;
		} else if(!strcmp(opStr,getOpName(SUB)) || !strcmp(opStr,getOpNameCDFG(SUB))) {
			nodeType=SUB;
		} else if(!strcmp(opStr,getOpName(LSH)) || !strcmp(opStr,getOpNameCDFG(LSH))) {
			nodeType=LSH;
		} else if(!strcmp(opStr,getOpName(RSH)) || !strcmp(opStr,getOpNameCDFG(RSH))) {
			nodeType=RSH;
		} else if(!strcmp(opStr,getOpName(REG))  || !strcmp(opStr,getOpNameCDFG(REG)) || !strcmp(opStr,REG_XML)) {
			nodeType=REG;
		} else if(!strcmp(opStr,getOpName(EQ)) || !strcmp(opStr,getOpNameCDFG(EQ))) {
			nodeType=EQ;
		} else if(!strcmp(opStr,getOpName(LOOPBACK)) || !strcmp(opStr,getOpNameCDFG(LOOPBACK))) {
			nodeType=LOOPBACK;
		}
		return nodeType;
	}

	const char* DfgOperator::getOpName(Type op) {
		//Note: END_OP_TYPE is the last element and it is only a placeholder to the index of the array
		static const char* strBuff[END_OP_TYPE] = {"UNKNOWN","CONST","VAR","VARCONST","DIV","ADD","MUL","SUB","LSH","RSH","REG","EQ","LOOPBACK"};
		return strBuff[op];
#if 0
		switch(op) {
		case DfgOperator::ADD: return "ADD";
		case DfgOperator::MUL: return "MUL";
		case DfgOperator::DIV: return "DIV";
		case DfgOperator::SUB: return "SUB";
		case DfgOperator::LSH: return "LSH";
		case DfgOperator::RSH: return "RSH";
		case DfgOperator::REG: return "REG";
		case DfgOperator::EQ: return  "EQ";
		case DfgOperator::LOOPBACK: return "LOOPBACK";
		default: return "UNKNOWN";
		}
#endif
	}

	const char* DfgOperator::getOpNameXML(Type op) {
		//Note: END_OP_TYPE is the last element and it is only a placeholder to the index of the array
		static const char* strBuff[END_OP_TYPE] = {"UNKNOWN","VARIABLE_VAL_UNKNOWN","VARIABLE_VAL_UNKNOWN","VARIABLE_VAL_UNKNOWN","DIV","ADD","MUL","SUB","SLL","SRL","DELAY","EQ","LOOPBACK"};
		return strBuff[op];
	}

	const char* DfgOperator::getOpNameCDFG(Type op) {
		//Note: END_OP_TYPE is the last element and it is only a placeholder to the index of the array
		static const char* strBuff[END_OP_TYPE] = {"unknown","const","var","varconst","div","add","mul","sub","sll","srl","reg","assign","loopback"};
		return strBuff[op];
	}
}
