/*
   l
 * =====================================================================================
 *
 *       Filename:  ATedNode.cc
 *    Description:
 *        Created:  11/10/2008 7:23:33 PM
 *         Author:  Daniel f. Gomez-Prado(dgomezpr), dgomezpr@ecs.umass.edu
 *        Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 239                                          $: Revision of last commit
 *  $Author:: daniel@win                                     $: Author of last commit
 *  $Date:: 2010-11-18 08:13:53 -0500(Thu, 18 Nov 2010)$: Date of last commit
 * =====================================================================================
 */

#include <cassert>
#include <list>
#include <iostream>
#include <sstream>
#include <algorithm>

#include "ATedNode.h"
#include "TedNode.h"
#include "TedMan.h"
#include "TedKids.h"
#include "TedVar.h"
#include "TedContainer.h"

#include "ETedNode.h"

namespace ted {

	TEDSet ATedNode::_transitoryContainer;
	TedContainer* ATedNode::_permanentContainer = NULL;

	void ATedNode::setContainer(TedContainer& tedCont) {
		_permanentContainer = &tedCont;
	}

	bool ATedNode::checkContainer(TedContainer& tedCont) {
		return(_permanentContainer == &tedCont);
	}

	void ATedNode::purge_transitory_container(void) {
		TEDSet::iterator it = _transitoryContainer.begin();
		while (it != _transitoryContainer.end()) {
			TedNode* pNode = (*it);
			assert(pNode);
			if (pNode->visited.isMarked()) {
				_transitoryContainer.erase(it++);
			} else {
				++it;
				delete pNode;
			}
		}
		//All nodes that remains are located in a TedMan::getContainer()
		//so it is safe to dispose of the _transitoryContainer
		_transitoryContainer.clear();
	}

	void ATedNode::garbage_collect_unmarked(bool (Mark<TedNode>::*functor)(void) const) {
		TEDSet::iterator it = _transitoryContainer.begin();
		while (it != _transitoryContainer.end()) {
			TedNode* pNode = (*it);
			assert(pNode);
			if (!(pNode->visited.*functor)()) {
				_transitoryContainer.erase(it++);
				delete pNode;
			} else {
				++it;
			}
		}
	}

	ATedNode* ATedNode::duplicate(void) const {
		// only copies the varname
		ATedNode* ret = this->clone();
		assert(ret->_tVar);
		assert(ret->_tKids==NULL);
		if (this->getKids()) {
			ret->setKids(this->getKids()->clone());
		}
		return ret;
	}

	/** @brief Duplicate a ETed recursively
	 *  @details The duplicated nodes are stored in the transitory container
	 **/
	ATedNode* ATedNode::duplicateRecursive(void) const {
		//Mark<TedNode>::newMark();
		map<const ATedNode*,ATedNode*> old2new;
		ATedNode* tmp = duplicateRecursiveRec(old2new);
		//Mark<TedNode>::newMark();
		return tmp;
	}

	ATedNode* ATedNode::duplicateRecursiveRec(map<const ATedNode*,ATedNode*>& old2new) const {
		//if(visited.isMarked()) {
		//	return old2new[this];
		//}
		if (old2new.find(this)!=old2new.end()) {
			return old2new[this];
		}
		//unsigned int index = 0;
		//ATedNode* pKid = NULL;
		//wedge      weight = 0;
		ATedNode* t = this->clone();
		_transitoryContainer.insert(t);
		if(!this->isConst()) {
			t->_tKids = TedKids::create_kids();
			FOREACH_KID_OF_NODE(this) {
				unsigned int index = _iterkids.getIndex();
				ATedNode* pKid = _iterkids.Node<ATedNode>();
				t->addKid(index,pKid->duplicateRecursiveRec(old2new),1);
			}
		}
		//visited.setMark();
		old2new[this] = t;
		return t;
	}

	void ATedNode::print_transitory_container(void) {
#if defined(_DEBUG)
		//required _transitorycontainer to be the TEDSet class defined while debugging
		_transitoryContainer.print2file();
#endif
	}

	void ATedNode::release_from_transitory_container(TedNode* pNode) {
		TEDSet::iterator it = _transitoryContainer.find(pNode);
		if (it!=_transitoryContainer.end()) {
			_transitoryContainer.erase(it);
		}
	}

	bool ATedNode::is_transitory_container_empty() {
		return _transitoryContainer.empty();
	}

	ATedNode* ATedNode::getDummyRoot(void) const {
		if (isDummy()) {
			wedge val = getKidWeight(0);
			ETedNode* ret = dynamic_cast<ETedNode*>((dynamic_cast<ATedNode*>(getKidNode(0)))->getDummyRoot());
			ATedNode* fin =	ret->mul_safe(new ETedNode(val));
			return fin;
		}
		return this->duplicate();
	}

	void ATedNode::purge_transitory_container_cone(ATedNode* atNode) {
		TEDVector vnodes;
		atNode->collectDFS(vnodes);
		for (TEDVector::iterator it = vnodes.begin(); it != vnodes.end(); it++) {
			TedNode* pnode = dynamic_cast<ATedNode*>(*it);
			if (!pnode || pnode->visited.isPMarked() || pnode==TedNode::getOne()) {
				continue;
			}
			delete pnode;
		}
		vnodes.clear();
	}

}
