/*
 * =====================================================================================
 *
 *        Filename:  TEDMan.cc
 *     Description:  Tedman class
 *         Created:  04/17/2007 12:24:18 PM EDT
 *          Author:  Daniel Gomez-Prado(from jan 08), Qian Ren(up to dec 07)(R&D Engineer), qren@ren-chao.com
 *         Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 272                                          $: Revision of last commit
 *  $Author:: daniel@win                                     $: Author of last commit
 *  $Date:: 2010-11-18 08:13:53 -0500(Thu, 18 Nov 2010)$: Date of last commit
 * =====================================================================================
 */

#include <iostream>
#include <sstream>
#include <fstream>
#include <cassert>
#include <algorithm>
#include <list>
#include <cmath>

#include "TedMan.h"
#include "TedNode.h"
#include "TedKids.h"
#include "TedVar.h"
#include "ETedNode.h"
#include "TedVarGroup.h"
#include "TedVarMan.h"
#include "TedNodeIterator.h"
#include "TedParents.h"
#include "TedDecompose.h"
#include "TedBitwidth.h"

#include "Bitwidth.h"
using namespace data;

#include "BinaryThreadedTree.h"
using dtl::BinaryThreadedTree;
using dtl::BinaryThreadedNode;

#include "Environment.h"
using namespace tds;

#include "DfgNode.h"
#include "DfgMan.h"
using dfg::DfgMan;
using dfg::DfgNode;
using dfg::DfgStat;

#include "util.h"
using util::Util;

#include "Convert.h"
using convert::Convert;

namespace ted {
#ifndef INLINE
#include "TedMan.inl"
#endif

	unsigned int TedMan::_factor_counter = 1;

	TedMan::const_dfs_iterator TedMan::const_dfs_begin(void)const{
		return const_dfs_iterator(this,const_dfs_iterator::begin);
	}

	TedMan::const_dfs_iterator TedMan::const_dfs_end(void)const{
		return const_dfs_iterator(const_dfs_iterator::end);
	}

	TedMan::dfs_iterator TedMan::dfs_begin(void) {
		return dfs_iterator(this,dfs_iterator::begin);
	}

	TedMan::dfs_iterator TedMan::dfs_end(void) {
		return dfs_iterator(dfs_iterator::end);
	}

	TedMan::TedMan(void): _vars(TedVarMan::instance()) {
		clusterA = NULL;
		clusterD = NULL;
		_const_limit = 0;
		ATedNode::setContainer(_container);
	}

	TedMan::~TedMan(void) {
		if (clusterA) {
			for( TedManCluster::iterator it = clusterA->begin(), it_end = clusterA->end(); it != it_end; it++) {
				TedMan* manager = it->second;
				delete manager;
			}
			clusterA->clear();
			delete clusterA;
			clusterA = NULL;
		}
		if (clusterD) {
			for( TedManCluster::iterator it = clusterD->begin(), it_end = clusterD->end(); it != it_end; it++) {
				TedMan* manager = it->second;
				delete manager;
			}
			clusterD->clear();
			delete clusterD;
			clusterD = NULL;
		}
		// Do nothing
		// the containers have to be deallocated on its own
		// the _vars has to be cleared by TedVarMan
	}

	void TedMan::purge_all_but(TedMan* p1) {
		assert(p1);
		Mark<TedNode>::newMark();
		Mark<TedKids>::newMark();
		p1->getContainer().update_TMark();
		ATedNode::purge_transitory_container();
		TedKids::purge_transitory_container();
	}

	void TedMan::purge_all_but(TedMan* p1,TedMan* p2) {
		assert(p1);
		assert(p2);
		Mark<TedNode>::newMark();
		Mark<TedKids>::newMark();
		p1->getContainer().update_TMark();
		p2->getContainer().update_TMark();
		ATedNode::purge_transitory_container();
		TedKids::purge_transitory_container();
	}

	void TedMan::purge_all_but(TedMan* p1,TedMan* p2, TedMan* p3) {
		assert(p1);
		assert(p2);
		assert(p3);
		Mark<TedNode>::newMark();
		Mark<TedKids>::newMark();
		p1->getContainer().update_TMark();
		p2->getContainer().update_TMark();
		p3->getContainer().update_TMark();
		ATedNode::purge_transitory_container();
		TedKids::purge_transitory_container();
	}

	bool TedMan::check_for_dangling_nodes(void) {
		TedParents parents;
		this->preOrderDFS(parents, &TedParents::collect);
		//this->preOrderDFSforPOs(parents, &TedParents::collect);
		for(TedContainer::const_iterator it = _container.begin(); it != _container.end(); it++) {
			for(TEDSet::const_iterator jt = it->second.second->begin(); jt != it->second.second->end(); jt ++) {
				TedNode* pNode = (*jt);
				assert(pNode);
				assert(pNode->getKids());
				if (parents.find(pNode) ==parents.end()) {
					if (!this->isPO(pNode)) {
#if _DEBUG

						_container[*pNode->getVar()].second->visualize(pNode,&parents,1,2);
#endif
						return false;
					}
				}
			}
		}
		return true;
	}

	/** @brief Normalize a ATedNode and put resulting TedNodes in manager*/
	TedNodeRoot TedMan::normalize(ATedNode* pANode) {
		pANode = _PropogateZero_Rec(pANode);
		if(pANode) {
			pair<TedNode*, wedge> pe = _Normalize_Rec(pANode);
			return TedNodeRoot(pe.first, pe.second);
		} else {
			return TedNodeRoot(TedNode::getOne(),0);
		}
	}

	/** @brief routine to remove Zeros in ATed*/
	ATedNode* TedMan::_PropogateZero_Rec(ATedNode* pANode) {
		if (!pANode)
			return NULL;

		unsigned int index = 0;
		ATedNode* pEKid = NULL,* pETemp = NULL;
		wedge weight = 0;

		if(pANode->isConst()) {
			if(pANode->getConst() == 0) {
				return NULL;
			} else {
				return pANode;
			}
		}
		list<unsigned int> index2erase;
		for(TedKids::iterator _iterkids = pANode->getKids()->begin(), _iterend = pANode->getKids()->end();
			_iterkids != _iterend; _iterkids++) {
			//FOREACH_KID_OF_ANODE(pANode, index, pEKid, weight)
			index = _iterkids->first;
			pEKid = (ATedNode*)_iterkids->second.Node();
			TedRegister r = _iterkids->second.getRegister();
			weight = _iterkids->second.getWeight();
			assert(pEKid);
			pETemp = _PropogateZero_Rec(pEKid);
			if(pETemp == NULL) {
				index2erase.push_back(index);
			} else {
				pANode->replaceKid(index, pETemp, 1,r);
			}
		}
		while(!index2erase.empty()) {
			pANode->removeKid(index2erase.back());
			index2erase.pop_back();
		}
		if(pANode->numKids() == 0)
			return NULL;
		if(pANode->numKids() == 1 && pANode->hasKid(0)) {
			pANode = (ATedNode*)pANode->getKidNode(0);
		}
		return pANode;
	}

	/** @brief routine to normalize a ATedNode, returning a Ted and the greatest common divisor*/
	pair <TedNode*, wedge > TedMan::_Normalize_Rec(TedNode* pANode) {
		assert(pANode);
		TedKids tKids;
		unsigned int index = 0;
		TedNode* pEKid = NULL;
		if(pANode->isConst()) {
			return pair<TedNode*, wedge>(TedNode::getOne(), pANode->getConst());
		}
		FOREACH_KID_OF_NODE(pANode) {
			index = _iterkids.getIndex();
			pEKid = _iterkids.Node<TedNode>();
			pair<TedNode*,wedge> pEdge = _Normalize_Rec(pEKid);
			tKids.insert(index, pEdge.first, pEdge.second,_iterkids.getRegister());
		}
		wedge gcd = tKids.extractGCD();
		const TedVar* pVar = pANode->getVar();
		TedNode* pNode = _container.getNode(*pVar, tKids, pANode->getRegister());
		return pair<TedNode*, wedge>(pNode, gcd);
	}

	/** @brief Link a TedNode to Po, with name poname, weight weight*/
	void TedMan::linkTedToPO(TedNode* pNode, wedge weight, const string& poname) {
		TedNodeRoot po(pNode, weight);
		linkTedToPO(po, poname);
	}

	/** @brief Link a TedNodeRoot to name poname, if poname is Null, it is auto assigned*/
	void TedMan::linkTedToPO(const TedNodeRoot& po, const string& poname) {
		string name;
		if(poname.empty()) {
			name = "F"+Util::itoa(_pos.size());
		} else {
			name = poname;
		}
		if (_pos.find(name) ==_pos.end()) {
			_pos.insert(pair<string,TedNodeRoot>(name,po));
		} else {
			_pos[name] = po;
		}
	}

	bool TedMan::isPO(TedNode* node) const {
		for(PrimaryOutputs::const_iterator p = _pos.begin(); p != _pos.end(); p++) {
			if(node == p->second.Node()) {
				return true;
			}
		}
		return false;
	}


	TedNodeRoot* TedMan::getPO(TedNode* node) {
		for(PrimaryOutputs::iterator p = _pos.begin(); p != _pos.end(); p++) {
			if(node == p->second.Node()) {
				return &(p->second);
			}
		}
		return NULL;
	}

	/** @brief Unlinking a primary output*/
	TedMan* TedMan::extractPO(set <string> se) {
		set <string > s;
		for(PrimaryOutputs::iterator p = _pos.begin(); p != _pos.end(); p++) {
			if(se.find(p->first) == se.end()) {
				s.insert(p->first);
			}
		}
		return _DuplicateUnlink(& s);
	}

	/** @brief Collect nodes in the order of DFS*/
	void TedMan::collectDFS(TEDVector& vNodes) {
		Mark<TedNode>::newMark();
		for(PrimaryOutputs::iterator p = _pos.begin(); p != _pos.end(); p++) {
			p->second.Node()->_collectDFS_rec(vNodes);
		}
	}



	/** @brief For substitution routine
	 *  @note the vector container forces the variables to be order by their names
	 **/
	void TedMan::getVars(vector<string>& vVars) {
		char** pvVars = (char**)malloc(sizeof(char*)* _container.size());
		for(TedContainer::iterator p = _container.begin(); p != _container.end(); p++) {
			*(pvVars + p->second.first-1) = (char*)p->first.getName().c_str();
		}
		for(TedContainer::size_type i = 0; i < _container.size(); i++) {
			string saveVar(*((char**)(pvVars + i)));
			vVars.push_back(saveVar);
		}
		free(pvVars);
	}

	/** @brief For substitution routine
	 *  @note the list container guarantees that the variables are order by their position in the TED
	 **/
	void TedMan::getVars(list<TedVar*>& vVars) {
		TedContainerOrder tnodesOrdered = _container.order();
		for(TedContainerOrder::const_iterator it = tnodesOrdered.begin(); it!= tnodesOrdered.end(); it++) {
			TedVar* pvar  = checkVariables(it->second.first->getName());
			vVars.push_back(pvar);
		}
	}

	/** @brief to duplicate TedManager, when poname is not NULL, the po is unlinked*/
	TedMan* TedMan::_DuplicateUnlink(set <string>* ponames) {
		assert(ATedNode::is_transitory_container_empty());
		assert(TedKids::is_transitory_container_empty());
		TedMan* pManNew = new TedMan();
		TedContainer::iterator msp;
		map <TedNode*, TedNode*> mVisited;
		TEDSet* sNodes;
		TedNode* pNode;
		set <const char*>::iterator p;
		map <unsigned int, unsigned int> mNewLevel;
		vector <TedVar> vVars;
		for(msp = _container.begin(); msp!= _container.end(); msp++) {
			sNodes = new TEDSet;
			pManNew->_container[TedVar(msp->first)] = pair<unsigned int, TEDSet*>(msp->second.first, sNodes);
		}
		for(PrimaryOutputs::iterator mp = _pos.begin(); mp!= _pos.end(); mp++) {
			if(ponames != NULL && ponames->size()!= 0) {
				if(ponames->find(mp->first.c_str())!= ponames->end())continue;
			}
			pNode = _Duplicate_Rec(pManNew, mp->second.Node(), mVisited);
			TedNodeRoot po(pNode,mp->second.getWeight(),mp->second.getType());
			pManNew->linkTedToPO(po, mp->first);

		}
		//removing empty levels
		if(ponames != NULL && ponames->size()!= 0) {
			for(msp = pManNew->_container.begin(); msp!= pManNew->_container.end(); msp++) {
				if(msp->second.second->size() == 0) {
					delete msp->second.second;
					vVars.push_back(msp->first);
					mNewLevel[msp->second.first] = 0;
				} else {
					mNewLevel[msp->second.first] = msp->second.first;
				}
			}
			for(vector<TedVar>::size_type index = 0; index < vVars.size(); index++) {
				pManNew->_container.erase(vVars[index]);
			}
			unsigned int i = 0;
			for(map <unsigned int, unsigned int>::iterator p = mNewLevel.begin(); p != mNewLevel.end(); p++) {
				if(p->second == 0) {
					i++;
				}
				p->second -= i;
			}
			for(msp = pManNew->_container.begin(); msp!= pManNew->_container.end(); msp++) {
				msp->second.first = mNewLevel[msp->second.first];
			}
		}
		pManNew->purge_transitory_containers();
		assert(this->check_for_dangling_nodes());
		assert(this->_container.check_children_are_included());
		return pManNew;
	}

	/** @brief Duplicate recursively a TedNode*/
	TedNode* TedMan::_Duplicate_Rec(TedMan* pManNew, TedNode* pNode, map <TedNode*, TedNode*> & mVisited) {
		map <TedNode*, TedNode*>::iterator mp;
		TedNode* pNodeNew = NULL,* pKid = NULL,*pKidNew = NULL;
		unsigned int index =0;
		wedge weight = 0;
		TedKids tKids;
		if((mp=mVisited.find(pNode))!= mVisited.end())
			return mp->second;
		if(pNode == TedNode::getOne())
			return pNode;
		if(pNode->isConst())
			return(new ETedNode(pNode->getConst()));
		FOREACH_KID_OF_NODE(pNode) {
			index = _iterkids.getIndex();
			pKid = _iterkids.Node<TedNode>();
			weight = _iterkids.getWeight();
			pKidNew = _Duplicate_Rec(pManNew, pKid, mVisited);
			TedRegister kidReg = _iterkids.getRegister();
			tKids.insert(index, pKidNew, weight,kidReg);
		}
		pNodeNew = new ETedNode(*pNode->getVar(), tKids, pNode->getRegister());
		pManNew->_container[*(pNode->getVar())].second->insert(pNodeNew);
		mVisited[pNode] = pNodeNew;
		return pNodeNew;
	}

	/** @brief Linearize a Ted by splitting non-linear variables*/
	TedMan* TedMan::linearize(void) {
		TedMan* pManNew = new TedMan();
		TedContainerOrder mNodesOrdered = _container.order();
		unsigned int maxOrder = 0;
		for(TedContainerOrder::iterator it = mNodesOrdered.begin(); it!= mNodesOrdered.end(); it++)
			if(it->second.first->isVar()&& !it->second.second->empty()) {
				maxOrder = 0;
				for(TEDSet::iterator jt = it->second.second->begin(); jt != it->second.second->end(); jt++) {
					if(maxOrder <(*jt)->getKids()->highestOrder()) {
						maxOrder = (*jt)->getKids()->highestOrder();
					}
				}
				assert(maxOrder>0);
				TedVar* currentVar = it->second.first;
				unsigned int keyLevel = pManNew->_container.getMaxLevel()+ 1;
				if(1 == maxOrder) { //It is already linear
					assert(_vars.getIfExist(currentVar->getName()));
					TEDSet* sNodes = new TEDSet;
					pManNew->_container[*currentVar] = pair<unsigned int, TEDSet*>(keyLevel, sNodes);
				} else { //It needs to be linearized
					TedVarGroupOwner* owner = _vars.createOwner(*currentVar);
					for(unsigned int i = 0; i < maxOrder; i++) {
						TEDSet* sNodes = new TEDSet;
						TedVarGroupMember* member = _vars.createMember(*owner);
						//member object will be cast to a simple TedVar object, therefore by looking at the
						//keys in the container we will not be able to see if the key was a member anymore.
						//we will need to retrieve this information through TedMan::checkVariables
						pManNew->_container[(*member)] = pair<unsigned int, TEDSet*>(keyLevel, sNodes);
						keyLevel++;
					}
				}
			}
		map <TedNode*, TedNode*> mVisited;
		for(PrimaryOutputs::iterator it = _pos.begin(); it!= _pos.end(); it++) {
			TedNode* pNode = _Linearize_Rec(pManNew, it->second.Node(), mVisited);
			TedNodeRoot root(pNode,it->second.getWeight());
			root.setRegister(it->second.getRegister());
			pManNew->linkTedToPO(root, it->first);
		}
		return pManNew;
	}

	/** @brief Linearize recursively a TedNode*/
	TedNode* TedMan::_Linearize_Rec(TedMan* pManNew, TedNode* pNode, map <TedNode*, TedNode*> & mVisited) {
		TedNode* pNodeNew = NULL,* pKid = NULL,*pKidNew = NULL;
		wedge weight = 0, gcd = 0;
		TedRegister retime_count = 0;
		TedKids tKids;
		string strNewVar;
		map<TedNode*,TedNode*>::iterator it = mVisited.find(pNode);
		if(it != mVisited.end()) {
			return it->second;
		}
		if(pNode == TedNode::getOne())
			return pNode;
		for(int i = pNode->getKids()->highestOrder(); i >= 0; i--) {
			if(i == pNode->getKids()->highestOrder()) {
				pKid = pNode->getKidNode(i);
				weight = pNode->getKidWeight(i);
				retime_count = pNode->getKidRegister(i);
				pKidNew = _Linearize_Rec(pManNew, pKid, mVisited);
				tKids.insert(1, pKidNew, weight,retime_count);
			} else {
				if(pNode->getKids()->has(i)) {
					pKid = pNode->getKidNode(i);
					weight = pNode->getKidWeight(i);
					retime_count = pNode->getKidRegister(i);
					pKidNew = _Linearize_Rec(pManNew, pKid, mVisited);
					tKids.insert(0, pKidNew, weight,retime_count);
				}
				gcd = tKids.extractGCD();
				const TedVar* var = _vars.getIfExist(pNode->getName());
				if(var) {
					if(var->isOwner()) {
						const TedVarGroupOwner* owner = dynamic_cast<const TedVarGroupOwner*>(var);
						var = owner->getMember(i+1);
					} //if var is member or simple, its var itself
				} else {
					var = pNode->getVar();
				}
				pNodeNew = pManNew->_container.getNode(*var, tKids,pNode->getRegister());
				tKids.clear();
				tKids.insert(1, pNodeNew, gcd,TedRegister(0));
			}
		}
		mVisited[pNode] = pNodeNew;
		return pNodeNew;
	}

	void TedMan::info(bool detailed) {
		TedParents parents;
		preOrderDFS(parents, &TedParents::collect);
		TedContainerOrder mNodesOrdered = _container.order();
		unsigned int goodContainers = 0;
		unsigned int nodeCount = 0;
		unsigned int total_kids = 0;
		if (detailed) {
			cout << "|---------------------------------------------------------------------------------|" << endl;
			cout << "|  Level  |     name     | node count | fanins(spread) | fanouts(spread) | output |" << endl;
			cout << "|---------------------------------------------------------------------------------|" << endl;
		}
		for(TedContainerOrder::iterator it = mNodesOrdered.begin(); it!= mNodesOrdered.end(); it++) {
			if (it->second.second->empty())
				continue;
			goodContainers++;
			nodeCount += it->second.second->size();
			if (detailed) {
				cout << "| ";
				cout.fill(' ');
				cout.width(7);
				cout << _container.size() + 1 - it->first;
				cout << " | ";
				cout.fill(' ');
				cout.width(12);
				cout << it->second.first->getName();
				cout << " | ";
				cout.fill(' ');
				cout.width(10);
				cout << it->second.second->size();
				unsigned int fanin_count = 0;
				set<unsigned int> fanin_spread;
				unsigned int fanout_count = 0;
				set<unsigned int> fanout_spread;
				unsigned int parent_count = 0;
				for (TEDSet::iterator jt = it->second.second->begin(); jt != it->second.second->end(); jt++) {
					TedNode* node = (*jt);
					fanin_count += node->numKids();
					total_kids += fanin_count;
					FOREACH_KID_OF_NODE(node) {
						TedNode* child = _iterkids.Node<TedNode>();
						fanin_spread.insert(_container.getLevel(*child->getVar()));
					}
					if (parents.hasParents(node)) {
						MyParent* fanout = parents[node];
						fanout_count += fanout->size();
						for(MyParent::iterator kt = fanout->begin(); kt != fanout->end(); kt++) {
							TedNode* parent = kt->first;
							fanout_spread.insert(_container.getLevel(*parent->getVar()));
						}
					} else {
						parent_count++;
					}
				}
				cout << " | ";
				cout.fill(' ');
				cout.width(6);
				cout << fanin_count;
				cout << "(";
				cout.fill(' ');
				cout.width(6);
				cout << fanin_spread.size();
				cout <<") | ";
				cout.fill(' ');
				cout.width(7);
				cout << fanout_count;
				cout << "(";
				cout.fill(' ');
				cout.width(6);
				cout << fanout_spread.size();
				cout <<") | ";
				cout.fill(' ');
				cout.width(6);
				cout << parent_count << " |";
				cout << endl;
			}
		}
		if (detailed) {
			cout << "|---------------------------------------------------------------------------------|" << endl;
		}
		cout << "TED has " << goodContainers << " variables and " << nodeCount << " nodes" << endl;
		cout << "|---------------------------------------------------------------------------------|" << endl;
		cout << "TedNode =" << sizeof(TedNode) << "bytes TedKids="<<sizeof(TedKids)<< "bytes TedKid="<<sizeof(TedKid) << "bytes" << endl;
		if (detailed) {
			cout << "Memory = " << sizeof(TedNode)*(nodeCount+1) + sizeof(TedKids)*(nodeCount)+sizeof(TedKid)*total_kids << "bytes" << endl;
		}
	}

	void TedMan::cost(bool avoid_ntl_extraction) {
		ReportCost repcost;
		if (avoid_ntl_extraction) {
			repcost.avoid_ntl_extraction();
		}
		FixCost* rc = repcost.compute_cost(*this);
		unsigned int edge0 = 0;
		unsigned int edgeN = 0;
		costNaive(edge0,edgeN);
		cout << "|--------------------------------------------|" << endl;
		cout << "|                   TED                      |" << endl;
		cout << "|--------------------------------------------|" << endl;
		cout << "|  node | edge0 |  edgeN  | factor |  width  |" << endl;
		cout << "|--------------------------------------------|" << endl;
		cout << "| ";
		cout.fill(' ');
		cout.width(5);
		cout << rc->nodes << " | ";
		cout.width(5);
		cout << edge0 << " | ";
		cout.width(7);
		cout << edgeN << " | ";
		cout.width(6);
		cout << rc->normB << " | ";
		cout.width(7);
		cout << rc->bitwidth << " | ";
		cout << endl;
		cout << "|--------------------------------------------|" << endl;
		cout << "|        DFG         |        Schedule       |" << endl;
		cout << "|--------------------|-----------------------|" << endl;
		cout << "| nMUL | nADD | nSUB | Latency | rMPY | rADD |" << endl;
		cout << "| ";
		cout.width(4);
		cout << rc->stat.nMul << " | ";
		cout.width(4);
		cout << rc->stat.nAdd << " | ";
		cout.width(4);
		cout << rc->stat.nSub << " | ";
		cout.width(7);
		cout << rc->stat.latency << " | ";
		cout.width(4);
		cout << rc->stat.rMPY << " | ";
		cout.width(4);
		cout << rc->stat.rADD << " | ";
		cout << endl;
		cout << "|--------------------------------------------|" << endl;
		cout << "|                   Gaut                     |" << endl;
		cout << "|--------------------------------------------|" << endl;
		cout << "|   Muxes  |  Latency  |  Register  |  Area  |" << endl;
		cout << "| ";
		cout.width(8);
		cout << rc->muxes << " | ";
		cout.width(9);
		cout << rc->latency << " | ";
		cout.width(10);
		cout << rc->registers << " | ";
		cout.width(6);
		cout << rc->area << " |" << endl;
		delete rc;
	}

	/** @brief Get the number of ADD edges and MPY edges*/
	void TedMan::costNaive(unsigned int& nbadd, unsigned int& nbmpy) {
		TedNode* pNode,* pKid;
		wedge weight;
		unsigned int index;
		TedContainer::iterator msp;
		set <TedNode*>::iterator sp;

		/*creating the references*/
		for(msp = _container.begin(); msp != _container.end(); msp ++) {
			for(sp = msp->second.second->begin(); sp != msp->second.second->end(); sp++) {
				pNode =*sp;
				FOREACH_KID_OF_NODE(pNode) {
					index = _iterkids.getIndex();
					pKid = _iterkids.Node<TedNode>();
					weight = _iterkids.getWeight();
					if(index==0) {nbadd++; }
					if(index==1&&pKid!=TedNode::getOne()) {nbmpy++; }
					if(weight!=-1 && weight!=1) {nbmpy++; }
				}
			}
		}
		//cout<<"#ADD="<<nbadd<<" #MPY="<<nbmpy<<endl;
	}

	/** @brief Get the number of factor candidates*/
	unsigned int TedMan::getNumOfCandidates(void) {
		TedNode* pNode = NULL,* pKid = NULL;
		//unsigned int index = 0;
		TedContainer::iterator msp;
		map <TedNode*, unsigned int> mRefs;
		map <TedNode*, unsigned int>::iterator mp;
		set <TedNode*>::iterator sp;
		unsigned int ret = 0;
		for(msp = _container.begin(); msp != _container.end(); msp ++) {
			for(sp = msp->second.second->begin(); sp != msp->second.second->end(); sp++) {
				pNode =*sp;
				FOREACH_KID_OF_NODE(pNode) {
					pKid = _iterkids.Node<TedNode>();
					mRefs[pKid] +=1;
				}
			}
		}
		for(mp = mRefs.begin(); mp != mRefs.end(); mp++) {
			if(mp->second <= 1) {
				continue;
			} else if(mp->first == TedNode::getOne()|| mp->first->isBasic()) {
				continue;
			} else {
				ret = ret + mp->second -1;
			}
		}
		return ret;
	}

	void TedMan::getParentNodes(TedNode* candidate, vector<TedNode*> &vparentnodes) {
		TedNode* pNode = NULL,*pKid = NULL;
		//unsigned int index = 0;
		TedContainer::iterator msp;
		set <TedNode*>::iterator sp;

		for(msp = _container.begin(); msp != _container.end(); msp ++) {
			for(sp = msp->second.second->begin(); sp != msp->second.second->end(); sp++) {
				pNode =*sp;
				FOREACH_KID_OF_NODE(pNode) {
					pKid = _iterkids.Node<TedNode>();
					if(pKid==candidate) {
						vparentnodes.push_back(pNode);
					}

				}
			}
		}
	}

	void TedMan::searchDecompNode(vector<pair<TedNode*, TedNode*> > & list_decomp_node) {
		TedNode* pNode = NULL,* pKid = NULL;
		wedge weight = 0;
		unsigned int index = 0;
		TedContainer::iterator msp;
		set <TedNode*>::iterator sp;
		map <TedNode*, vector<pair <TedNode*, pair< int, int> > > > linkofnodes;
		map <TedNode*, vector<pair <TedNode*, pair< int, int> > > >::iterator it_linkofnodes;
		vector<pair< TedNode*, pair<int,int> > >::iterator it_parent;
		vector<pair<TedNode*, TedNode*> > it_vector;
		list_decomp_node.clear();	 //nettoyage de la liste
		/*creation d'une liste de noeud avec enfants pour references*/
		for(msp = _container.begin(); msp != _container.end(); msp ++) {
			for(sp = msp->second.second->begin(); sp != msp->second.second->end(); sp++) {
				pNode =*sp;
				//pour chacun des enfants je met le noeud et dans la liste associs ses enfants
				FOREACH_KID_OF_NODE(pNode) {
					index = _iterkids.getIndex();
					pKid = _iterkids.Node<TedNode>();
					weight = _iterkids.getWeight();
					linkofnodes[pKid].push_back(pair<TedNode*, pair<int,int> >(pNode, pair<int,int>(index,weight)));
				}
			}
		}
		//	it_vector=arg_candidate_list.begin();
		/* pour chacun des noeud de  la liste*/
		string name_first_child;
		bool samename;
		for(it_linkofnodes=linkofnodes.begin();it_linkofnodes!=linkofnodes.end();it_linkofnodes++) {
			if(it_linkofnodes->first==TedNode::getOne())
				continue;
			if(it_linkofnodes->second.size()<2)
				continue;
			samename=true;
			name_first_child=it_linkofnodes->second.begin()->first->getName();
			//lister les parents
			for(it_parent=it_linkofnodes->second.begin();it_parent!=it_linkofnodes->second.end();it_parent++) {
				if(name_first_child!=it_parent->first->getName()) {
					samename=false;
					break;
				}
			}
			if(samename==false) {
				cout<<"------------------------------------------"<<endl;
				cout<<it_linkofnodes->first->getName()<<" "<<it_linkofnodes->first<<" is referenced by"<<endl;

				for(it_parent=it_linkofnodes->second.begin();it_parent!=it_linkofnodes->second.end();it_parent++) {
					cout<<it_parent->first->getName()<<" "<< it_parent->first<<" with a "<<it_parent->second.first<<"-edge"<<endl;
				}
			}
		}
		pNode=NULL;
	}

	void TedMan::collectParents(void) {
		TedParents parents;
		preOrderDFS(parents,&TedParents::collect);
		parents.numParents(TedNode::getOne());
	}

	void TedMan::computeBitwidth(void) {
		TedBitwidth tedwidth(this);
		tedwidth.compute();

	}

	Bitwidth* TedMan::getMaxBitwidth(void) {
		Bitwidth* retval = NULL;
		for(PrimaryOutputs::iterator it = _pos.begin(); it!= _pos.end(); it++) {
			Bitwidth* prec = it->second.getBitwidth();
			if(!retval) {
				retval = prec->clone();
			} else if(retval->isLessThan(*prec)) {
				retval->copy(*prec);
			}
		}
		return retval;
	}

	void TedMan::moveSupportTo(TedMan* destiny) {
		//O(n+n.pt+n.st)
		Mark<TedNode>::newMark();
		Mark<TedVar>::newMark();
		vector<TedNode*> vNodes;
		vector<string> terms;
		for(PrimaryOutputs::iterator it = _pos.begin(); it!=_pos.end(); it++) {
			if(it->second.getType() ==TedNodeRoot::PT || it->second.getType() ==TedNodeRoot::ST) {
				it->second.Node()->_collectDFS_rec(vNodes);
				destiny->linkTedToPO(it->second,it->first);
				terms.push_back(it->first);
			}
		}
		//unlink the POs from PT or ST
		while(!terms.empty()) {
			_pos.erase(terms.back());
			terms.pop_back();
		}
		BinaryThreadedTree<int,TedVar> bt;
		for(TedContainer::iterator it = _container.begin(); it != _container.end(); it++) {
			TedVar* var = _vars.getIfExist(it->first.getName());
			if(var && var->visited.isMarked())
				bt.insert(-1*it->second.first, it->first);
		}
		BinaryThreadedTree<int,TedVar>::inorder_iterator iot = bt.begin();
		BinaryThreadedTree<int,TedVar>::inorder_iterator iot_end = bt.end();
		for(;iot!=iot_end;++iot) {
			destiny->_container.registerVarAtTop((*iot)->getData());
			TEDSet* nodes = _container[(*iot)->getData()].second;
			TEDSet tempset =*nodes;
			for(TEDSet::iterator jt = tempset.begin(); jt != tempset.end(); jt++) {
				if((*jt)->visited.isMarked()) {
					//if a node is shared between the root TED and its decomposed TED, this will
					//remove the node from the root TED container. Any method that traverses the root TED
					//container will work, but the node won't be physically registered there.
					destiny->_container[(*iot)->getData()].second->insert(*jt);
					nodes->erase(*jt);
				}
			}
#define INMIDIATE_REMOVE_EMPTY_LEVELS 1
#ifdef INMIDIATE_REMOVE_EMPTY_LEVELS
			if(nodes->empty()) {
				_container.removeLevel((*iot)->getData());
			}
#endif
		}
#ifdef NODE_BELONGS_TO_2_CONTAINERS
		TEDSet missing(_container.all_missing_children());
		for (TEDSet::iterator jt = missing.begin(); jt != missing.end(); jt++) {
			TedNode* node = *jt;
			assert(node);
			ETedNode* enode = dynamic_cast<ETedNode*>(node);
			assert(enode);
			_container.getNode(enode);
		}
#endif
#ifndef INMIDIATE_REMOVE_EMPTY_LEVELS
		_container.remove_empty_levels();
#endif
	}

	TedMan* TedMan::decomposeAll(bool force) {
		TedCompareCost* compare = new MinLatency();
		TedMan* pTedMan = decomposeAll(compare,force);
		delete compare;
		return pTedMan;
	}

	TedMan* TedMan::decomposeAll(TedCompareCost* compare,bool force) {
		TedMan* pTedMan = duplicate();
		bool continueDecomposing = !pTedMan->decompose(force);
		size_t watch_dog = pTedMan->_container.nodeCount();
		while(continueDecomposing && watch_dog-- > 0) {
			TedOrderStrategy* orderst = new TedOrderProper(pTedMan);
			set<TedVar> window(orderst->getStridePO());
			TedMan* newTedMan = orderst->permute(compare,&window);
			delete orderst;
			if(newTedMan!=pTedMan) {
				delete pTedMan;
				pTedMan = newTedMan;
			}
			continueDecomposing = !pTedMan->decompose(force);
		}
		return pTedMan;
	}

	TedMan* TedMan::decomposeAllwithSupport(TedCompareCost* compare,bool force) {
		TedMan* pTedMan = duplicate();
		TedMan* support = new TedMan();
		bool continueDecomposing = !pTedMan->decompose(force);
		size_t watch_dog = pTedMan->_container.nodeCount();
		while(continueDecomposing && watch_dog-- > 0) {
			pTedMan->moveSupportTo(support);
			TedOrderStrategy* orderst = new TedOrderProper(pTedMan);
			TedMan* newTedMan = pTedMan->permuteOrdering(orderst, compare,true);
			//TedMan* newTedMan = pTedMan->siftAll(orderst, compare);
			delete orderst;
			if(newTedMan!=pTedMan) {
				delete pTedMan;
				pTedMan = newTedMan;
			}
			continueDecomposing = !pTedMan->decompose(force);
		}
		support->moveSupportTo(pTedMan);
		delete support;
		return pTedMan;
	}


	bool TedMan::verify(const string& output1, const string& output2) {
		PrimaryOutputs::iterator out1 = _pos.find(output1);
		PrimaryOutputs::iterator out2 = _pos.find(output2);
		return (out1 != _pos.end() && out2 != _pos.end() && out1->second.Node() == out2->second.Node());
	}

	bool TedMan::regressionTest(stringstream& message) {
		bool ret = true;
		int passed = 0;
		int failed = 0;
		stringstream annotateFail;
		stringstream annotatePass;
		for(PrimaryOutputs::iterator it = _pos.begin(); it != _pos.end(); it++) {
			if((it->second).getWeight()!=0) {
				failed++;
				ret = false;
				annotateFail << ":" << it->first;
			} else {
				passed++;
				annotatePass << ":" << it->first;
			}
		}
		int total = passed+failed;
		if(total==passed) {
			message << ",all(" << passed << "/" << total << "),none,";
		} else if(total==failed) {
			message << ",none,all(" << failed << "/" << total << "),";
		} else {
			message << "," << passed << "/" << total << " " << annotatePass.str()<< ",";
			message << "," << failed << "/" << total << " " << annotateFail.str()<< ",";
		}
		return ret;
	}

	void TedMan::foreach(void(TedVar::*functor)(void)const) {
		TedContainerOrder tnodesOrdered = _container.order();
		for(TedContainerOrder::const_iterator it = tnodesOrdered.begin(); it!= tnodesOrdered.end(); it++) {
			TedVar* pvar  = checkVariables(it->second.first->getName());
			(pvar->*functor)();
		}
	}

	void TedMan::foreach(void(TedNode::*functor)(void)) {
		for(TedContainer::iterator p = _container.begin(); p != _container.end(); p++) {
			for(TEDSet::iterator sp = p->second.second->begin(); sp != p->second.second->end(); sp++) {
				((*sp)->*functor)();
			}
		}
	}

	void TedMan::listOutputs(void) {
		for(PrimaryOutputs::iterator p = _pos.begin(); p != _pos.end(); p++) {
			cout << p->first;
			switch(p->second.getType()) {
			case TedNodeRoot::PO:
				cout << " PrimaryOutput ";
				break;
			case TedNodeRoot::PT:
				cout << " ProductTerm ";
				break;
			case TedNodeRoot::ST:
				cout << " SumTerm ";
				break;
			case TedNodeRoot::BC:
				cout << " BottomCut ";
				break;
			case TedNodeRoot::TC:
				cout << " TopCut ";
				break;
			case TedNodeRoot::EV:
				cout << " ExpectedValue ";
				break;
			case TedNodeRoot::ET:
				cout << " ErrorTED ";
				break;
			default:
				cout << " Unkwnown ";
				break;
			}
			//TedNode* rootNode = p->second.Node();
			Mark<TedNode>::newMark();
			vector <TedNode*> vNodes;
			p->second.Node()->_collectDFS_rec(vNodes);
			cout << vNodes.size()<< endl;
		}
	}

	void TedMan::recoverRegisterVars(void) {
		TedParents parents;
		preOrderDFS(parents, &TedParents::collect);
		for(TedContainer::iterator p = _container.begin(); p != _container.end(); p++) {
			TEDSet::iterator sp_last = p->second.second->begin();
			TEDSet::iterator sp = p->second.second->begin();
			while (sp != p->second.second->end()) {
				sp_last = sp;
				sp_last++;
				TedNode* oldnode  = (*sp);
				if (oldnode->getRegister()>0 && !oldnode->getVar()->getBase()) {
					TedVar* rvar = _vars.createRetimedVar(*oldnode->getVar(),oldnode->getRegister());
					TedNode* newnode = _container.getNode(*rvar,*oldnode->getKids(),oldnode->getRegister());
					updateAllParents(oldnode,newnode,parents);
					if (oldnode!=newnode) {
						p->second.second->erase(sp);
						delete oldnode;
					}
				}
				sp = sp_last;
			}
			if (p->second.second->empty()) {
				//empty variable container
			}
		}
	}

#if 0
	//
	// MOVED to TedRetime and then inlined in TedMan.inl
	//
	void TedMan::retimeForward(void) {
		TedContainerOrder tnodesOrdered = _container.order();
		// from bottom register to top register
		TedRetiming retime(this);
		for(TedContainerOrder::reverse_iterator it = tnodesOrdered.rbegin(); it!= tnodesOrdered.rend(); it++) {
			TedVar* pvar  = checkVariables(it->second.first->getName());
			retime.up(*pvar);
		}
	}

	void TedMan::retimeBackward(void) {
		TedContainerOrder tnodesOrdered = _container.order();
		// from bottom register to top register
		TedRetiming retime(this);
		for(TedContainerOrder::iterator it = tnodesOrdered.begin(); it!= tnodesOrdered.end(); it++) {
			TedVar* pvar  = checkVariables(it->second.first->getName());
			retime.down(*pvar);
		}
	}
#endif

	void TedMan::deriveSchedule(void) {
		bool isDecomposed = false;
		for(PrimaryOutputs::iterator p = _pos.begin(); p != _pos.end(); p++) {
			if(p->second.getType() == TedNodeRoot::PT || p->second.getType() == TedNodeRoot::ST) {
				isDecomposed = true;
				break;
			}
		}
		if(!isDecomposed) {
			cout << "TED must be decomposed first in order to derive its schedule" << endl;
			return;
		}
		for(PrimaryOutputs::iterator p = _pos.begin(); p != _pos.end(); p++) {
			if(p->second.getType() == TedNodeRoot::PT || p->second.getType() == TedNodeRoot::ST) {
				isDecomposed = true;
				break;
			}
		}
	}

	void TedMan::purge_transitory_containers(void) {
		Mark<TedNode>::newMark();
		Mark<TedKids>::newMark();
		this->getContainer().update_TMark();
		TedKids::purge_transitory_container();
		ATedNode::purge_transitory_container();
	}

	class AdditiveElement {
	public:
		TedNode* start;
		TedNode* end;
		unsigned int index;
		wedge weight;
		TedRegister reg;
		AdditiveElement(TedNode*s,TedNode*e,unsigned int i, wedge w,TedRegister r) : start(s), end(e), index(i), weight(w), reg(r) {};
		~AdditiveElement(void) {};
	};

	unsigned int TedMan::precompute_critical_path(void) {
		multimap<unsigned int, TedNode*> all_paths(precompute_all_latency_paths());
		return all_paths.rbegin()->first;
	}

	multimap<unsigned int, TedNode*> TedMan::precompute_all_latency_paths(void) {
		typedef multimap<unsigned int,AdditiveElement> AdditiveChain;
		TedParents parents;
		preOrderDFS(parents, &TedParents::collect);
		multimap<unsigned int, TedNode*> paths;
		map<TedNode*,unsigned int> delayMap;
		map<TedNode*, AdditiveChain > delayAdditive;
		map<TedNode*, unsigned int > delayMultiplicative;

		unsigned int delay_mul = tds::Environment::getInt(tds::Environment::DELAYMPY);
		unsigned int delay_add = tds::Environment::getInt(tds::Environment::DELAYADD);
		unsigned int delay_reg = 0;
		//
		// register outputs are assumed to be available as primary inputs therefore
		// its delay is disregarded for the computation of the critical latency.
		// tds::Environment::getInt(tds::Environment::DELAYREG);
		//
		for (dfs_iterator it= dfs_begin(); it != dfs_end(); it++) {
			TedNode* node = (*it);
			if (node==TedNode::getOne()) {
				continue;
			}
			bool connects_to_additive_tree = false;
			bool connects_to_multiplicative_tree = false;
			if ((parents.numParents(node) ==1) && !isPO(node)) {
				MyParent::iterator parent_it = parents[node]->begin();
				unsigned int parent_index = parent_it->second->at(0);
				assert(parent_it->first->hasKid(parent_index));
				if (parent_it->first->getKidRegister(parent_index) ==0) {
					if (parent_index==0) {
						//additive could have any number of kids but parent must have unitary weight
						connects_to_additive_tree = (parent_it->first->getKidWeight(0) ==1 || parent_it->first->getKidWeight(0) ==-1);
					} else if (node->numKids() ==1) {
						//multiplicative could have any parent index other than 0, but must have only one kid
						connects_to_multiplicative_tree = true;
					}
				}
			}
			TedKids* kids = node->getKids();
			AdditiveChain to_be_added;
			for(TedKids::iterator iterkid = kids->begin(), iterend = kids->end(); iterkid != iterend; iterkid++) {
				unsigned int index  = iterkid.getIndex();
				TedNode*	pKid = iterkid.Node<TedNode>();
				wedge weight = iterkid.getWeight();
				TedRegister reg = iterkid.getRegister();
				unsigned int k = (weight!=1 && weight!=-1) ? 1 : 0;
				float tree_level = log10((float)(index+k))/log10((float)2);
				unsigned int latency = ceil(tree_level)*delay_mul;
				if (pKid != TedNode::getOne()) {
					bool retrieve_tree = (reg==0) && !isPO(pKid) && (parents.numParents(pKid) ==1);
					bool retrieve_additive_tree = retrieve_tree && (index==0) && (k==0);
					bool retrieve_multiplicative_tree = retrieve_tree && (index!=0) && (pKid->numKids() ==1);
					if (retrieve_additive_tree) {
						// inserts all additive tree to this node
						// the tree could growth if this node connects
						// to an additive tree or it could be resolved
						// if it does not
						AdditiveChain& refa = delayAdditive[pKid];
						to_be_added.insert(refa.begin(),refa.end());
					} else if (retrieve_multiplicative_tree) {
						//this segments assumes we are resolving the multiplicative tree
						//and no further propagation is needed
						unsigned int prior_latency = delayMap[pKid];
						unsigned int prior_tree_level = delayMultiplicative[pKid];
						prior_tree_level += (index+k);
						tree_level = log10((float)(prior_tree_level))/log10((float)2);
						latency = ceil(tree_level)*delay_mul;
						if (latency<prior_latency) {
							latency=prior_latency+delay_mul;
						} else {
							unsigned int adjustment = (prior_latency!=0 ) ? 1 : 0;
							tree_level = log10((float)(prior_tree_level+adjustment))/log10((float)2);
							latency = ceil(tree_level)*delay_mul;
						}
						AdditiveElement ae(node,pKid,index,weight,reg);
						to_be_added.insert(AdditiveChain::value_type(latency,ae));
					} else {
						unsigned int prior_latency = delayMap[pKid];
						if (reg!=0) {
							paths.insert(pair<unsigned int,TedNode*>(prior_latency,pKid));
							prior_latency = reg*delay_reg;
						}
						if (latency<prior_latency) {
							latency=prior_latency+delay_mul;
						} else {
							unsigned int adjustment = (reg==0 ) ? 1 : 0;
							tree_level = log10((float)(index+k+adjustment+1))/log10((float)2);
							latency = ceil(tree_level)*delay_mul;
						}
						AdditiveElement ae(node,pKid,index,weight,reg);
						to_be_added.insert(AdditiveChain::value_type(latency,ae));
					}
				} else {
					AdditiveElement ae(node,pKid,index,weight,reg);
					to_be_added.insert(AdditiveChain::value_type(latency,ae));
				}
			}
			//
			// individual branches have been resolved
			// now we need to compute the collective delay of all
			// branches if the node does not require preservation
			//
			assert(!to_be_added.empty());
			if (connects_to_additive_tree) {
				assert(!isPO(node));
				delayAdditive.insert(pair<TedNode*, AdditiveChain>(node,to_be_added));
			} else if (connects_to_multiplicative_tree) {
				assert(!isPO(node));
				unsigned int onlyKid_index = node->getKids()->highestOrder();
				TedNode* onlyKid = node->getKidNode(onlyKid_index);
				TedRegister reg = node->getKidRegister(onlyKid_index);
				delayMap[node] = (reg==0) ? delayMap[onlyKid] : reg*delay_reg;
				wedge weight = node->getKidWeight(onlyKid_index);
				unsigned int k = (weight!=1 && weight!=-1) ? 1 : 0;
				unsigned int prior_tree_level = delayMultiplicative[onlyKid];
				prior_tree_level += (onlyKid_index+k);
				delayMultiplicative.insert(pair<TedNode*,unsigned int>(node,prior_tree_level));
			} else {
				unsigned int maxout = to_be_added.rbegin()->first;
				if (to_be_added.size() ==1) {
					delayMap[node] = maxout;
				} else if (maxout<delay_add) {
					float len = log10((float)to_be_added.size())/log10((float)2);
					delayMap[node] = ceil(len)*delay_add;
				} else {
					maxout-=delay_add;
					AdditiveChain::iterator ms1 = to_be_added.begin();
#if 1
					//attempting to resolve cases in which chain of adders
					//are on top of a chain of multipliers
					// need to figure out a way to compute the latency of
					// each component from their equivalent node.
					map<TedNode*,multiset<unsigned int> > regroup;
					while(ms1!=to_be_added.end()) {
						unsigned int k2 = (ms1->second.weight != 1 && ms1->second.weight != -1) ? 1 : 0;
						unsigned int adj = (ms1->second.reg == 0) ? 1: 0;
						unsigned int tree2 = ms1->second.index + k2 + adj;
						float level2 = log10((float)(tree2))/log10((float)2);
						unsigned int latency2 = ceil(level2)*delay_mul;
						if (regroup.find(ms1->second.end) == regroup.end()) {
							multiset<unsigned int> one_tree;
							one_tree.insert(latency2);
							regroup.insert( pair<TedNode*,multiset<unsigned int> >(ms1->second.end,one_tree) );
						} else {
							multiset<unsigned int>& one_tree = regroup[ms1->second.end];
							one_tree.insert(latency2);
						}
						ms1++;
					}
					multiset<unsigned int> final_add;
					for (map<TedNode*,multiset<unsigned int> >::iterator px = regroup.begin(), px_end = regroup.end(); px != px_end; px++) {
						multiset<unsigned int>& pxset = px->second;
						if (pxset.size()>1) {
							multiset<unsigned int>::iterator iter1 = pxset.begin();
							multiset<unsigned int>::iterator iter2 = iter1++;
							while (iter2!=pxset.end()) {
								pxset.erase(iter1);
								unsigned int val = *iter2 + delay_add;
								pxset.erase(iter2);
								pxset.insert(val);
								iter1 = pxset.begin();
								iter2 = iter1;
								iter2++;
							}
							unsigned int new_latency = delay_mul;
							unsigned int prior_latency = delayMap[px->first];
							if (new_latency>prior_latency) {
								new_latency += *(pxset.begin());
							} else {
								new_latency += prior_latency;
							}
							final_add.insert(new_latency);
						} else {
							for(AdditiveChain::iterator pxx = to_be_added.begin(), pxe = to_be_added.end(); pxx!=pxe; pxx++) {
								if (pxx->second.end == px->first) {
									final_add.insert(pxx->first);
									break;
								}
							}
						}
					}
					multiset<unsigned int>::iterator iterA = final_add.begin();
					multiset<unsigned int>::iterator iterB = iterA;
					iterB++;
					while (iterB!=final_add.end()) {
						final_add.erase(iterA);
						unsigned int val = *iterB + delay_add;
						final_add.erase(iterB);
						final_add.insert(val);
						iterA = final_add.begin();
						iterB = iterA;
						iterB++;
					}
					delayMap[node] = *final_add.begin();
#else
					AdditiveChain::iterator ms2 = ms1;
					ms2++;
					while ( ms2!=to_be_added.end() && (ms2->first)<=maxout ) {
						to_be_added.erase(ms1);
						unsigned int val = (ms2->first)+delay_add;
						to_be_added.erase(ms2);
						AdditiveElement ae(NULL,NULL,0,0,0);
						to_be_added.insert(AdditiveChain::value_type(val,ae));
						ms1 = to_be_added.begin();
						ms2 = ms1;
						ms2++;
					}
					float len = log10((float)to_be_added.size())/log10((float)2);
					delayMap[node] = ceil(len)*delay_add+to_be_added.rbegin()->first;
#endif
				}
				if (isPO(node)) {
					paths.insert(pair<unsigned int,TedNode*>(delayMap[node],node));
				}
			}
		}
		return paths;
	}



	void TedMan::siftAll(TedOrderStrategy* order, TedCompareCost* less, TedManCluster* cluster) {
		TedManCluster clusterReClassify;
		list<TedMan*> todel;
		for (TedManCluster::iterator it = cluster->begin(), it_end = cluster->end(); it != it_end; it++) {
			TedMan* test = it->second;
			TedOrderStrategy* orderCluster = new TedOrderProper(test);
			orderCluster->stop_criteria(order->stop_criteria());
			TedMan* test_out = test->siftAll(orderCluster,less);
			Cost* compare = less->compute_cost(*test_out);
			unsigned int final_cost = compare->get_cost();
			clusterReClassify.insert(TedManCluster::value_type(final_cost,test_out));
			delete orderCluster;
			delete compare;
			if (test_out != test) {
				todel.push_back(test);
			}
		}
		cluster->clear();
		cluster->insert(clusterReClassify.begin(),clusterReClassify.end());
		clusterReClassify.clear();
		while (!todel.empty()) {
			delete todel.back();
			todel.pop_back();
		}
	}

	void TedMan::siftGroupedAll(TedOrderStrategy* order, TedCompareCost* less, TedManCluster* cluster) {
		TedManCluster clusterReClassify;
		list<TedMan*> todel;
		for (TedManCluster::iterator it = cluster->begin(), it_end = cluster->end(); it != it_end; it++) {
			TedMan* test = it->second;
			TedOrderStrategy* orderCluster = new TedOrderProper(test);
			orderCluster->stop_criteria(order->stop_criteria());
			TedMan* test_out = test->siftGroupedAll(orderCluster,less);
			Cost* compare = less->compute_cost(*test_out);
			unsigned int final_cost = compare->get_cost();
			clusterReClassify.insert(TedManCluster::value_type(final_cost,test_out));
			delete orderCluster;
			delete compare;
			if (test_out != test) {
				todel.push_back(test);
			}
		}
		cluster->clear();
		cluster->insert(clusterReClassify.begin(),clusterReClassify.end());
		clusterReClassify.clear();
		while (!todel.empty()) {
			delete todel.back();
			todel.pop_back();
		}
	}


	void TedMan::permuteOrdering(TedOrderStrategy* order, TedCompareCost* less, TedManCluster* cluster) {
		TedManCluster clusterReClassify;
		list<TedMan*> todel;
		for (TedManCluster::iterator it = cluster->begin(), it_end = cluster->end(); it != it_end; it++) {
			TedMan* test = it->second;
			TedOrderStrategy* orderCluster = new TedOrderProper(test);
			orderCluster->stop_criteria(order->stop_criteria());
			TedMan* test_out = test->permuteOrdering(orderCluster,less,false);
			Cost* compare = less->compute_cost(*test_out);
			unsigned int final_cost = compare->get_cost();
			clusterReClassify.insert(TedManCluster::value_type(final_cost,test_out));
			delete orderCluster;
			delete compare;
			if (test_out != test) {
				todel.push_back(test);
			}
		}
		cluster->clear();
		cluster->insert(clusterReClassify.begin(),clusterReClassify.end());
		clusterReClassify.clear();
		while (!todel.empty()) {
			delete todel.back();
			todel.pop_back();
		}
	}

	void TedMan::annealing(TedOrderStrategy* order, TedCompareCost* less, TedManCluster* cluster, bool stride_backtrack) {
		TedManCluster clusterReClassify;
		list<TedMan*> todel;
		for (TedManCluster::iterator it = cluster->begin(), it_end = cluster->end(); it != it_end; it++) {
			TedMan* test = it->second;
			TedOrderStrategy* orderCluster = new TedOrderProper(test);
			orderCluster->stop_criteria(order->stop_criteria());
			TedMan* test_out = test->annealing(orderCluster,less,false,stride_backtrack);
			Cost* compare = less->compute_cost(*test_out);
			unsigned int final_cost = compare->get_cost();
			clusterReClassify.insert(TedManCluster::value_type(final_cost,test_out));
			delete orderCluster;
			delete compare;
			if (test_out != test) {
				it->second = NULL;
				todel.push_back(test);
			}
		}
		cluster->clear();
		cluster->insert(clusterReClassify.begin(),clusterReClassify.end());
		clusterReClassify.clear();
		while (!todel.empty()) {
			delete todel.back();
			todel.pop_back();
		}
	}


#ifdef FUNCTIONS_DEPRECATED
	/** @brief Build a Ted From DFG*/
	TedMan* TedMan::buildTedFromDFG(TedMan* pTedMan, DfgMan* pDfgMan) {
		assert(ATedNode::checkContainer(pTedMan->getContainer()));
		map <DfgNode*, ATedNode*> mVisited;
		for(map <string, DfgNode*>::iterator p = pDfgMan->_mPos.begin(); p != pDfgMan->_mPos.end(); p++) {
			ATedNode* pANode = _BuildTedFromDfg_Rec(p->second, mVisited, pTedMan);
			TedNodeRoot pin(pTedMan->normalize(pANode));
			pTedMan->linkTedToPO(pin.Node(), pin.getWeight(), p->first);
		}
		return pTedMan;
	}

	ATedNode* _BuildTedFromDfg_Rec(DfgNode* pDfgNode, map<DfgNode*,ATedNode*> & mVisited, TedMan* pTedMan) {
		ATedNode* pANode = NULL,* pELeft=NULL,* pERight=NULL;
		if(mVisited.find(pDfgNode)!= mVisited.end()) {
			return mVisited[pDfgNode]->duplicateRecursive();
		}
		if(pDfgNode->isConst()) {
			return new ETedNode(pDfgNode->getConst());
		} else if(pDfgNode->isPI()|| pDfgNode->isVarConst()) {
			//not needed
			//TedVar* pvar = NULL;
			//	//pvar = pTedMan->checkVariables(pDfgNode->getName());			   -----\
			//pvar = TedVarMan::instance().createVar(TedVar(pDfgNode->getName()));	 <--/
#ifdef CAUSE_BUG_102
			if(pvar->isOwner()) {
				return new ETedNode(dynamic_cast<TedVarGroupOwner*>(pvar)->getMember()->getName());
			}
#endif
			return new ETedNode(pDfgNode->getName());
		} else {
			assert(pDfgNode->isOp());
		}
		pELeft = _BuildTedFromDfg_Rec(pDfgNode->getLeft(), mVisited, pTedMan);
		pERight= _BuildTedFromDfg_Rec(pDfgNode->getRight(), mVisited, pTedMan);
		if(pDfgNode->isAdd()) {
			pANode = pELeft->add(pERight);
		} else if(pDfgNode->isSub()) {
			// I believe this would have require duplication
			pANode = pELeft->sub(pERight);
		} else if(pDfgNode->isMul()) {
			pANode = pELeft->mul(pERight);
		} else if(pDfgNode->isLSH()) {
			assert(pERight->isConst());
			int val = 2;
			int exponent = pERight->getConst();
			for(int i = 1; i < exponent; i++) {
				val*= 2;
			}
			pERight->setConst(val);
			pANode = pELeft->mul(pERight);
		}
		return pANode;
	}

	void  _UpdateLowerUpperBoundSet(TedMan* pMan, multiset <_UpLowBound>* sTeds, int* minUpb) {
		int upb, lowb;
		_UpLowBound pin;
		multiset <_UpLowBound>::iterator ppin;
		pMan->getUpLowMulBound(&upb, &lowb);
		if(lowb >=*minUpb) {
			return;
		} else if(upb >=*minUpb) {
			pin.lowb=lowb;
			pin.upb = upb;
			pin.pTedMan = pMan->duplicate();
			sTeds->insert(pin);
		} else {
			*minUpb = upb;
			pin.lowb = upb;
			ppin = sTeds->insert(pin);
			sTeds->erase(ppin, sTeds->end());
			pin.lowb=lowb;
			pin.upb = upb;
			pin.pTedMan = pMan->duplicate();
			sTeds->insert(pin);
		}
	}
#endif

}
