#include "TedMan.h"
#include "TedContainer.h"
#include "TedVar.h"
#include "TedNode.h"
#include "ETedNode.h"

#include <vector>
#include <set>
#include <bitset>

namespace ted {

	vector< set<TedVar> > TedMan::getStrides(bool create_clusters,unsigned int bottom_window_index, unsigned int top_window_index) {
		if (top_window_index<=bottom_window_index) {
			top_window_index = _container.getMaxLevel();
		}
		if (create_clusters) {
			return clusterPOsByVarSupport(bottom_window_index,top_window_index);
		} else {
			set<TedVar> all_vars;
			for (TedContainer::iterator it = _container.begin(), end = _container.end(); it != end; it++) {
				// ENFORCE WINDOW SIZE
				unsigned int varLevel = _container.getLevel(it->first);
				if (varLevel >= bottom_window_index && varLevel <= top_window_index) {
					all_vars.insert(it->first);
				}
			}
			vector< set<TedVar> > one_stride;
			one_stride.push_back(all_vars);
			return one_stride;
		}
	}

	//Traverses all POs and collects for each PO the support variables
	vector< set<TedVar> > TedMan::clusterPOsByVarSupport(unsigned int bottom_window_index, unsigned int top_window_index) {
		if (top_window_index<=bottom_window_index) {
			top_window_index = _container.getMaxLevel();
		}
		map<const TedNode*, set< TedVar > > varlist;
		const int var_size = 1024*3;
		const int max_level = _container.getMaxLevel();
		assert(max_level<var_size);
		vector < bitset<var_size> > vb;
		vector < set<TedVar> > vs;

		for(TedMan::PrimaryOutputs::iterator p = getPOs().begin(); p != getPOs().end(); p++) {
			const TedNode* pNode = p->second.Node();
			getNodeSupport(pNode,varlist,bottom_window_index,top_window_index);
			bitset<var_size> bs;
			set<TedVar> st = varlist[pNode];
			for(set<TedVar>::iterator it = st.begin(), end = st.end(); it != end; it++) {
				unsigned int level = _container.getLevel(*it);
				bs.set(level);
			}
			if(bs.count() == (max_level-1)) {
				//there is one PO that uses all variables
				//level 0 correspond to variable ONE which is never used
				vb.clear();
				vs.clear();
				vs.push_back(st);
				return vs;
			}
			bool append = true;
			bool found = false;
			if(!vb.empty()) {
				vector< bitset<var_size> >::iterator it = vb.begin();
				vector< bitset<var_size> >::iterator it_end = vb.end();
				vector< set<TedVar> >::iterator jt = vs.begin();
				vector< set<TedVar> >::iterator jt_end = vs.end();
				for(int index=0; it != it_end; it++,jt++,index++) {
					bitset<var_size> test =*it;
					if(test == bs) {
						append = false;
						if (found) {
							// a previous equal stride was found
							// current iterator is a duplicate
							vb.erase(it);
							vs.erase(jt);
						}
						found = true;
						//					break;
					} else {
						bitset<var_size> discriminant = (test ^ bs);
						bitset<var_size> test_contains_bs = (discriminant & test);
						bitset<var_size> bs_contains_test = (discriminant & bs);
						if(test_contains_bs == discriminant) {
							// test contains bs
							append = false;
							if (found) {
								vb.erase(it);
								vs.erase(jt);
							} else {
								//continue the search for duplicates
								//with the most complete stride
								bs =*it;
								st =*jt;
							}
							found = true;
							//						break;
						} else if(bs_contains_test == discriminant) {
							//bs contains test
							append = false;
							if (found) {
								vb.erase(it);
								vs.erase(jt);
							} else {
								//replace in place
								vb[index] = bs;
								vs[index] = st;
							}
							//break;
						}
					}

				}
			}
			if(append) {
				vb.push_back(bs);
				vs.push_back(st);
			}
		}
		for(int i = 0; i <vb.size(); i++) {
			string strade = vb[i].to_string();
			cout << "strade="<< strade.substr(var_size-max_level-1)<< endl;
		}
		return vs;
	}

	//recursively traverse a node annotating its support
	void TedMan::getNodeSupport(const TedNode* pNode, map<const TedNode*,set<TedVar> >& varlist, unsigned int bottom_window_index, unsigned int top_window_index) {
		if(varlist.find(pNode)!= varlist.end()|| pNode == TedNode::getOne()) {
			return;
		}
		set<TedVar> newVars;
		// ENFORCE WINDOW SIZE
		unsigned int level = _container.getLevel(*pNode->getVar());
		if (level >= bottom_window_index && level <= top_window_index) {
			newVars.insert(*pNode->getVar());
		}
		//unsigned int index=0;
		TedNode* pKid=NULL;
		//int weight=0;
		FOREACH_KID_OF_NODE(pNode) {
			pKid = _iterkids.Node<TedNode>();
			getNodeSupport(pKid,varlist,bottom_window_index,top_window_index);
			newVars.insert(varlist[pKid].begin(), varlist[pKid].end());
		}
		varlist.insert(pair<const TedNode*,set<TedVar> >(pNode,newVars));
	}

	//Traverses all POs and collects for each PO the support variables
	set<TedVar> TedMan::getPOSupport(unsigned int bottom_window_index, unsigned int top_window_index) {
		if (top_window_index<=bottom_window_index) {
			top_window_index = _container.getMaxLevel();
		}
		map<const TedNode*, set< TedVar > > varlist;
		const int var_size = 1024;
		const int max_level = _container.getMaxLevel();
		assert(max_level<var_size);
		set<TedVar> po_vars;
		for(TedMan::PrimaryOutputs::iterator p = getPOs().begin(); p != getPOs().end(); p++) {
			if(p->second.getType() !=TedNodeRoot::PO) {
				continue;
			}
			const TedNode* pNode = p->second.Node();
			getNodeSupport(pNode,varlist,bottom_window_index,top_window_index);
			set<TedVar>& st = varlist[pNode];
			po_vars.insert(st.begin(),st.end());
		}
		return po_vars;
	}

	void TedMan::generateClusters(void) {
		if (clusterA) return;
		if (clusterD) return;
		
		multimap<unsigned int,TedNode*> paths(precompute_all_latency_paths());
		unsigned int critical_path = paths.rbegin()->first;
		unsigned int critical_threshold = (critical_path*75)/100;
		for (multimap<unsigned int,TedNode*>::iterator it = paths.begin(), it_end=paths.end(); it != it_end; it++) {
			TedNode* node = it->second;
			TedNodeRoot* root = getPO(node);
			if (root) {
				string po_name;
				for (PrimaryOutputs::iterator jt = _pos.begin(), jt_end = _pos.end(); jt != jt_end; jt++) {
					if ( &(jt->second) == root) {
						po_name = jt->first;
						break;
					}
				}
				assert(!po_name.empty());
				if (it->first >= critical_threshold) {
					if (!clusterD) {
						clusterD = new TedManCluster;
					}
					moveToCluster(po_name,clusterD);
				} else {
					if (!clusterA) {
						clusterA = new TedManCluster;
					}
					moveToCluster(po_name,clusterA);
				}
			}
		}
	}

	void TedMan::moveToCluster(string poName, TedManCluster* cluster) {
		assert(cluster);
		PrimaryOutputs::iterator it = _pos.find(poName);
		if(it != _pos.end()) {
			TedMan* toCluster = new TedMan();
			toCluster->_container.instance_of(this->_container);
			map <TedNode*, TedNode*> mVisited;
			TedNodeRoot& root = it->second;
			TedNode* top = moveToCluster(toCluster,root.Node(),mVisited);
			TedNodeRoot poCluster(top,root.getWeight(),root.getType());
			toCluster->linkTedToPO(poCluster,poName);
			toCluster->_container.remove_empty_levels();
			toCluster->purge_transitory_containers();
			assert(toCluster->check_for_dangling_nodes());
			assert(toCluster->getContainer().check_children_are_included());
			cluster->insert(TedManCluster::value_type(UINT_MAX,toCluster));
		}
	}

	// modified from _Duplicate_Rec
	TedNode* TedMan::moveToCluster(TedMan* pManNew, TedNode* pNode, map <TedNode*, TedNode*> & mVisited) {
		map <TedNode*, TedNode*>::iterator mp;
		TedNode* pNodeNew = NULL,* pKid = NULL,*pKidNew = NULL;
		unsigned int index =0;
		wedge weight = 0;
		TedKids tKids;
		if((mp=mVisited.find(pNode))!= mVisited.end())
			return mp->second;
		if(pNode == TedNode::getOne())
			return pNode;
		if(pNode->isConst())
			return(new ETedNode(pNode->getConst()));
		FOREACH_KID_OF_NODE(pNode) {
			index = _iterkids.getIndex();
			pKid = _iterkids.Node<TedNode>();
			weight = _iterkids.getWeight();
			pKidNew = moveToCluster(pManNew, pKid, mVisited);
			TedRegister kidReg = _iterkids.getRegister();
			tKids.insert(index, pKidNew, weight,kidReg);
		}
		pNodeNew = new ETedNode(*pNode->getVar(), tKids, pNode->getRegister());
		pManNew->_container.registerVarAtTop(*pNode->getVar());
		pManNew->_container[*(pNode->getVar())].second->insert(pNodeNew);
		mVisited[pNode] = pNodeNew;
		return pNodeNew;
	}

}