/*
 * =====================================================================================
 *
 *        Filename:  TEDManWriteDot.cc
 *     Description:  Tedman class
 *         Created:  04/17/2007 12:24:18 PM EDT
 *          Author:  Daniel Gomez-Prado
 *         Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 213                                          $: Revision of last commit
 *  $Author:: daniel@win                                     $: Author of last commit
 *  $Date:: 2010-11-18 08:13:53 -0500(Thu, 18 Nov 2010)$: Date of last commit
 * =====================================================================================
 */

#include <iostream>
#include <sstream>
#include <fstream>
#include <cassert>
#include <algorithm>
#include <list>

#include "TedMan.h"
#include "TedNode.h"
#include "TedKids.h"
#include "TedVar.h"
#include "ETedNode.h"
#include "TedVarGroup.h"
#include "TedVarMan.h"
#include "TedNodeIterator.h"
#include "TedParents.h"
#include "TedDecompose.h"
#include "TedBitwidth.h"

#include "Bitwidth.h"
using namespace data;

#include "BinaryThreadedTree.h"
using dtl::BinaryThreadedTree;
using dtl::BinaryThreadedNode;

#include "Environment.h"
using namespace tds;

#include "DfgNode.h"
#include "DfgMan.h"
using dfg::DfgMan;
using dfg::DfgNode;
using dfg::DfgStat;

#include "util.h"
using util::Util;

#include "Convert.h"
using convert::Convert;

namespace ted {


	/** @brief Show Dot File*/
	void TedMan::writeDot(TedManCluster* cluster, ofstream& ofile, bool bigFont, bool compact_graph) {
		string poFontSize = "";
		string oneFontSize = "";
		if(bigFont) {
			poFontSize = "fixedsize=\"true\", fontsize=14, ";
			oneFontSize = "fixedsize=\"true\", fontsize=12, ";
		}
		ofile << "Digraph G {\n\tsize=\"10, 7.5\"; center = \"true\"; \n";
		ofile << "\t{\n";
		ofile << "\t\tnode [shape = plaintext];\n";
		ofile << "\t\tedge [style = invis];\n";
		ofile << "\t\tLevelPO [label=\"\"];\n";
		ofile << "\n\t}\n";

		for(TedManCluster::iterator it = cluster->begin(), it_end = cluster->end(); it != it_end; it++) {
			TedMan* current = it->second;
			PrimaryOutputs::iterator p = current->_pos.begin();
			ofile << "// ouput from cluster\n";
			ofile << "\t{rank = same; LevelPO; Node_" << p->first;
			ofile << " [" << poFontSize << "label=\""<<p->first<<"\", shape=invtriangle]; }\n";
			ofile << "\tNode_"<<p->first << " -> \"Node_" << p->second.Node()->obfuscate()<< "\" [label=\"";
			if(p->second.getWeight()!= 1)
				ofile << p->second.getWeight();
			if(p->second.getBitwidth())
				ofile << "Q("<<p->second.getBitwidth()->at() << ")";
			if(p->second.getRegister()!=0) {
				ofile << "\\n---" << p->second.getRegister()<< "R---";
			}
			ofile << "\"]\n";			
			for(TedContainer::iterator ps = current->_container.begin(); ps != current->_container.end(); ps++) {
				for(TEDSet::iterator sp = ps->second.second->begin(); sp != ps->second.second->end(); sp++) {
					ofile << (*sp)->getDotStringSimple(current->getLevel(*sp),bigFont,compact_graph);
				}
			}
		}
		ofile << "\n\t{rank = same; Level0; \"Node_" << TedNode::getOne()->obfuscate()<< "\"";
		ofile << " [" << oneFontSize << "label=\"ONE\", shape=triangle]; }\n";
		ofile << "}";
	}


	/** @brief Show Dot File*/
	void TedMan::writeDot(ofstream& ofile, bool bigFont, bool compact_graph) {
		string poFontSize = "";
		string oneFontSize = "";
		if(bigFont) {
			poFontSize = "fixedsize=\"true\", fontsize=14, ";
			oneFontSize = "fixedsize=\"true\", fontsize=12, ";
		}
		ofile << "Digraph G {\n\tsize=\"10, 7.5\"; center = \"true\"; \n";
		ofile << "\t{\n";
		ofile << "\t\tnode [shape = plaintext];\n";
		ofile << "\t\tedge [style = invis];\n";
		ofile << "\t\tLevelPO [label=\"\"];\n";
		unsigned int maxLevel = _container.getMaxLevel();
		ofile << "\t\tLevel0 [label=\"\"];\n";
		for(unsigned int i = 1; i <= maxLevel; i++) {
			TEDSet* used = _container[_container.getVarAtLevel(i)].second;
			if(used && !used->empty()) {
				ofile << "\t\tLevel" << i << " [label=\"";
				if(Environment::getBool("show_level"))
					ofile <<  i << ":" << _container.getVarAtLevel(i).getName();
				if(_vars.getIfExist(_container.getVarAtLevel(i).getName())->getBitwidth())
					ofile << " Q("<< _vars.getIfExist(_container.getVarAtLevel(i).getName())->getBitwidth()->at() << ")";
				ofile << "\"];\n";
			}
		}
		ofile << "\t\tLevelPO";
		for(unsigned int i = maxLevel; i > 0; i--) {
			TEDSet* used = _container[_container.getVarAtLevel(i)].second;
			if(used && !used->empty()) {
				ofile << "-> Level" << i;
			}
		}
		ofile << "\n\t}\n";
		for(PrimaryOutputs::iterator p = _pos.begin(); p != _pos.end(); p++) {
			ofile << "\t{rank = same; LevelPO; Node_" << p->first;
			ofile << " [" << poFontSize << "label=\""<<p->first<<"\", shape=invtriangle]; }\n";
			ofile << "\tNode_"<<p->first << " -> \"Node_" << p->second.Node()->obfuscate()<< "\" [label=\"";
			if(p->second.getWeight()!= 1)
				ofile << p->second.getWeight();
			if(p->second.getBitwidth())
				ofile << "Q("<<p->second.getBitwidth()->at() << ")";
			if(p->second.getRegister()!=0) {
				ofile << "\\n---" << p->second.getRegister()<< "R---";
			}
			ofile << "\"]\n";
		}
		for(TedContainer::iterator p = _container.begin(); p != _container.end(); p++) {
			for(TEDSet::iterator sp = p->second.second->begin(); sp != p->second.second->end(); sp++) {
				ofile << (*sp)->getDotStringSimple(getLevel(*sp),bigFont,compact_graph);
			}
		}
		ofile << "\n\t{rank = same; Level0; \"Node_" << TedNode::getOne()->obfuscate()<< "\"";
		ofile << " [" << oneFontSize << "label=\"ONE\", shape=triangle]; }\n";
		ofile << "}";
	}

#if 0
	/** @brief The dot string of the node*/
	string TedMan::dotString(const TedNode* pNode, bool bigFont, bool compact_graph)const {
		string nodeFontSize = "";
		if(bigFont) {
			nodeFontSize = "fixedsize=\"true\", fontsize=18, ";
		}
		ostringstream oss;
		unsigned int index=0;
		TedNode* pKid=NULL;
		wedge weight=0;
		if(pNode->getRegister()> 0 || pNode->getBitwidth()) {
			oss << "\n\tsubgraph \"reg_" << obfuscate(pNode)<< "\" {";
			oss << "\n\trank = same; Level" << getLevel(pNode)<<"; \"node_reg_" << obfuscate(pNode)<< "\" [label=\"";
			if(pNode->getRegister()> 0)
				oss << pNode->getRegister()<< "R ";
			if(pNode->getBitwidth())
				oss << "b=" << pNode->getBitwidth()->at();
			oss << "\", shape = plaintext]";
		}
		oss << "\n\t{rank = same; Level" << getLevel(pNode)<< "; \"Node_" << obfuscate(pNode)<< "\" [";
		oss << nodeFontSize << "label=\"";
		if(compact_graph && pNode->getName().length()> 4)
			oss << " ";
		else
			oss << pNode->getName();
		oss << "\", ";
		oss	<< "shape=" <<(pNode->isConst() ?"triangle":"ellipse")<< "]; }\n";
		if(pNode->getRegister()> 0 || pNode->getBitwidth()) {
			oss << "\n\t\"Node_" << obfuscate(pNode)<< "\" -> \"node_reg_" << obfuscate(pNode)<< "\" [arrowhead=inv, minlen=0]";
			oss << "\n\t}";
		}
		FOREACH_KID_OF_NODE(pNode) {
			index = _iterkids.getIndex();
			pKid = _iterkids.Node<TedNode>();
			weight = _iterkids.getWeight();
			oss << "\t\"Node_"<<obfuscate(pNode)<<"\" -> \"Node_"<<obfuscate(pKid)<<"\" [style="<<(index==0?"dashed":"solid");
			oss	<< ", label=\"";
			if(index>1) {
				oss << "^"<<index<<" ";
			}
			if(weight!=1) {
				oss << weight;
			}
			if(_iterkids->second.getRegister()!=0) {
				oss << "\\n---" << _iterkids->second.getRegister()<< "R---";
			}
			if(_iterkids->second.getBitwidth()) {
				oss << "b=" << _iterkids->second.getBitwidth()->at();
			}
			oss << "\"];\n";
		}
		return oss.str();
	}

	string TedMan::obfuscate(const TedNode* ptr)const {
		ostringstream oss;
		string name(ptr->getName());
		/* size_t found = name.find_first_of('.');
		   while(found!=string::npos) {
		   name.replace(found,1,"\\.");
		   found = name.find_first_of('.',found+3);
		   }*/
#ifdef _DEBUG
		oss << name << "_" << ptr;
#else
		const size_t distance = sizeof(TedNode);
		const size_t val = ((size_t)ptr)/distance;
		oss << name << "_" << hex << val;
#endif
		return oss.str();
	}
#endif

	void TedMan::visualize(void) {
		visualize(NULL);
	}

	void TedMan::visualize(const TedNode* pTouch) {
		static unsigned int index = 0;
		ofstream ofile;
		string filename = Environment::getStr("show_directory")+ "debugTed" + Util::itoa(index++)+ ".dot";
		ofile.open(filename.c_str());
		if(!ofile.is_open()) {
			throw(string("04014. Cannot open temporary file for writing.\n"));
		}
		writeDotDetailed(ofile,pTouch);
		ofile.close();
		Util::showDotFile(filename.c_str());
	}

	void TedMan::writeDotDetailed(ofstream& ofile) {
		writeDotDetailed(ofile,NULL);
	}

	/** @brief Show Dot File*/
	void TedMan::writeDotDetailed(ofstream& ofile, const TedNode* pTouch) {
		ofile << "Digraph G {\n\tsize=\"10, 7.5\"; center = \"true\"; \n";
		ofile << "\t{\n";
		ofile << "\t\tnode [shape = plaintext];\n";
		ofile << "\t\tedge [style = invis];\n";
		ofile << "\t\tLevelPO [label=\"\"];\n";
		unsigned int maxLevel = _container.getMaxLevel();
		for(unsigned int i = 0; i <= maxLevel; i++) {
			ofile << "\t\tLevel" << i << " [label=\""<< i <<"\"];\n";
		}
		ofile << "\t\tLevelPO";
		for(unsigned int i = 0; i <= maxLevel; i++) {
			ofile << "-> Level" << maxLevel-i;
		}
		ofile << "\n\t}\n";
		for(PrimaryOutputs::iterator p = _pos.begin(); p != _pos.end(); p++) {
			ofile << "\t{rank = same; LevelPO; Node_" << p->first;
			ofile << " [shape=record, label=\"{ " << p->first << " | " << &(p->second)<< "} | {";
			switch(p->second.getType()) {
			case TedNodeRoot::PO :
				ofile << "PO";
				break;
			case TedNodeRoot::PT :
				ofile << "PT";
				break;
			case TedNodeRoot::ST :
				ofile << "ST";
				break;
			case TedNodeRoot::TC :
				ofile << "TC";
				break;
			case TedNodeRoot::BC :
				ofile << "BC";
				break;
			default :
				ofile << "err";
				break;
			}
			ofile << " | r=" << p->second.getRegister()<< " | w=" << p->second.getWeight();
			if (p->second.getBitwidth()) {
				ofile << " | Q(" << p->second.getBitwidth()->at() << ")";
			}
			ofile << "}\"]; }\n";
			ofile << "\tNode_"<<p->first << " -> Node_" << p->second.Node()<< "[label=\"";
			ofile << "\"]\n";
		}
		for(TedContainer::iterator p = _container.begin(); p != _container.end(); p++) {
			for(TEDSet::iterator sp = p->second.second->begin(); sp != p->second.second->end(); sp++) {
				ofile << (*sp)->getDotStringDetailed(getLevel(*sp),pTouch);
			}
		}
		ofile << "\n\t{rank = same; Level0; Node_" << TedNode::getOne();
		ofile << " [label=\"ONE\", shape=triangle]; }\n";
		ofile << "}";
	}

#if 0
	/** @brief The dot string of the node*/
	string TedMan::dotStringDetailed(const TedNode* pNode, const TedNode* pTouch) const {
		assert(pNode);
		ostringstream oss;
		unsigned int index=0;
		TedNode* pKid=NULL;
		wedge weight=0;
		oss << "\n\t{rank = same; Level" << getLevel(pNode)<< "; Node_" << pNode << " [shape=record,";
		oss << "label=\" {";
		oss << pNode << " | { {";
		oss << pNode->getName();
		if(pNode->getVar()->isConst())
			oss << "(C)";
		else if(pNode->getVar()->isVar())
			oss << "(V)";
		else
			oss << "(VC)";
		if (pNode->getVar()->isBinary()) {
			oss << "(B)";
		}
		oss << "\\n";
		oss << "b=" <<(pNode->getVar()->getBitwidth() ? pNode->getVar()->getBitwidth()->at(): "X")<< "\\n";
		oss << pNode->getVar()<< "} | {";
		oss <<(pNode->visited.isMarked() ? "M" : "U")<< pNode->visited.getMark();
		oss << " | r=" << pNode->getRegister();
		oss << " | b=" <<(pNode->getBitwidth() ? pNode->getBitwidth()->at(): "X");
		oss << "} }";
		if(NULL!=pNode->getBackptr()) {
			oss << " | ptr=" << pNode->getBackptr();
		}
		oss <<" } \"]; }\n";
		FOREACH_KID_OF_NODE(pNode) {
			index = _iterkids.getIndex();
			pKid = _iterkids.Node<TedNode>();
			weight = _iterkids.getWeight();
			oss << "\tNode_"<<pNode<<" -> Node_"<<pKid<<" [style="<<(index==0?"dashed":"solid");
			if (pKid==pTouch) {
				oss << ", color=\"red\"";
			} else if (pNode==pTouch) {
				oss << ", color=\"blue\"";
			}
			oss	<< ", label=\"";
			oss << "^,w,r,b\\n";
			oss << index <<"," << weight << "," << _iterkids->second.getRegister()<< ",";
			if(_iterkids->second.getBitwidth())
				oss << _iterkids->second.getBitwidth()->at();
			else
				oss << "X";
			oss << "\"];\n";
		}
		return oss.str();
	}
#endif
}
