/*
 * =====================================================================================
 *
 *        Filename:  TEDManWriteDot.cc
 *     Description:  Tedman class
 *         Created:  04/17/2007 12:24:18 PM EDT
 *          Author:  Daniel Gomez-Prado
 *         Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 213                                          $: Revision of last commit
 *  $Author:: daniel@win                                     $: Author of last commit
 *  $Date:: 2010-11-18 08:13:53 -0500(Thu, 18 Nov 2010)$: Date of last commit
 * =====================================================================================
 */

#include <iostream>
#include <sstream>
#include <fstream>
#include <cassert>
#include <algorithm>
#include <list>

#include "TedNode.h"
#include "TedKids.h"
#include "TedVar.h"
#include "ETedNode.h"
#include "TedVarGroup.h"
#include "TedVarMan.h"
#include "TedNodeIterator.h"
#include "TedBitwidth.h"

#include "Bitwidth.h"
using namespace data;

#include "BinaryThreadedTree.h"
using dtl::BinaryThreadedTree;
using dtl::BinaryThreadedNode;

#include "Environment.h"
using namespace tds;

#include "util.h"
using util::Util;

namespace ted {
	/** @brief The dot string of the node*/
	string TedNode::getDotStringSimple(size_t level, bool bigFont, bool compact_graph) const {
		const TedNode* pNode = this;
		string nodeFontSize = "";
		if(bigFont) {
			nodeFontSize = "fixedsize=\"true\", fontsize=18, ";
		}
		ostringstream oss;
		unsigned int index=0;
		TedNode* pKid=NULL;
		wedge weight=0;
		if(pNode->getRegister()> 0 || pNode->getBitwidth()) {
			oss << "\n\tsubgraph \"reg_" << pNode->obfuscate()<< "\" {";
			oss << "\n\trank = same; Level" << level <<"; \"node_reg_" << pNode->obfuscate()<< "\" [label=\"";
			if(pNode->getRegister()> 0)
				oss << pNode->getRegister()<< "R ";
			if(pNode->getBitwidth())
				oss << "b=" << pNode->getBitwidth()->at();
			oss << "\", shape = plaintext]";
		}
		oss << "\n\t{rank = same; Level" << level << "; \"Node_" << pNode->obfuscate()<< "\" [";
		oss << nodeFontSize << "label=\"";
		if(compact_graph && pNode->getName().length()> 4)
			oss << " ";
		else
			oss << pNode->getName();
		oss << "\", ";
		oss	<< "shape=" <<(pNode->isConst() ?"triangle":"ellipse")<< "]; }\n";
		if(pNode->getRegister()> 0 || pNode->getBitwidth()) {
			oss << "\n\t\"Node_" << pNode->obfuscate()<< "\" -> \"node_reg_" << pNode->obfuscate()<< "\" [arrowhead=inv, minlen=0]";
			oss << "\n\t}";
		}
		FOREACH_KID_OF_NODE(pNode) {
			index = _iterkids.getIndex();
			pKid = _iterkids.Node<TedNode>();
			weight = _iterkids.getWeight();
			oss << "\t\"Node_"<<pNode->obfuscate()<<"\" -> \"Node_"<<pKid->obfuscate()<<"\" [style="<<(index==0?"dashed":"solid");
			oss	<< ", label=\"";
			if(index>1) {
				oss << "^"<<index<<" ";
			}
			if(weight!=1) {
				oss << weight;
			}
			if(_iterkids->second.getRegister()!=0) {
				oss << "\\n---" << _iterkids->second.getRegister()<< "R---";
			}
			if(_iterkids->second.getBitwidth()) {
				oss << "b=" << _iterkids->second.getBitwidth()->at();
			}
			oss << "\"];\n";
		}
		return oss.str();
	}

	string TedNode::obfuscate(void) const {
		const TedNode* ptr = this;
		ostringstream oss;
		string name(ptr->getName());
#ifdef _DEBUG
		oss << name << "_" << ptr;
#else
		const size_t distance = sizeof(TedNode);
		const size_t val = ((size_t)ptr)/distance;
		oss << name << "_" << hex << val;
#endif
		return oss.str();
	}

	/** @brief The dot string of the node*/
	string TedNode::getDotStringDetailed(size_t level, const TedNode* pTouch, bool useLevel) const {
		const TedNode* pNode = this;
		assert(pNode);
		ostringstream oss;
		unsigned int index=0;
		TedNode* pKid=NULL;
		wedge weight=0;
		oss << "\n\t{";
		if (useLevel) {
			oss << "rank = same; Level" << level << ";"; 
		}
		oss << "Node_" << pNode << " [shape=record,";
		oss << "label=\" {";
		oss << pNode << " | { {";
		oss << pNode->getName();
		if(pNode->getVar()->isConst())
			oss << "(C)";
		else if(pNode->getVar()->isVar())
			oss << "(V)";
		else
			oss << "(VC)";
		if (pNode->getVar()->isBinary()) {
			oss << "(B)";
		}
		oss << "\\n";
		oss << "b=" <<(pNode->getVar()->getBitwidth() ? pNode->getVar()->getBitwidth()->at(): "X")<< "\\n";
		oss << pNode->getVar()<< "} | {";
		oss <<(pNode->visited.isMarked() ? "M" : "U")<< pNode->visited.getMark();
		oss << " | r=" << pNode->getRegister();
		oss << " | b=" <<(pNode->getBitwidth() ? pNode->getBitwidth()->at(): "X");
		oss << "} }";
		if(NULL!=pNode->getBackptr()) {
			oss << " | ptr=" << pNode->getBackptr();
		}
		oss <<" } \"]; }\n";
		FOREACH_KID_OF_NODE(pNode) {
			index = _iterkids.getIndex();
			pKid = _iterkids.Node<TedNode>();
			weight = _iterkids.getWeight();
			oss << "\tNode_"<<pNode<<" -> Node_"<<pKid<<" [style="<<(index==0?"dashed":"solid");
			if (pKid==pTouch) {
				oss << ", color=\"red\"";
			} else if (pNode==pTouch) {
				oss << ", color=\"blue\"";
			}
			oss	<< ", label=\"";
			oss << "^,w,r,b\\n";
			oss << index <<"," << weight << "," << _iterkids->second.getRegister()<< ",";
			if(_iterkids->second.getBitwidth())
				oss << _iterkids->second.getBitwidth()->at();
			else
				oss << "X";
			oss << "\"];\n";
		}
		return oss.str();
	}

}
