/*
 * =====================================================================================
 *
 *       Filename:  TedParents.cc
 *    Description:
 *        Created:  08/21/2011 11:38:24 PM
 *         Author:  Daniel F. Gomez-Prado (dgomezpr), dgomezpr@ecs.umass.edu
 *        Company:  UMASS
 *
 * =====================================================================================
 *  $Revision::                                              $: Revision of last commit
 *  $Author::                                                $: Author of last commit
 *  $Date::                                                  $: Date of last commit
 * =====================================================================================
 */

#include <algorithm>

using namespace std;

#include "TedParents.h"
#include "TedMan.h"

namespace ted {


	void TedParents::print_parents(void) const {
		for(TedParents::const_iterator it = this->begin(), it_end = this->end(); it != it_end; it++) {
			TedNode* node = it->first;
			MyParent* parents = it->second;
			print_parents(node,parents);
		}
	}

	void TedParents::print_parents(TedNode* pNode) const {
		TedParents::const_iterator it  = find(pNode);
		if (it != end()) {
			TedNode* node = it->first;
			MyParent* parents = it->second;
			print_parents(node,parents);
		}
	}

	void TedParents::print_parents(TedNode* node, MyParent* parents) const {
		cout << " ===== " << node << " = " << parents->size() << " ==== " <<endl;
		for(MyParent::const_iterator pit = parents->begin(), pit_end = parents->end(); pit != pit_end; pit++) {
			TedNode* parent = pit->first;
			MyIndex* index = pit->second;
			cout << parent << " (" << index->size() << ") ";
			for(MyIndex::const_iterator jt = index->begin(), jt_end = index->end(); jt!=jt_end; jt++) {
				cout <<*jt << ",";
			}
			cout << endl;
		}
		cout << " ========================================" << endl;
	}

	void TedParents::print_output_cone(TedNode* pNode,const TedMan* tMan) const {
		const TedMan::PrimaryOutputs& pos = tMan->getPOs();
		for(TedMan::PrimaryOutputs::const_iterator p = pos.begin(); p != pos.end(); p++) {
			if(pNode == p->second.Node()) {
				cout << "Output cone = " << p->first << endl;
				return;
			}
		}
		if (!hasParents(pNode)) { 
			cout << "missing parent at Node " << pNode << endl;
			return;
		}
		MyParent* mp = this->at(pNode);
		if (!mp) {
			cout << "empty parent container for node " << pNode << endl;
			return;
		} 
		for(MyParent::const_iterator pit = mp->begin(), pit_end = mp->end(); pit != pit_end; pit++) {
			TedNode* parent = pit->first;
			print_output_cone(parent,tMan);
		}				
	}

	bool TedParents::hasAllParents(TedNode* pNode, const TedMan* tMan) const {
		bool ret = true;
		if (!tMan->isPO(pNode)) {
			if (!hasParents(pNode)) { 
				ret = false;
			} else {
				MyParent* mp = this->at(pNode);
				if (!mp) {
					ret = false;
				} else {
					for(MyParent::const_iterator pit = mp->begin(), pit_end = mp->end(); pit != pit_end; pit++) {
						TedNode* parent = pit->first;
						if (!hasAllParents(parent,tMan)) {
							ret = false;
							break;
						}
					}
				}
			}
		}
		return ret;
	}

	bool TedParents::hasSiblings(TedNode* pNode) const {
		TedParents::const_iterator it  = find(pNode);
		if (it != end()) {
			MyParent* mp = it->second;
			for(MyParent::const_iterator pit = mp->begin(), pit_end = mp->end(); pit != pit_end; pit++) {
				TedNode* parent = pit->first;
				FOREACH_KID_OF_NODE(parent) {
					if (pNode!=_iterkids.Node<TedNode>()) {
						return true;
					}
				}
			}
		} 
		return false;
	}
}