/*
 * =====================================================================================
 *
 *        Filename:  TedTypes.cc
 *     Description:  TedNode class header
 *         Created:  04/17/2012 12:13:04 PM EDT
 *          Author:  Daniel Gomez-Prado
 *         Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 215                                          $: Revision of last commit
 *  $Author:: daniel@win                                     $: Author of last commit
 *  $Date:: 2010-01-12 14:44:42 +0100(Tue, 12 Jan 2010)$: Date of last commit
 * =====================================================================================
 */

#include <iostream>
#include <sstream>
#include <fstream>
#include <cassert>
#include <algorithm>
#include <list>

#include "TedTypes.h"
#include "TedNode.h"
#include "TedMan.h"
#include "TedContainer.h"
#include "TedKids.h"
#include "TedVarMan.h"
#include "TedNodeIterator.h"
#include "TedParents.h"

#include "Bitwidth.h"
using namespace data;

#include "DfgNode.h"
#include "DfgMan.h"
using dfg::DfgMan;
using dfg::DfgNode;
using dfg::DfgStat;

#include "Environment.h"
using namespace tds;

#include "util.h"
using util::Util;

namespace ted {

#if defined(_DEBUG)
	void TEDSet::visualize(void) {
		static unsigned int index = 0;
		ofstream ofile;
		string filename = Environment::getStr("show_directory")+ "debugTEDSet" + Util::itoa(index++)+ ".dot";
		ofile.open(filename.c_str());
		if(!ofile.is_open()) {
			throw(string("04014. Cannot open temporary file for writing.\n"));
		}
		writeTEDSet(ofile,NULL);
		ofile.close();
		Util::showDotFile(filename.c_str());
	}

	/** @brief Show Dot File*/
	void TEDSet::writeTEDSet(ofstream& ofile, const TedNode* pTouch) {
		ofile << "Digraph G {\n\tsize=\"10, 7.5\"; center = \"true\"; \n";
		ofile << "\t{\n";
		ofile << "\t\tnode [shape = plaintext];\n";
		ofile << "\t\tedge [style = invis];\n";
		ofile << "\n\t}\n";
		for(TEDSet::iterator sp = begin(); sp != end(); sp++) {
			ofile << (*sp)->getDotStringDetailed(1,pTouch);
		}
		ofile << "\n\t{rank = same; Level0; Node_" << TedNode::getOne();
		ofile << " [label=\"ONE\", shape=triangle]; }\n";
		ofile << "}";
	}

	void TEDSet::visualize(const TedNode* pTouch, TedParents* parents, unsigned int link_parents, unsigned int link_children) {
		static unsigned int index = 0;
		ofstream ofile;
		string filename = Environment::getStr("show_directory")+ "debugTEDSetWindow" + Util::itoa(index++)+ ".dot";
		ofile.open(filename.c_str());
		if(!ofile.is_open()) {
			throw(string("04014. Cannot open temporary file for writing.\n"));
		}
		writeWindow(ofile,pTouch,parents,link_parents,link_children);
		ofile.close();
		Util::showDotFile(filename.c_str());
	}

	void TEDSet::writeWindow(ofstream& ofile, const TedNode* pTouch, TedParents* parents, unsigned int link_parents, unsigned int link_children) {
		ofile << "Digraph G {\n\tsize=\"10, 7.5\"; center = \"true\"; \n";
		ofile << "\t{\n";
		ofile << "\t\tnode [shape=box, style=filled fillcolor=green];\n";
		ofile << "\t\tedge [style = invis];\n";
		ofile << "\t\tLevel1 [label=\"Current Set\"];\n";
		ofile << "\n\t}\n";

		unsigned int set_level = link_children + 1;
		for(TEDSet::iterator sp = begin(); sp != end(); sp++) {
			ofile << "//\tUp to " << link_parents << " parents of node " << (*sp) << endl;
			writeUpper(ofile,(*sp),pTouch,parents,set_level+1,set_level+link_parents);
			ofile << "//\tThe node " << (*sp) << " itself" << endl;
			ofile << (*sp)->getDotStringDetailed(1,pTouch,true);
			ofile << "//\tDown to " << link_children << " children of node " << (*sp) << endl;
			writeLower(ofile,(*sp),pTouch,set_level-1,0);
		}
		ofile << "\n\t{Node_" << TedNode::getOne() << " [label=\"ONE\", shape=triangle]; }" << endl;
		ofile << "}";
	}

	void TEDSet::writeUpper(ofstream& ofile, const TedNode* pNode, const TedNode* pTouch, TedParents* parents, unsigned int index_up, const unsigned int stop) {
		if (index_up>stop) {
			return;
		}
		assert(pNode);
		TedParents& parentRef = *parents;
		if (parentRef.hasParents(const_cast<TedNode*>(pNode))) {
			MyParent* mp = parentRef[const_cast<TedNode*>(pNode)];
			if (mp) {
				for(MyParent::const_iterator it = mp->begin(), end = mp->end(); it != end; it++) {
					const TedNode* topParent = it->first;
					writeUpper(ofile,topParent,pTouch,parents,index_up+1,stop);
					ofile << topParent->getDotStringDetailed(index_up,pTouch,false);
				}
			} else {
				ofile << "MISSING_PARENT [style=solid, color=orange];"<< endl;
				ofile << "MISSING_PARENT -> Node_" << pNode << " [style=solid, color=orange];"<< endl;
			}
		} else {
			ofile << "No_Parent [style=solid, color=orange];" << endl;
			ofile << "\tNo_Parent -> Node_"<<pNode<<" [style=solid, color=\"orange\"];"<< endl;
		}
	}

	void TEDSet::writeLower(ofstream& ofile, const TedNode* pNode, const TedNode* pTouch, unsigned int index_down, const unsigned int stop) {
		if (index_down==stop || pNode==TedNode::getOne()) {
			return;
		}
		assert(pNode);
		TedNode* pKid = NULL;
		FOREACH_KID_OF_NODE(pNode) {
			pKid =_iterkids.Node<TedNode>();
			if (pKid) {
				ofile << pKid->getDotStringDetailed(index_down,pTouch,false);
				writeLower(ofile,pKid,pTouch,index_down-1,stop);
			} else {
				ofile << "Dangling_Child [style=solid, color=orange];"<<endl;
				ofile << "\t Node_"<<pNode<<" -> Dangling_Child [style=solid, color=\"orange\"];"<< endl;
			}
		}
	}

	void TEDSet::visualizeUpperCone(const TedNode* pCone, TedParents* parents, unsigned int link_parents, unsigned int link_children) {
		static unsigned int index = 0;
		ofstream ofile;
		string filename = Environment::getStr("show_directory")+ "debugTEDSetCone" + Util::itoa(index++)+ ".dot";
		ofile.open(filename.c_str());
		if(!ofile.is_open()) {
			throw(string("04014. Cannot open temporary file for writing.\n"));
		}
		writeCone(ofile,pCone,parents,link_parents,link_children);
		ofile.close();
		Util::showDotFile(filename.c_str());
	}

	void TEDSet::writeCone(ofstream& ofile, const TedNode* pCone, TedParents* parents, unsigned int link_parents, unsigned int link_children) {
		ofile << "Digraph G {\n\tsize=\"10, 7.5\"; center = \"true\"; \n";
		ofile << "\t{\n";
		ofile << "\t\tnode [shape=box, style=filled fillcolor=green];\n";
		ofile << "\t\tedge [style = invis];\n";
		ofile << "\t\tLevel1 [label=\"Current Set\"];\n";
		ofile << "\n\t}\n";

		unsigned int set_level = link_children + 1;

		ofile << "//\tUp to " << link_parents << " parents of node " << pCone << endl;
		writeUpper(ofile,pCone,NULL,parents,set_level+1,set_level+link_parents);
		ofile << "//\tThe node " << pCone << " itself" << endl;
		ofile << pCone->getDotStringDetailed(1,NULL,true);
		ofile << "//\tDown to " << link_children << " children of node " << pCone << endl;
		writeLower(ofile,pCone,NULL,set_level-1,0);

		ofile << "\n\t{Node_" << TedNode::getOne() << " [label=\"ONE\", shape=triangle]; }" << endl;
		ofile << "}";
	}

	void TEDSet::print2file(void) {
		static unsigned int index = 0;
		string filename = "tedset_container";
		filename += Util::itoa(index++);
		filename += ".txt";
		ofstream out(filename.c_str());
		out << " node_pointer | Tmark | Pmark | variable_ptr | kid_pointer " << endl;
		unsigned tmarked = 0;
		unsigned untmarked = 0;
		unsigned pmarked = 0;
		unsigned unpmarked = 0;
		for(TEDSet::const_iterator it = begin(); it != end(); it++) {
			TedNode* pNode = (*it);
			out << " ";
			out.width(12);
			out.fill(' ');
			out << pNode << " | ";
			if (pNode->visited.isMarked()) {
				tmarked++;
				out << " MM " << " | ";
			} else {
				untmarked++;
				out << " UU " << " | ";
			}
			if (pNode->visited.isPMarked()) {
				pmarked++;
				out << " MM " << " | ";
			} else {
				unpmarked++;
				out << " UU " << " | ";
			}
			out.width(12);
			out.fill(' ');
			out << pNode->getVar() << " | ";
			out.width(11);
			out.fill(' ');
			out << pNode->getKids();
			out << " | " << hex << pNode->safe_guard;
			out << dec;
			out << endl;
		}
		out << "size=" << size()<< endl;
		out << "tmarked=" << tmarked << "  untmarked=" << untmarked << endl;
		out << "pmarked=" << pmarked << "  unpmarked=" << unpmarked << endl;
		out.close();
		//	cout << "ATedNode report in transitory container printed" << endl;
	}

	void TEDSet::print2fileDetailed(void) {
		static unsigned int index = 0;
		string filename = "tedset_detailed";
		filename += Util::itoa(index++);
		filename += ".txt";
		ofstream out(filename.c_str());
		out << " node_pointer | var_name | const / kid_pointer " << endl;
		for(TEDSet::const_iterator it = begin(); it != end(); it++) {
			TedNode* pNode = (*it);
			out << " ";
			out.width(12);
			out.fill(' ');
			out << pNode << " | ";			
			out.width(12);
			out.fill(' ');
			out << pNode->getVar()->getName() << " | ";
			out.width(11);
			out.fill(' ');
			if (pNode->isConst()) {
				out << "value=" << pNode->getConst();
			} else if (pNode->isZero()) {
				out << "isZero";
			} else if (pNode->isDummy())  {
				out << "[i=0,w="<<pNode->getKidWeight(0) << ",k=" <<pNode->getKidNode(0) << "]";
			} else {
				out << "  [";
				FOREACH_KID_OF_NODE(pNode) {
					out << "i="<<_iterkids.getIndex() << ",w=" << _iterkids.getWeight() << ",k=" <<  _iterkids.Node<TedNode>() << "][";
				}
				out << "]";
			}
			out << endl;
		}
		out << "size=" << size()<< endl;
		out.close();
		//	cout << "ATedNode report in transitory container printed" << endl;
	}
#else
	namespace {
		const bool Avoid_empty_translation_unit = true;
	}
#endif

}
